/*
 * emdf_output.cpp
 *
 * Class to output something either in XML or in console-format.
 *
 * Ulrik Petersen
 * Created: 10/21-2001
 * Last update: 6/21-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/



#include <emdf_output.h>
#include <string_func.h>
#include <cassert>

static const char* s_charset_names[] = {
  "UTF-8",       // kCSASCII
  "ISO-8859-1",  // kCSISO_8859_1
  "ISO-8859-8",  // kCSISO_8859_8
  "UTF-8",       // kCSUTF8
  "ISO-8859-2",  // kCSISO_8859_2
  "ISO-8859-3",  // kCSISO_8859_3
  "ISO-8859-4",  // kCSISO_8859_4
  "ISO-8859-5",  // kCSISO_8859_5
  "ISO-8859-6",  // kCSISO_8859_6
  "ISO-8859-7",  // kCSISO_8859_7
  "ISO-8859-9",  // kCSISO_8859_9
  "ISO-8859-10", // kCSISO_8859_10
  "ISO-8859-13", // kCSISO_8859_13
  "ISO-8859-14", // kCSISO_8859_14
  "ISO-8859-15", // kCSISO_8859_15
  "ISO-8859-16"  // kCSISO_8859_16

  // NOTE: I tried including UTF-16, but Solaris doesn't do 
  // std::ostream << std::wstring, so I dropped the idea.
};

void CEMdFOutput::emit_attributes(const AttributePairList& attributes)
{
  AttributePairList::const_iterator ci(attributes.begin());
  AttributePairList::const_iterator cend(attributes.end());
  while (ci != cend) {
    // Emit name
    out(" " + ci->first + "=");

    // Figure out delimiter.
    // If the string to output has a `"', delimiter is `''.
    // Otherwise, it is `"'.
    std::string str_delimiter;
    if (has_double_quote(ci->second)) {
      str_delimiter = "\'";
    } else {
      str_delimiter = "\"";
    }

    // output value
    out(str_delimiter);
    out_char_data(ci->second);
    out(str_delimiter);

    // Advance iterator
    ++ci;
  }
}

////////////////////////////////////////////////////////
//
// CEMdFOutput
//
///////////////////////////////////////////////////////

CEMdFOutput::CEMdFOutput(eCharsets charset, std::ostream *pStream, eOutputKind output_kind, int indent_chars) : m_charset(charset), 
  m_output_kind(output_kind),
  m_pStream(pStream), 
  m_current_indent(0), 
  m_indent_chars(indent_chars)
{
  // This causes the C++ library (at least on Linux) not to interleave
  // C and C++ output, which causes a performance increase (at least
  // on Linux).  See B. Stroustrup, "The C++ Programming language",
  // 3rd edition, p. 651.
  std::ios::sync_with_stdio(false);
}

CEMdFOutput::~CEMdFOutput()
{
  // Don't destroy pStream!  It could be cout!

  // But flush stream so we don't have anything pending
  flush();
}

void CEMdFOutput::increase_indent()
{
  ++m_current_indent;
}

void CEMdFOutput::decrease_indent()
{
  m_current_indent--;
  assert(m_current_indent >= 0);
}

void CEMdFOutput::out(std::string s)
{
  *m_pStream << s;
}

// Convert '&' to '&amp;' 
//     and '<' to '&lt;'
// before out'ing. 
// Useful for outputting character data which may contain these
// characters.
void CEMdFOutput::out_char_data(std::string s)
{
  std::string::const_iterator p(s.begin());
  std::string::const_iterator pend(s.end());
  std::string sout;
  while (p != pend) {
    if (*p == '&') {
      sout += "&amp;";
    } else if (*p == '<') {
      sout += "&lt;";
    } else {
      sout += *p;
    }
    ++p;
  }
  out(sout);
}

void CEMdFOutput::newline()
{
  int spaces = m_current_indent * m_indent_chars;
  // Output newline
  *m_pStream << '\n';

  // NOTE: The following also flushes the buffer (on Linux, at least)
  // therefore, this is slower than writing a '\n' as above.
  // *m_pStream << std::endl;

  // Output spaces for indent
  for (int i = 0; i < spaces; ++i) {
    *m_pStream << ' ';
  }
}

void CEMdFOutput::print_XMLDecl()
{
  out("<?xml "); 
  out("version='1.0' ");
  out("encoding='");
  out(s_charset_names[m_charset]);
  out("' ");
  out("standalone='yes' ");
  out("?>");

}

void CEMdFOutput::print_DTD_start(std::string root_element)
{
  out("<!DOCTYPE ");
  out(root_element);
  out(" [");
  newline();
}

void CEMdFOutput::print_DTD_end()
{
  out("]>");
  newline();
}


void CEMdFOutput::start_tag(std::string name, bool newline_before)
{
  if (newline_before)
    newline();
  increase_indent();
  out("<");
  out(name);
  out(">");
}

// Must have pairs of (attribute name, attribute value) 
void CEMdFOutput::start_tag(std::string name, const AttributePairList& attributes, bool newline_before)
{
  if (newline_before)
    newline();
  increase_indent();
  out("<");
  out(name);
  emit_attributes(attributes);
  out(">");
}

// for tags of type <tag/>
void CEMdFOutput::start_single_tag(std::string name, bool newline_before) 
{
  if (newline_before)
    newline();
  out("<");
  out(name);
  out("/>");
}

// for tags of type <tag/>
void CEMdFOutput::start_single_tag(std::string name, const AttributePairList& attributes, bool newline_before) 
{
  if (newline_before)
    newline();
  out("<");
  out(name);
  emit_attributes(attributes);
  out("/>");
}

void CEMdFOutput::end_tag(std::string name, bool newline_before)
{
  decrease_indent();
  if (newline_before)
    newline();
  out("</");
  out(name);
  out(">");
}


