/*
 * mysqlconn.cpp
 *
 * MySQL Database connection 
 *
 * Ulrik Petersen
 * Created: 1/27-2001
 * Last update: 6/24-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#ifdef _WIN32
#include <windows.h>
#endif

#include <mysqlconn.h>
#include <iostream>
#include <cstring>
#include <logging.h>
#include <utils.h>


MySQLEMdFConnection::MySQLEMdFConnection(const std::string& host, const std::string& user, const std::string& passwd, const std::string& db)
{
  // We aren't doing any transaction
  bTransactionInProgress = false;

  // We don't have any result
  pResultSet = NULL;
  m_Row = NULL;

  // Get db
  if (db.length() == 0) {
    m_db = "emdf";
  } else {
    ConvertDBName(db, m_db);
  }


  // Make MYSQL structure
  pConn = mysql_init(NULL);


  // Connect
  MYSQL *res;
  if ((res = mysql_real_connect(pConn, host.c_str(), user.c_str(), passwd.c_str(), m_db.c_str(), 0, NULL, 0)) != pConn) {
    std::cerr << "Error: could not connect.  Return value was " << res << "." << std::endl;
  }

  
#ifdef EMDF_DEBUG
  if (!ConnectionOk()) {
    std::cerr << "mysqlconn.cpp: Connection not OK connecting to database '"
	      << db << "'." << std::endl;
    std::string error_message;
    bool res_err;
    res_err = ErrorMessage(error_message);
    std::cout << "Error message = '" << error_message << "'\n";
    std::cout << "res_err = " << res_err << "\n";
  }
#endif // EMDF_DEBUG
}

MySQLEMdFConnection::~MySQLEMdFConnection()
{
  // Commit transaction if in progress
  if (bTransactionInProgress)
    CommitTransaction();

  // Do away with result set
  FreeResult();

  // Close the connection and deallocate the MYSQL structure
  mysql_close(pConn);
}


// Status of connection
// Returns true on connection good, false on connection bad
// Should be called right after constructor to see if we
// succeeded in setting up the connection.
bool MySQLEMdFConnection::ConnectionOk(void)
{
  if (pConn == NULL)
    return false;
  else {
    return mysql_ping(pConn) == 0;
  }
}


// Executing
bool MySQLEMdFConnection::Exec(const std::string& query)
{
  if (pConn == 0)
    return false;
  else {
    // Do away with previous pResultSet
    FreeResult();

    // Write to log
    LOG_WRITE_TIME("MySQLEMdFConnection::Exec", std::string("Starting query.\n") + query);

    // Execute query
    if (mysql_query(pConn, query.c_str()) != 0) {
      // Write to log
      LOG_WRITE_TIME("MySQLEMdFConnection::Exec", "Query finished with failure.");

      // Error: mysql_query did not return 0.
      return false;
    } else {
      // No error

      // Write to log
      LOG_WRITE_TIME("MySQLEMdFConnection::Exec", "Query finished with success.");

      // Store result
      LOG_WRITE_TIME("MySQLEMdFConnection::Exec", "Storing result...");
      pResultSet = mysql_store_result(pConn);
      LOG_WRITE_TIME("MySQLEMdFConnection::Exec", "Storing result finished.");
      if (pResultSet != NULL) {
	// There are rows

	// Get number of fields
	m_num_fields = mysql_field_count(pConn);

	// Fetch first row
	m_current_tuple_no = 0;
	mysql_data_seek(pResultSet, 0);
	m_Row = mysql_fetch_row(pResultSet);

	// Return success
	return true;
      } else {
	if (mysql_field_count(pConn) == 0) {
	  // Query does not contain data (it was not a SELECT)
	  m_command_rows = mysql_affected_rows(pConn);

	  // Return success
	  return true;
	} else {
#ifdef EMDF_DEBUG
	  std::cerr << "mysqlconn.cpp: The following query should have returned results, but didn't:" << std::endl;
	  std::cerr << "---------------------------------" << std::endl;
	  std::cerr << query << std::endl;	
	  std::cerr << "---------------------------------" << std::endl;
#endif // EMDF_DEBUG

	  // Return failure
	  return false;
	}
      }
    }
  }
}


// Get last error message
// Returns true on error message available 
// (even if it is "OK"), false on error message not available.
bool MySQLEMdFConnection::ErrorMessage(std::string& out)
{
  if (pConn == 0) {
    out = "not connected.";
    return true;
  } else {
    out = mysql_error(pConn);

    if (out.length() > 0)
      return true;
    else
      return false;
  }
}


// Transactions
bool MySQLEMdFConnection::BeginTransaction(void)
{
  if (pConn == 0)
    return false;
  else {
    return false; // Transactions not supported
  }
}

bool MySQLEMdFConnection::CommitTransaction(void)
{
  if (pConn == 0)
    return false;
  else {
    std::cerr << "Warning: MySQLEMdFConnection::CommitTransaction: Transactions not supported." << std::endl << "... Something is wrong in the logic! ..." << std::endl;
    return false;
  }
}

bool MySQLEMdFConnection::AbortTransaction(void)
{
  if (pConn == 0)
    return false;
  else {
    std::cerr << "Warning: MySQLEMdFConnection::AbortTransaction: Transactions not supported." << std::endl << "... Something is wrong in the logic! ..." << std::endl;
    return false;
  }
}


// Tuple-counts
// Both return -1 on not appliccable or error
long MySQLEMdFConnection::NoOfResultTuples(void)  // SELECT
{
  if (pConn == 0)
    return -1;
  else {
    if (pResultSet == NULL)
      return -1;
    else {
      return (long) mysql_num_rows(pResultSet);
    }
  }
}

long MySQLEMdFConnection::NoOfCommandTuples(void) // INSERT, UPDATE, DELETE
{
  if (pConn == 0)
    return -1;
  else {
    return m_command_rows;
  }
}


// Field count
// Returns -1 on no fields.
int MySQLEMdFConnection::NoOfFields(void)
{
  if (pConn == 0)
    return -1;
  else {
    if (pResultSet == NULL)
      return -1;
    else {
      return (int) mysql_num_fields(pResultSet);
    }
  }
}


bool MySQLEMdFConnection::UseDatabase(const std::string& db_name)
{
  if (pConn == 0)
    return false;
  else {
    int result = mysql_select_db(pConn, db_name.c_str());
    FreeResult();
    return result == 0;
  }
}

void MySQLEMdFConnection::ReleaseResults(void)
{
  FreeResult();
}

void MySQLEMdFConnection::FreeResult(void)
{
  if (pResultSet != NULL) {
    mysql_free_result(pResultSet);
    pResultSet = NULL;
    m_Row = NULL;
  }
}
