/*
 * table.cpp
 *
 * Tables
 *
 * Ulrik Petersen
 * Created: 5/5-2001
 * Last update: 6/21-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <table.h>
#include <string_func.h>
#include <cassert>
#include <iostream>

/////////////////////////////////////////////////////////////////
//
// TableRowIterator
//
/////////////////////////////////////////////////////////////////

TableRowIterator::TableRowIterator(TableRow *pMotherTableRow)
  : m_pMotherTableRow(pMotherTableRow)
{
  m_iter = m_pMotherTableRow->begin();
}

TableRowIterator::TableRowIterator()
  :  m_pMotherTableRow(0)
{
}

TableRowIterator::TableRowIterator(const TableRowIterator& other)
{
  assign(other);
}


TableRowIterator::~TableRowIterator()
{
}

TableRowIterator& TableRowIterator::operator=(const TableRowIterator& other)
{
  assign(other);
  return *this;
}

void TableRowIterator::assign(const TableRowIterator& other)
{
  m_pMotherTableRow = other.m_pMotherTableRow;
  m_iter = other.m_iter;
}


bool TableRowIterator::hasNext() const
{
  assert(m_pMotherTableRow != 0);
  return m_iter != m_pMotherTableRow->end();
}


std::string TableRowIterator::next()
{
  assert(m_pMotherTableRow != 0);
  assert(hasNext());
  std::string strResult = *m_iter;
  ++m_iter;
  return strResult;
}

std::string TableRowIterator::previous()
{
  assert(m_pMotherTableRow != 0);
  --m_iter;
  return *m_iter;
}

std::string TableRowIterator::current()
{
  assert(m_pMotherTableRow != 0);
  return *m_iter;
}

std::string TableRowIterator::getColumn(int column)
{
  assert(m_pMotherTableRow != 0);
  std::list<std::string>::iterator iter, iterend;
  iter = m_pMotherTableRow->begin();
  iterend = m_pMotherTableRow->end();
  int nCol = 1;
  while (nCol != column && iter != iterend) {
    ++nCol;
    ++iter;
  }
  return *iter;
}






/////////////////////////////////////////////////////////////////
//
// TableRow
//
/////////////////////////////////////////////////////////////////

TableRow::TableRow()
{
}

TableRow::TableRow(std::list<std::string> other)
  : std::list<std::string>(other)
{
}


TableRow::~TableRow()
{
}


TableRowIterator TableRow::iterator()
{
  return TableRowIterator(this);
}


std::string TableRow::getColumn(int column)
{
  return TableRowIterator(this).getColumn(column);
}


int TableRow::size() const
{
  return ((std::list<std::string>*) this)->size();
}



/////////////////////////////////////////////////////////////////
//
// TableIterator
//
/////////////////////////////////////////////////////////////////

TableIterator::TableIterator()
  : m_pMotherTable(0)
{
}

TableIterator::TableIterator(CTable *pMotherTable)
  : m_pMotherTable(pMotherTable)
{
  m_iter = m_pMotherTable->m_list->begin();
}


TableIterator::TableIterator(const TableIterator& other)
{
  assign(other);
}


TableIterator::~TableIterator()
{
}

TableIterator& TableIterator::operator=(const TableIterator& other)
{
  assign(other);
  return *this;
}

void TableIterator::assign(const TableIterator& other)
{
  m_pMotherTable = other.m_pMotherTable;
  m_iter = other.m_iter;
}


bool TableIterator::hasNext() const
{
  assert(m_pMotherTable != 0);
  return m_iter != m_pMotherTable->m_list->end();
}


TableRow& TableIterator::next()
{
  assert(m_pMotherTable != 0);
  TableRow& trResult = *m_iter;
  ++m_iter;
  return trResult;
}


TableRow& TableIterator::previous()
{
  assert(m_pMotherTable != 0);
  --m_iter;
  return *m_iter;
}

TableRow& TableIterator::current()
{
  return *m_iter;
}


std::string TableIterator::next(int column)
{
  assert(m_pMotherTable != 0);
  assert(hasNext());
  std::string strResult = current().getColumn(column);
  next();
  return strResult;
}


std::string TableIterator::getColumn(int column)
{
  assert(m_pMotherTable != 0);
  assert(hasNext());
  return current().getColumn(column);
}






/////////////////////////////////////////////////////////////////
//
// CTable
//
/////////////////////////////////////////////////////////////////


std::string TCT2String(TableColumnType tc_type)
{
  std::string strResult;
  switch (tc_type) {
  case kTCString:
    strResult = "string";
    break;
  case kTCInteger:
    strResult = "integer";
    break;
  case kTCID_D:
    strResult = "id_d";
    break;
  case kTCMonad_m:
    strResult = "monad_m";
    break;
  case kTCBool:
    strResult = "bool";
    break;
  case kTCEnum:
    strResult = "enum";
    break;
  default:
    std::cerr << "tc_type = " << tc_type << std::endl;
    assert(false);
    strResult = "";
  }
  return strResult;
}

// For lists
CTable::CTable()
{
  m_header_enum_types = new TableRow;
  m_header_types = new TCTList;
  m_list = new TableRowList;
  m_header_names = new TableRow;
}

CTable::~CTable()
{
  delete m_list;
  delete m_header_names;
  delete m_header_enum_types;
  delete m_header_types;
}

TableIterator CTable::iterator()
{
  return TableIterator(this);
}


//
// - Append str to the last list of strings 
//   in the list of lists of strings.
//
// - If the list of lists is empty, 
//   first make a new list of strings.
//
// - To use this function, call newline() when you
//   have added as many strings as are necessary 
//   for one tuple.
//
void CTable::append(const std::string& str)
{
  if (m_list->empty())
    newline();
  m_list->rbegin()->push_back(str);
}

//
// Append l to the list of lists of strings
//
void CTable::append(const std::list<std::string>& l)
{
  m_list->push_back((TableRow) l);
}

void CTable::append_header(std::string header_name, TableColumnType tc_type, std::string enum_type)
{
  m_header_names->push_back(header_name);
  m_header_types->push_back(tc_type);
  m_header_enum_types->push_back(enum_type);
}

//
// - Make an empty list of strings and append it
//   to the list of lists of strings
void CTable::newline()
{
  TableRow l;
  m_list->push_back(l);
}

//
// First column is 1.
// case sensitive according to whether bCompareCaseInsensitively is true or not
TableIterator CTable::find(const std::string& str, int column_index, bool bCompareCaseInsensitively)
{
  TableIterator ti = iterator();
  while (ti.hasNext()) {
    TableRowIterator tri = ti.current().iterator();
    bool found = false;
    int col_index = 1;
    while (tri.hasNext() && !found) {
      if (col_index > column_index)
	break;
      if (col_index == column_index) {
	if (bCompareCaseInsensitively) {
	  if (strcmp_nocase(tri.current(), str) == 0) {
	    found = true;
	  }
	} else {
	  if (strcmp_case(tri.current(), str) == 0) { 
	    found = true;
	  }
	}
      }
      ++col_index;
      tri.next();
    }
    if (found)
      break;
    ti.next();
  }

  return ti;
}

//
// Logical and
// First column is 1.
// not case sensitive.
TableIterator CTable::find(const std::string& str1, int column_index1, const std::string& str2, int column_index2, bool bCompareCaseInsensitively)
{
  TableIterator ti = iterator();
  while (ti.hasNext()) {
    TableRowIterator tri = ti.current().iterator();
    bool found = false;
    bool found1 = false;
    bool found2 = false;
    int col_index = 1;
    while (tri.hasNext() && !found) {
      if (col_index > column_index1 && col_index > column_index2)
	break;
      const std::string& strString = tri.current();

      if (col_index == column_index1) {
	if (bCompareCaseInsensitively) {
	  if (strcmp_nocase(strString, str1) == 0)
	    found1 = true;
	} else {
	  if (strcmp_case(strString, str1) == 0)
	    found1 = true;
	}
      }

      if (col_index == column_index2) {
	if (bCompareCaseInsensitively) {
	  if (strcmp_nocase(strString, str2) == 0)
	    found2 = true;
	} else {
	  if (strcmp_case(strString, str2) == 0)
	    found2 = true;
	}
      }

      if (found1 && found2)
	found = true;

      ++col_index;
      tri.next();
    }
    if (found)
      break;
    ti.next();
  }

  return ti;
}


void CTable::erase(TableIterator& d)
{
  m_list->erase(d.m_iter);
}

void CTable::update_row(TableIterator& i, const std::string new_value, int column_index)
{
  TableRowIterator tri = i.current().iterator();
  int col_index = 1;
  while (tri.hasNext()) {
    if (col_index == column_index) {
      (*(tri.m_iter)) = new_value;
      return;
    }
    ++col_index;
    tri.next();
  }
  std::cerr << "Error: table.cpp:CTable::update_row(): column_index '" << column_index << "' out of range. col_index = '" << col_index << "'." << std::endl;
  assert(false);
  return;
}

std::string CTable::get_column(TableIterator& i, int column_index)
{
  return i.current().getColumn(column_index);
}

long CTable::size() const
{
  return m_list->size();
}

void CTable::start_new_row()
{
  // Only add a new line if the list is not empty
  if (m_list->begin() != m_list->end()) {
    newline();
  }
}

void CTable::PrintConsole(CEMdFOutput* pOut) const
{
  std::list<std::string::size_type> max_lengths;
  TableRow headers;
  create_headers(headers);
  get_max_lengths_all(max_lengths, headers);
  dump_results_all(pOut, max_lengths, headers);
}

void CTable::create_headers(TableRow& headers) const
{
  // Make sure we are pristine
  headers.clear();

  //
  // Build headers
  //
  TableRow::const_iterator ci_header_names(m_header_names->begin());
  TableRow::const_iterator ci_header_names_end(m_header_names->end());
  TCTList::const_iterator ci_header_types(m_header_types->begin());
  TableRow::const_iterator ci_header_enum_types(m_header_enum_types->begin());
  while (ci_header_names != ci_header_names_end) {
    std::string header;

    // header-name
    header = *ci_header_names + " : ";

    // header-type
    header += TCT2String(*ci_header_types);

    // Possible enum-type
    if (*ci_header_types == kTCEnum) {
      header += "(" + *ci_header_enum_types +")";
    }

    // Push it back
    headers.push_back(header);

    // Advance iterators
    ++ci_header_names;
    ++ci_header_types;
    ++ci_header_enum_types;
  }
}

void emit_minuses(CEMdFOutput* pOut, std::string::size_type no_of_minuses)
{
  std::string minuses(no_of_minuses, '-');
  pOut->out(minuses);
}

void CTable::emit_horizontal_line(CEMdFOutput* pOut, const std::list<std::string::size_type>& max_lengths) const
{
  // Output leading "+"
  pOut->out("+");

  // Output each of the max_lengths
  std::list<std::string::size_type>::const_iterator ci(max_lengths.begin());
  std::list<std::string::size_type>::const_iterator cend(max_lengths.end());
  while (ci != cend) {
    // Emit minuses
    emit_minuses(pOut, *ci + 2);

    // Emit "+"
    pOut->out("+");

    // Advance iterator
    ++ci;
  }

  // Emit trailing newline
  pOut->newline();
}

void CTable::dump_results_all(CEMdFOutput* pOut, const std::list<std::string::size_type>& max_lengths, const TableRow& headers) const 
{
  // Emit horizontal line
  emit_horizontal_line(pOut, max_lengths);

  // Dump header
  dump_results(pOut, headers, max_lengths);

  // Emit horizontal line
  emit_horizontal_line(pOut, max_lengths);

  // Dump results
  TableRowList::const_iterator ci(m_list->begin());
  TableRowList::const_iterator cend(m_list->end());
  while (ci != cend) {
    dump_results(pOut, *ci, max_lengths);
    ++ci;
  }

  // Emit horizontal line
  emit_horizontal_line(pOut, max_lengths);
}

void CTable::dump_results(CEMdFOutput* pOut, const TableRow& strlist, const std::list<std::string::size_type>& max_lengths) const 
{
  // Output initial "|".
  pOut->out("|");

  // Declare and initialize iterators
  std::list<std::string::size_type>::const_iterator mli;
  TableRow::const_iterator sli;
  TableRow::const_iterator sliend;
  mli = max_lengths.begin();
  sli = strlist.begin();
  sliend = strlist.end();

  // Iterate through lists
  while (sli != sliend) {
    // Output leading " " and string
    pOut->out(" ");
    pOut->out(*sli);

    // Output padding
    std::string spaces(*mli - sli->length() + 1, ' ');
    pOut->out(spaces);

    // Output trailing "|"
    pOut->out("|");

    // Advance iterators
    ++mli;
    ++sli;
  }

  // Output trailing endl
  pOut->newline();
}


void CTable::get_max_lengths(const TableRow& strlist, std::list<std::string::size_type>& max_lengths) const
{
  // Resize max_lengths if necessary to add members
  if (max_lengths.size() < strlist.size()) {
    max_lengths.resize(strlist.size(), 0);
  }

  // Iterate over strlist
  std::list<std::string::size_type>::iterator mli;
  TableRow::const_iterator sli;
  TableRow::const_iterator sliend;
  mli = max_lengths.begin();
  sli = strlist.begin();
  sliend = strlist.end();
  while (sli != sliend) {
    if (sli->length() > *mli)
      *mli = sli->length();

    ++mli;
    ++sli;
  }
}


void CTable::get_max_lengths_all(std::list<std::string::size_type>& max_lengths, const TableRow& headers) const
{
  // Erase so we are sure we begin fresh
  max_lengths.erase(max_lengths.begin(), max_lengths.end());

  // Do headers
  get_max_lengths(headers, max_lengths);

  // Do results
  TableRowList::const_iterator ci(m_list->begin());
  TableRowList::const_iterator cend(m_list->end());
  while (ci != cend) {
    get_max_lengths(*ci, max_lengths);
    ++ci;
  }
}


void CTable::clear()
{
  m_list->clear();
}

void CTable::PrintXML(CEMdFOutput *pOut) const
{
  pOut->start_tag("table", true);
  PrintXML_Header(pOut);
  PrintXML_Rows(pOut);
  pOut->end_tag("table", true);
}

void CTable::PrintXML_Header(CEMdFOutput *pOut) const
{
  pOut->start_tag("theader", true);
  TableRow::const_iterator ci_header_names(m_header_names->begin());
  TableRow::const_iterator ci_header_names_end(m_header_names->end());
  TCTList::const_iterator ci_header_types(m_header_types->begin());
  TableRow::const_iterator ci_header_enum_types(m_header_enum_types->begin());
  while (ci_header_names != ci_header_names_end) {
    // Start theadercolumn
    pOut->start_tag("theadercolumn", true);

    // tcaption
    pOut->start_tag("tcaption", false); 
    pOut->out_char_data(*ci_header_names);
    pOut->end_tag("tcaption", false);

    // ttype
    AttributePairList attribute_list;
    attribute_list.push_back(AttributePair(std::string("type"), TCT2String(*ci_header_types)));
    if (*ci_header_types == kTCEnum) {
      attribute_list.push_back(AttributePair(std::string("enum_type"), *ci_header_enum_types));
    }
    pOut->start_single_tag("ttype", attribute_list, false);

    // End theadercolumn
    pOut->end_tag("theadercolumn", false);

    // Advance iterators
    ++ci_header_names;
    ++ci_header_types;
    ++ci_header_enum_types;
  }

  // End theader tag
  pOut->end_tag("theader", true);
}

void CTable::PrintXML_Rows(CEMdFOutput *pOut) const
{
  TableRowList::const_iterator ci(m_list->begin());
  TableRowList::const_iterator cend(m_list->end());
  while (ci != cend) {
    PrintXML_Row(pOut, *ci);
    ++ci;
  }
}

void CTable::PrintXML_Row(CEMdFOutput *pOut, const TableRow& sl) const
{
  // Start trow tag
  pOut->start_tag("trow", true);
  TableRow::const_iterator ci(sl.begin());
  TableRow::const_iterator cend(sl.end());
  while (ci != cend) {
    // Output tcolumn
    pOut->start_tag("tcolumn", true);
    pOut->out_char_data(*ci);
    pOut->end_tag("tcolumn", false);

    // Advance iterator
    ++ci;
  }

  // End trow tag
  pOut->end_tag("trow", true);
}

void CTable::print_DTD(CEMdFOutput *pOut)
{
  pOut->out("  <!ELEMENT table (theader, trow*) >"); 
  pOut->newline();
  pOut->out("  <!ELEMENT theader (theadercolumn)* >"); 
  pOut->newline();
  pOut->out("  <!ELEMENT theadercolumn (tcaption, ttype) >"); 
  pOut->newline();
  pOut->out("  <!ELEMENT tcaption (#PCDATA)* >"); 
  pOut->newline();
  pOut->out("  <!ELEMENT ttype EMPTY >"); 
  pOut->newline();
  pOut->out("  <!ELEMENT trow (tcolumn)* >"); 
  pOut->newline();
  pOut->out("  <!ELEMENT tcolumn (#PCDATA)* >"); 
  pOut->newline();
  pOut->out("  <!ATTLIST table >"); 
  pOut->newline();
  pOut->out("  <!ATTLIST theader >"); 
  pOut->newline();
  pOut->out("  <!ATTLIST theadercolumn >"); 
  pOut->newline();
  pOut->out("  <!ATTLIST tcaption >"); 
  pOut->newline();
  pOut->out("  <!ATTLIST ttype "); 
  pOut->newline();
  pOut->out("     type (bool|string|integer|id_d|enum) #REQUIRED"); 
  pOut->newline();
  pOut->out("     enum_type CDATA #IMPLIED >"); 
  pOut->newline();
  pOut->out("  <!ATTLIST trow >"); 
  pOut->newline();
  pOut->out("  <!ATTLIST tcolumn >"); 
  pOut->newline();
}


