/*
 * mql_enumeration_statements.cpp
 *
 * MQL ENUMERATION statements
 *
 * Ulrik Petersen
 * Created: 3/6-2001 (March 6, 2001)
 * Last update: 3/21-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <mql_enumeration_statements.h>


/////////////////////////////////////////////////////////////
//
// CEnumerationStatement
//
// base class
//
////////////////////////////////////////////////////////////
CEnumerationStatement::~CEnumerationStatement()
{
  delete m_enumeration_name;
}

//
// - Checks that the enumeration exists.  Result in bResult.
//
// - If it exists, gets its enum_id into m_enum_id.
//
bool CEnumerationStatement::SymbolEnumerationExists(bool& bResult)
{
  // Ask DB
  if (!m_pEE->pDB->EnumExists(*m_enumeration_name, bResult, m_enum_id)) {
    // Return with DB error
    return false;
  } 

  if (!bResult) {
    m_pEE->pError->mql_append_error("Enumeration " + *m_enumeration_name + " did not exist.");
  }

  // If we got this far, there were no DB errors
  return true;
}




/////////////////////////////////////////////////////////////
//
// CREATE ENUMERATION
//
// CCreateEnumerationStatement
//
////////////////////////////////////////////////////////////
CCreateEnumerationStatement::CCreateEnumerationStatement(
    CMQL_execution_environment *pEE, 
    std::string* enumeration_name, 
    CECDeclaration* ec_declaration_list) 
  : CEnumerationStatement(pEE, enumeration_name), 
  m_ec_declarations(ec_declaration_list)
{
}

CCreateEnumerationStatement::~CCreateEnumerationStatement()
{
  delete m_ec_declarations;
}

void CCreateEnumerationStatement::Weed(bool& bResult)
{
  m_ec_declarations->WeedDefault(m_pEE, bResult);
  return;
}

bool CCreateEnumerationStatement::Symbol(bool& bResult)
{
  // Check that enumeration does not exist
  bool bEnumerationExists;
  if (!SymbolEnumerationExists(bEnumerationExists)) {
    return false;
  }

  if (bEnumerationExists) {
    bResult = false;
    m_pEE->pError->mql_append_error("Enumeration " + *m_enumeration_name + " exists already!\n");
    return true;
  }

  // Check that no enumeration constant already exists in 
  // the same enumeration by the name given in any of the 
  // ec_declarations
  if (!m_ec_declarations->SymbolExistenceOfConstants(m_pEE, m_enum_id, bResult))
    return false;
  if (!bResult) 
    return true;

  // Assign a value to each ec_declaration, either
  // based on its position in the sequence, or based
  // on its initialization
  m_ec_declarations->SymbolAssignValues();

  // Return with success
  bResult = true;
  return true;
}

bool CCreateEnumerationStatement::Exec()
{
  // Create enumeration and get enum_id
  id_d_t enum_id;
  if (!m_pEE->pDB->CreateEnum(*m_enumeration_name, enum_id)) {
    m_pEE->pError->mql_append_error("Could not create numeration.\n");
    return false;
  }

  if (!m_ec_declarations->CreateConstants(m_pEE, enum_id)) {
    m_pEE->pError->mql_append_error("Could not create constants.\n");
    return false;
  }

  return true;
}



/////////////////////////////////////////////////////////////
//
// UPDATE ENUMERATION
//
// CUpdateEnumerationStatement
//
////////////////////////////////////////////////////////////
CUpdateEnumerationStatement::CUpdateEnumerationStatement(
   CMQL_execution_environment *pEE, 
   std::string* enumeration_name,
   CECUpdate* ec_update_list) : CEnumerationStatement(pEE, enumeration_name),
				m_ec_update_list(ec_update_list)
{
}

CUpdateEnumerationStatement::~CUpdateEnumerationStatement()
{
  delete m_ec_update_list;
}

//
// - Check that at most one declaration has the DEFAULT keyword.
//
void CUpdateEnumerationStatement::Weed(bool& bResult)
{
  m_ec_update_list->WeedDefault(m_pEE, bResult);
  return;
}

//
// - Check that the enumeration exists already
//
// - Check that for all additions, the constants to be added 
//   do not exist
//
// - Check that for all removals and updates, the constants
//   involved do exist.
//
// - Check whether the current default is being removed.
//   If it is, then another default should be specified,
//   either as an update or as an addition.
//
bool CUpdateEnumerationStatement::Symbol(bool& bResult)
{
  // Check that enumeration does exist
  bool bEnumerationExists;
  if (!SymbolEnumerationExists(bEnumerationExists))
    return false;

  if (!bEnumerationExists) {
    bResult = false;
    return true;
  }

  // - Check that for all additions, the constants to be added 
  //   do not exist
  //
  // - Check that for all removals and updates, the constants
  //   involved do exist.
  if (!m_ec_update_list->SymbolConstantExistence(m_pEE, m_enum_id, bResult))
    return false;
  if (!bResult)
    return true;

  //
  // - Check whether the current default is being removed.
  //   If it is, then another default should be specified,
  //   either as an update or as an addition.
  bool bDefaultIsRemoved;
  if (!m_ec_update_list->SymbolDefaultIsRemoved(m_pEE, m_enum_id, bDefaultIsRemoved))
    return false;
  if (bDefaultIsRemoved) {
    if (!m_ec_update_list->has_default()) {
      bResult = false;
      m_pEE->pError->mql_append_error("Default is removed without specifying new default.");
      return true;
    }
  }

  // Set bResult to no compiler error and return no database error
  bResult = true;
  return true;
}

//
// - Add all the constants to be added.
// 
// - Update all the constants to be updated.
//
// - Remove all the constants to be removed.
//
// - If there was a new specification of "default":
//   - Remove the "is_default" flag from the current default.
//   - Set the new default.
// 
bool CUpdateEnumerationStatement::Exec()
{
  // - Add all the constants to be added.
  // - Update all the constants to be updated.
  // - Remove all the constants to be removed.
  if (!m_ec_update_list->ExecAddUpdateRemove(m_pEE, m_enum_id)) 
    return false;

  //
  // - If there was a new specification of "default":
  //   - Remove the "is_default" flag from the current default.
  //   - Set the new default.
  if (m_ec_update_list->has_default()) {
    if (!m_pEE->pDB->SetDefaultEnumConst(m_enum_id,
				  m_ec_update_list->get_default()))
      return false;
  }

  return true;
}



/////////////////////////////////////////////////////////////
//
// DROP ENUMERATION
//
// CDropEnumerationStatement
//
////////////////////////////////////////////////////////////
CDropEnumerationStatement::CDropEnumerationStatement(CMQL_execution_environment *pEE, 
std::string* enumeration_name) : CEnumerationStatement(pEE, enumeration_name)
{
}

CDropEnumerationStatement::~CDropEnumerationStatement()
{
}


//
// - Checks that the enumeration exists.  Result in bResult.
//
// - If it exists, gets its enum_id into m_enum_id.
//
bool CDropEnumerationStatement::Symbol(bool& bResult)
{
  return SymbolEnumerationExists(bResult);
}

bool CDropEnumerationStatement::Exec()
{
  // Drop in database
  return m_pEE->pDB->DropEnum(m_enum_id);
}




