/*
 * mql_helper_classes.cpp
 *
 * MQL helper classes
 *
 * Ulrik Petersen
 * Created: 2/27-2001
 * Last update: 6/21-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <mql_types.h>
#include <mql_error.h>
#include <string_func.h>
#include <sstream>
#include <mql_execution_environment.h>
#include <mql_object_type_statements.h>
#include <mql_monads_statements.h>
#include <mql_enumeration_statements.h>
#include <mql_object_statements.h>
#include <monads.h>


//////////////////////////////////////////////////////////////
//
// CType 
//
//////////////////////////////////////////////////////////////
CType::CType(std::string* Enumeration) // For Enumeration
{
  m_type = kEnumeration;
  m_Enumeration = Enumeration;
  m_string_length = 0;
}

CType::CType(eType type, long string_length)  // for String and ASCII
{
  assert(type == kString || type == kASCII);
  m_type = type;
  m_Enumeration = 0;
  m_string_length = string_length;
}


CType::CType(eType type) // For all others
{
  assert(type != kString && type != kASCII);
  m_type = type;
  m_Enumeration = 0;
  m_string_length = 0;
}

CType::~CType()
{
  delete m_Enumeration;
}

//
// - Checks that length of ASCII and STRING strings 
//   is within the limits.
//
void CType::WeedStringLength(CMQL_execution_environment *pEE, bool& bResult)
{
  switch (m_type) {
  case kInteger:
  case kID_D:
  case kEnumeration:
    bResult = true;
    break;
  case kString:
    bResult = (m_string_length > 0) && (m_string_length <= MAX_STRING_LENGTH*2);
    if (!bResult) {
      pEE->pError->mql_append_error("STRING-declarations must have length between 1 and " + int2string(MAX_STRING_LENGTH) + ".\n");
    }
    break;
  case kASCII:
    bResult = (m_string_length > 0) && (m_string_length <= MAX_STRING_LENGTH);
    if (!bResult) {
      pEE->pError->mql_append_error("ASCII-declarations must have length between 1 and " + int2string(MAX_STRING_LENGTH) + ".\n");
    }
    break;
  };
  return;
}

//
// - Checks that the enumeration exists.  Result in bResult.
//
// - If it exists, gets its enum_id into m_type_id.
//
bool CType::SymbolEnumerationExists(CMQL_execution_environment *pEE, bool& bResult)
{
  assert(m_type == kEnumeration);
  return pEE->pDB->EnumExists(*m_Enumeration, bResult, m_type_id);
}

//
// Assign type id to m_type_id.
//
bool CType::TypeAssignTypeID(CMQL_execution_environment *pEE)
{
  bool bResult; // Dummy: It should already have been checked
  switch (m_type) {
  case kInteger:
    m_type_id = encode_feature_type_id(FEATURE_TYPE_INTEGER);
    break;
  case kString:
    m_type_id = encode_feature_type_id(FEATURE_TYPE_STRING, m_string_length);
    break;
  case kASCII:
    m_type_id = encode_feature_type_id(FEATURE_TYPE_ASCII, m_string_length);
    break;
  case kID_D:
    m_type_id = encode_feature_type_id(FEATURE_TYPE_ID_D);
    break;
  case kEnumeration:
    if (!pEE->pDB->EnumExists(*m_Enumeration, bResult, m_type_id))
      return false;
    break;
  };
  return true;
}



//////////////////////////////////////////////////////////////
//
// CExpression 
//
//////////////////////////////////////////////////////////////
CExpression::CExpression(long integer) // For kExprInteger
{
  m_string = 0;
  m_type = kExprInteger;
  m_integer = integer;
}

CExpression::CExpression(std::string* str, eExprType type) // For kExprString and kExprIdentifier
{
  if (type == kExprString || type == kExprIdentifier) {
    m_string = str;
    m_type = type;
  }
  else
    assert(false);
}

CExpression::~CExpression()
{
  delete m_string;
}

long CExpression::get_integer()
{
  assert(m_type == kExprInteger);
  return m_integer;
}

const std::string& CExpression::get_string()
{
  assert(m_type == kExprString);
  return *m_string;
}

const std::string& CExpression::get_identifier()
{
  assert(m_type == kExprIdentifier);
  return *m_string;
}

bool CExpression::TypeTypeCompatibility(CMQL_execution_environment *pEE, CType* other_type, bool& bResult)
{
  switch(m_type) {
  case kExprInteger:
    bResult = (other_type->get_type() == kInteger
	       || other_type->get_type() == kID_D);
    break;
  case kExprString:
    bResult = other_type->get_type() == kString || other_type->get_type() == kASCII;
    break;
  case kExprIdentifier:
    // Is other an enumeration?
    bResult = other_type->get_type() == kEnumeration;

    // If other isn't enumeration, we won't bother checking existence
    if (!bResult)
      return true;

    // If other is an enumeration, we must check that the enumeration
    // constant exists.
    // This assumes that we have run TypeAssignTypeID on the other_type.
    id_d_t enum_id = other_type->get_type_id();
    long dummy_value;      // Dummy variable
    bool dummy_is_default; // Dummy variable
    if (!pEE->pDB->EnumConstExists(*m_string, enum_id, bResult, 
				   dummy_value, dummy_is_default))
      return false;

    // If it didn't exist, we return now
    if (!bResult)
      return true;

    // Else, the result is true.
    bResult = true;

    // Set m_enum_id at the same time
    m_enum_id = enum_id;
    break;
  };

  return true;
}

bool CExpression::get_as_string(CMQL_execution_environment *pEE, std::string& result, bool bConvertEnumConstToInteger)
{
  std::ostringstream strinteger;
  bool bEnumConstExists;
  bool dummy_is_default;
  long value;
  switch(m_type) {
  case kExprInteger:
    strinteger << m_integer;
    result = strinteger.str();
    break;
  case kExprString:
    result = *m_string;
    break;
  case kExprIdentifier:
    if (bConvertEnumConstToInteger) {
      if (!pEE->pDB->EnumConstExists(*m_string,
				m_enum_id,
				bEnumConstExists,
				value,
				dummy_is_default))
	return false;
      strinteger << value;
      result = strinteger.str();
    } else {
      result = *m_string;
    }
    break;
  };

  // If we got this far, there were no DB errors
  return true;
}






//////////////////////////////////////////////////////////////
//
// CFeatureDeclaration 
//
//////////////////////////////////////////////////////////////
CFeatureDeclaration::CFeatureDeclaration(std::string* feature_name,
					 CType* feature_type,
					 CExpression* default_specification,
					 bool is_computed,
					 CFeatureDeclaration* next)
{
  m_feature_name = feature_name;
  m_type = feature_type;
  m_default_specification = default_specification;
  m_is_computed = is_computed;
  m_next = next;
}

CFeatureDeclaration::~CFeatureDeclaration()
{
  delete m_next;
  delete m_feature_name;
  delete m_type;
  delete m_default_specification;
}

//
// - Check that declarations of STRING and ASCII have lengths
//   within the prescribed limits.
//
void CFeatureDeclaration::WeedStringLength(CMQL_execution_environment *pEE, bool& bResult) 
{ 
  m_type->WeedStringLength(pEE, bResult); 
  return; 
}

//
// - Check that "self" is not declared
//
// - Check that declarations of STRING and ASCII have lengths
//   within the prescribed limits.
//
void CFeatureDeclaration::Weed(CMQL_execution_environment *pEE, bool& bResult)
{
  // First deal with next
  if (m_next != 0) {
    m_next->Weed(pEE, bResult);
    if (!bResult)
      return;
  }

  // Then deal with ourselves

  //
  // - Check that "self" is not declared
  //
  WeedSelfNotDeclared(pEE, bResult);
  if (!bResult)
    return;

  //
  // - Check that declarations of STRING and ASCII have lengths
  //   within the prescribed limits.
  //
  WeedStringLength(pEE, bResult);
  if (!bResult)
    return;

  // If we got this far, there were no compiler errors
  bResult = true;
  return;
}

//
// - Check that "self" is not declared
//
void CFeatureDeclaration::WeedSelfNotDeclared(CMQL_execution_environment *pEE, bool& bResult)
{
  bResult = strcmp_nocase(get_feature_name(), std::string("self")) 
    != 0;
  if (!bResult) {
    pEE->pError->mql_append_error("Feature 'self' is implicitly declared and cannot be declared explicitly.\n");
  }
  return;
}

bool CFeatureDeclaration::SymbolEnumerationsExist(CMQL_execution_environment *pEE, bool& bResult)
{
  // Deal with next first
  if (m_next != 0) {
    if (!m_next->SymbolEnumerationsExist(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }

  // Then deal with outselves
  if (m_type->get_type() == kEnumeration) {
    if (!m_type->SymbolEnumerationExists(pEE, bResult))
      return false;
    if (!bResult) {
      pEE->pError->mql_append_error("Enumeration " + m_type->get_Enumeration() + " used as the type of feature " + *m_feature_name + " does not exist");
      return true;
    }
  } else { // Not enumeration
    bResult = true;
  }
  return true;
}

bool CFeatureDeclaration::SymbolEnumConstantsExist(CMQL_execution_environment *pEE, bool& bResult)
{
  // Deal with next first
  if (m_next != 0) {
    if (!m_next->SymbolEnumConstantsExist(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }

  // Then deal with outselves
  if (m_type->get_type() == kEnumeration) {
    if (m_default_specification != 0 
	&& m_default_specification->get_kind() == kExprIdentifier) {
      id_d_t enum_id = m_type->get_type_id();
      long value;
      bool is_default;
      if (!pEE->pDB->EnumConstExists(m_default_specification->get_identifier(),
				enum_id, bResult, value, is_default))
	return false;
      else
	bResult = true;
    } else { // Not identifier, or empty: Will be caught in type-checker
      bResult = true;
    }
  } else { // Not enumeration
    bResult = true;
  }
  return true;
}

//
// - Check that none of the features exist for the object type 
//   object_type_id.
//
bool CFeatureDeclaration::SymbolFeaturesMustNotExist(CMQL_execution_environment *pEE, bool& bResult, id_d_t object_type_id)
{
  // Deal with next first
  if (m_next != 0) {
    if (!m_next->SymbolFeaturesMustNotExist(pEE, bResult, object_type_id))
      return false;
    if (!bResult)
      return true;
  }

  bool bFeatureExists;
  id_d_t feature_type_id;      // Dummy variable
  std::string default_value;   // Dummy variable
  bool is_computed;            // Dummy variable
  if (!pEE->pDB->FeatureExists(*m_feature_name,
			  object_type_id,
			  bFeatureExists,
			  feature_type_id,
			  default_value,
			  is_computed))
    return false;

  bResult = !bFeatureExists;
  return true;
}

//
// - Assign type-id to each feature.
//
bool CFeatureDeclaration::TypeAssignTypeID(CMQL_execution_environment *pEE)
{
  // Deal with next first
  if (m_next != 0) {
    if (!m_next->TypeAssignTypeID(pEE))
      return false;
  }

  // Then deal with outselves
  if (!m_type->TypeAssignTypeID(pEE))
    return false;

  return true;
}

//
// - Check that there is type-compatibility between the feature
//   and the default specification, and make a default specification
//   if it does not exist.
//
bool CFeatureDeclaration::TypeTypeCompatibility(CMQL_execution_environment *pEE, bool& bResult)
{
  // Deal with next first
  if (m_next != 0) {
    if (!m_next->TypeTypeCompatibility(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }

  // Then deal with ourselves
  if (m_default_specification != 0) {
    if (!m_default_specification->TypeTypeCompatibility(pEE, m_type, bResult))
      return false;
  } else { // m_default_specification == 0
    std::string *pDefault_string;
    switch (m_type->get_type()) {
    case kInteger:
      m_default_specification = new CExpression(0);
      break;
    case kString:
      pDefault_string = new std::string("");
      m_default_specification = new CExpression(pDefault_string, kExprString);
      break;
    case kASCII:
      pDefault_string = new std::string("");
      m_default_specification = new CExpression(pDefault_string, kExprString);
      break;
    case kID_D:
      m_default_specification = new CExpression(NIL);
      break;
    case kEnumeration:
      pDefault_string = new std::string;
      if (!pEE->pDB->GetDefault(m_type->get_type_id(),
			   bResult,
			   *pDefault_string))
	return false;
      else
	m_default_specification = new CExpression(pDefault_string, kExprIdentifier);
      if (!bResult) {
	pEE->pError->mql_append_error(std::string("Could not find default for enumeration ") + *m_feature_name + ".\n");
      }
      break;
    }
  }

  return true;
}

bool CFeatureDeclaration::get_feature_info(CMQL_execution_environment *pEE, feature_info& fi)
{
  std::string feature_value;
  if (!get_default_specification()->get_as_string(pEE, feature_value, false))
    return false;
  fi = feature_info(get_feature_name(), 
		    get_type()->get_type_id(),
		    feature_value,
		    get_is_computed());
  return true;
}


bool CFeatureDeclaration::ExecAdd(CMQL_execution_environment *pEE, const std::string& object_type_name, id_d_t object_type_id)
{
  // Deal with next first
  if (m_next != 0) {
    if (!m_next->ExecAdd(pEE, object_type_name, object_type_id))
      return false;
  }

  // Then deal with ourselves
  feature_info fi;
  if (!get_feature_info(pEE, fi))
    return false;
  if (!pEE->pDB->AddFeature(object_type_name, object_type_id, fi))
    return false;


  // If we got this far, there were no DB errors
  return true;
}


//////////////////////////////////////////////////////////////
//
// CObjectTypeName 
//
//////////////////////////////////////////////////////////////

CObjectTypeName::CObjectTypeName(std::string* object_type_name, 
				 CObjectTypeName* next)
{
  m_object_type_name = object_type_name;
  m_next = next;
}

CObjectTypeName::~CObjectTypeName()
{
  delete m_next;
  delete m_object_type_name;
}

//
// - Checks that all object type names exist.
//
// - Sets m_object_type_id to the object type_id of
//   the object type (if it exists).
//
bool CObjectTypeName::SymbolTypesExist(CMQL_execution_environment *pEE, bool& bResult)
{
  // First deal with next
  if (m_next != 0) {
    if (!m_next->SymbolTypesExist(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }

  // Then deal with ourselves
  if (!pEE->pDB->ObjectTypeExists(*m_object_type_name,
			     bResult,
			     m_object_type_id))
    return false;

  // If we got this far, there were no compiler errors 
  // and no database errors
  bResult = true;
  return true;
}

bool CObjectTypeName::Exec(CMQL_execution_environment *pEE, id_d_t object_type_id, 
			   eAddUpdateRemove AddRemove, eCBBF cbbf)
{
  // First deal with next
  if (m_next != 0) {
    if (!m_next->Exec(pEE, object_type_id, AddRemove, cbbf))
      return false;
  }

  // Then deal with ourselves
  switch (AddRemove) {
  case kAdd:
    switch (cbbf) {
    case kCovered_by:
      if (!pEE->pDB->CreateCoveredByRelationship(object_type_id, m_object_type_id))
	return false;
      break;
    case kBuildable_from:
      if (!pEE->pDB->CreateBuildableFromRelationship(object_type_id, m_object_type_id))
	return false;
      break;
    default:
      assert(false);
      break;
    }
    break;
  case kRemove:
    switch (cbbf) {
    case kCovered_by:
      if (!pEE->pDB->DropCoveredByRelationship(object_type_id, m_object_type_id))
	return false;
      break;
    case kBuildable_from:
      if (!pEE->pDB->DropBuildableFromRelationship(object_type_id, m_object_type_id))
	return false;
      break;
    default:
      assert(false);
      break;
    }
    break;
  default:
    assert(false);
    break;
  }

  // If we got this far, there were no DB errors
  return true;
}

//////////////////////////////////////////////////////////////
//
// CCBBFList 
//
//////////////////////////////////////////////////////////////

CCBBFList::CCBBFList(CObjectTypeName* object_type_name_list, 
		     eAddUpdateRemove AddRemove, 
		     eCBBF cbbf, 
		     CCBBFList* next) // only Add or Remove
{
  assert(AddRemove == kAdd || AddRemove == kRemove);
  m_object_type_name_list = object_type_name_list;
  m_AddRemove = AddRemove;
  m_cbbf = cbbf;
  m_next = next;
}

CCBBFList::~CCBBFList()
{
  delete m_next;
  delete m_object_type_name_list;
}

bool CCBBFList::SymbolTypesExist(CMQL_execution_environment *pEE, bool& bResult)
{
  // First deal with next
  if (m_next != 0) {
    if (!m_next->SymbolTypesExist(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }

  // Then deal with ourselves
  if (!m_object_type_name_list->SymbolTypesExist(pEE, bResult))
    return false;
  else
    return true;
}

bool CCBBFList::Exec(CMQL_execution_environment *pEE, const std::string& object_type_name, id_d_t object_type_id)
{
  // First deal with next
  if (m_next != 0) {
    if (!m_next->Exec(pEE, object_type_name, object_type_id))
      return false;
  }

  // Then deal with ourselves
  if (!m_object_type_name_list->Exec(pEE, object_type_id, m_AddRemove, m_cbbf))
    return false;


  // If we got this far, there were no DB errors
  return true;
}




//////////////////////////////////////////////////////////////
//
// CFeatureUpdate 
//
//////////////////////////////////////////////////////////////

CFeatureUpdate::CFeatureUpdate(CFeatureDeclaration* feature_declaration, CFeatureUpdate* next)
{
  m_AddOrRemove = kAdd;
  m_feature_addition = feature_declaration;
  m_feature_removal = 0;
  m_next = next;
}

CFeatureUpdate::CFeatureUpdate(std::string* feature_name, CFeatureUpdate* next)
{
  m_AddOrRemove = kRemove;
  m_feature_addition = 0;
  m_feature_removal = feature_name;
  m_next = next;
}

CFeatureUpdate::~CFeatureUpdate()
{
  delete m_next;
  delete m_feature_addition;
  delete m_feature_removal;
}

void CFeatureUpdate::Weed(CMQL_execution_environment *pEE, bool& bResult)
{
  // First do next
  if (m_next != 0) {
    m_next->Weed(pEE, bResult);
    if (!bResult)
      return;
  }

  // Then do ourselves
  switch (m_AddOrRemove) {
  case kAdd:
    // The CFeatureDeclaration knows how to do this
    m_feature_addition->Weed(pEE, bResult);
    return;
    break;
  case kRemove:
    bResult = strcmp_nocase(get_feature_removal(), std::string("self")) 
      != 0;
    if (!bResult) {
      pEE->pError->mql_append_error("Cannot remove feature 'self'.\n");
    }
    return;
  default:
    assert(false);
    break;
  };
  if (!bResult)
    return;
  
  // If we got this far, there were no errors
  bResult = true;
  return;
}

// 
// - Check that all the features that are to be removed do exist.
//   
// - Check that all the features that are to be added do not exist.
//
bool CFeatureUpdate::SymbolExistenceOfFeatures(CMQL_execution_environment *pEE, 
					       bool& bResult, 
					       id_d_t object_type_id)
{
  // First do next
  if (m_next != 0) {
    if (!m_next->SymbolExistenceOfFeatures(pEE, bResult, object_type_id))
      return false;
    if (!bResult)
      return true;
  }

  // Then do ourselves
  id_d_t feature_type_id;      // Dummy variable
  std::string default_value;   // Dummy variable
  bool is_computed;            // Dummy variable
  switch (m_AddOrRemove) {
  case kAdd:
    // Features for addition must not exist.
    return m_feature_addition->SymbolFeaturesMustNotExist(pEE, bResult, object_type_id);
    break;
  case kRemove:
    // Features for removal must exist.
    if (!pEE->pDB->FeatureExists(*m_feature_removal,
			    object_type_id,
			    bResult,
			    feature_type_id,
			    default_value,
			    is_computed))
      return false;
    return true;
  default:
    assert(false);
    return false; // Just so we don't get a warning
    break;
  };
}

//
// - Check that the enumerations exist for the types   
//   whose features are enumerations
// - Check that, within these enumerations, any default
//   specification which is an enumeration constant exists
//   in that enumeration.
bool CFeatureUpdate::SymbolEnumerations(CMQL_execution_environment *pEE, bool& bResult)
{
  // First do next
  if (m_next != 0) {
    if (!m_next->SymbolEnumerations(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }

  // Then do ourselves
  switch (m_AddOrRemove) {
  case kAdd:
    //
    // - Check that the enumerations exist for the types   
    //   whose features are enumerations
    if (!m_feature_addition->SymbolEnumerationsExist(pEE, bResult))
      return false;
    if (!bResult) 
      return true;


    // - Check that, within these enumerations, any default
    //   specification which is an enumeration constant exists
    //   in that enumeration.
    if (!m_feature_addition->SymbolEnumConstantsExist(pEE, bResult))
      return false;
    if (!bResult)
      return true;
    break;
  case kRemove:
    bResult = true;
    return true;
  default:
    assert(false);
    return false; // Just so we don't get a warning
    break;
  };

  return true;
}

bool CFeatureUpdate::TypeAssignTypeID(CMQL_execution_environment *pEE)
{
  // Deal with next first
  if (m_next != 0) {
    if (!m_next->TypeAssignTypeID(pEE))
      return false;
  }

  // Then deal with outselves
  if (m_feature_addition != 0) {
    if (!m_feature_addition->TypeAssignTypeID(pEE))
      return false;
  }

  return true;
}

bool CFeatureUpdate::TypeTypeCompatibility(CMQL_execution_environment *pEE, bool& bResult)
{
  // Deal with next first
  if (m_next != 0) {
    if (!m_next->TypeTypeCompatibility(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }

  // Then deal with ourselves
  if (m_feature_addition != 0) {
    if (!m_feature_addition->TypeTypeCompatibility(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }

  return true;
}

bool CFeatureUpdate::Exec(CMQL_execution_environment *pEE, const std::string& object_type_name, id_d_t object_type_id)
{
  // First do next
  if (m_next != 0) {
    if (!m_next->Exec(pEE, object_type_name, object_type_id))
      return false;
  }

  // Then do ourselves
  switch (m_AddOrRemove) {
  case kAdd:
    if (!m_feature_addition->ExecAdd(pEE, object_type_name, object_type_id))
      return false;
    break;
  case kRemove:
    if (!pEE->pDB->DropFeature(object_type_name, *m_feature_removal, object_type_id))
      return false;
    break;
  default:
    assert(false);
    return false; // Just so we don't get a warning
    break;
  };
  

  // If we got this far, there were no DB errors
  return true;
}


//////////////////////////////////////////////////////////////
//
// CMonadSetElement 
//
//////////////////////////////////////////////////////////////

CMonadSetElement::CMonadSetElement(monad_m mse_first, 
				   monad_m mse_last, 
				   CMonadSetElement* next)
{
  m_mse_first = mse_first;
  m_mse_last = mse_last;
  m_next = next;
}

CMonadSetElement::CMonadSetElement(monad_m mse_singleton, 
				   CMonadSetElement* next)
{
  m_mse_first = m_mse_last = mse_singleton;
  m_next = next;
}

CMonadSetElement::~CMonadSetElement()
{
  delete m_next;
}

//
// - Checks that each monad set element is monotonic.
//   I.e., that mse_last >= mse_first.
//
void CMonadSetElement::WeedMonotonic(CMQL_execution_environment *pEE, bool& bResult)
{
  // First deal with next
  if (m_next != 0) {
    m_next->WeedMonotonic(pEE, bResult);
    if (!bResult)
      return;
  }

  // Then deal with ourselves
  bResult = m_mse_last >= m_mse_first;
  if (!bResult) {
    pEE->pError->mql_append_error("Range in monad set must be monotonic (i.e., in 'A - B',\nB must be greater than or equal to A).\n");
    pEE->pError->mql_append_error("The offending mse was " + monad_m2string(m_mse_first) + "-" + monad_m2string(m_mse_last) +".\n");
  }
  return;
}

//
// - Checks that both mse_first and mse_last are positive.
// 
void CMonadSetElement::WeedPositive(CMQL_execution_environment *pEE, bool& bResult)
{
  // First deal with next
  if (m_next != 0) {
    m_next->WeedPositive(pEE, bResult);
    if (!bResult)
      return;
  }

  // Then deal with ourselves
  bResult = (m_mse_first >= 1) && (m_mse_last >= 1);
  if (!bResult) {
    pEE->pError->mql_append_error("All monads in a monad set must be positive (i.e., greater than or equal to 1).\n");
    pEE->pError->mql_append_error("The offending mse was " + monad_m2string(m_mse_first) + "-" + monad_m2string(m_mse_last) +".\n");
  }
  return;
}

void CMonadSetElement::MonadsBuildSet(CSet_of_monad_ms& monads)
{
  // First deal with next
  if (m_next != 0) {
    m_next->MonadsBuildSet(monads);
  }

  // Then deal with ourselves
  monads.add(m_mse_first, m_mse_last);
}


//////////////////////////////////////////////////////////////
//
// CMonadSetChainElement
//
//////////////////////////////////////////////////////////////

CMonadSetChainElement::CMonadSetChainElement(CMonadSetElement *myset, eSetOperator op, CMonadSetChainElement *next)
{
  m_set = myset;
  m_operator = op;
  m_next = next;
}

CMonadSetChainElement::~CMonadSetChainElement()
{
  delete m_set;
  delete m_next;
}

//
// - Checks that each monad set element is monotonic.
//   I.e., that mse_last >= mse_first.
//
void CMonadSetChainElement::WeedMonotonic(CMQL_execution_environment *pEE, bool& bResult)
{
  // First deal with next
  if (m_next != 0) {
    m_next->WeedMonotonic(pEE, bResult);
    if (!bResult)
      return;
  }

  // Then deal with ourselves
  assert(m_set);
  m_set->WeedMonotonic(pEE, bResult);

  // Return
  return;
}

//
// - Checks that both mse_first and mse_last are positive.
// 
void CMonadSetChainElement::WeedPositive(CMQL_execution_environment *pEE, bool& bResult)
{
  // First deal with next
  if (m_next != 0) {
    m_next->WeedPositive(pEE, bResult);
    if (!bResult)
      return;
  }

  // Then deal with ourselves
  assert(m_set);
  m_set->WeedPositive(pEE, bResult);
  return;
}

void CMonadSetChainElement::Exec(CSet_of_monad_ms& monads)
{
  // First deal with next
  if (m_next != 0) {
    m_next->Exec(monads);
  } else { // m_next == 0
    // This is the first in the chain.
    // Just build monad set, and return.
    m_set->MonadsBuildSet(monads);
    return;
  }

  //
  // Then deal with ourselves
  //

  // Get our own monads
  CSet_of_monad_ms self_monads;
  m_set->MonadsBuildSet(self_monads);
  
  // Apply operation
  switch (m_operator) {
  case kSOUnion:
    monads.union_with(self_monads);
    break;
  case kSOIntersect:
    monads = CSet_of_monad_ms::intersect(monads, self_monads);
    break;
  case kSODifference:
    monads.difference(self_monads);
    break;
  case kSONone:
    assert(false);
    break;
  default:
    assert(false);
    break;
  }
  return;
}




//////////////////////////////////////////////////////////////
//
// CECDeclaration 
//
//////////////////////////////////////////////////////////////

CECDeclaration::CECDeclaration(bool is_default, 
			       std::string* ec_name, 
			       long* opt_ec_initialization, 
			       CECDeclaration* next)
{
  m_is_default = is_default;
  m_ec_name = ec_name;
  m_opt_ec_initialization = opt_ec_initialization;
  m_next = next;
}

CECDeclaration::~CECDeclaration()
{
  delete m_next;
  delete m_ec_name;
  delete m_opt_ec_initialization;
}

bool CECDeclaration::get_is_default()
{
  return m_is_default;
}

long CECDeclaration::get_opt_ec_initialization()
{
  return *m_opt_ec_initialization;
}

const std::string& CECDeclaration::get_ec_name()
{
  return *m_ec_name;
}

void CECDeclaration::WeedDefault(CMQL_execution_environment *pEE, bool& bResult)
{
  bool has_met_default = false;
  WeedDefaultPrivate(pEE, bResult, has_met_default);
  return;
}

//
// - Check that no constant exists already in the same enumeration
// 
bool CECDeclaration::SymbolExistenceOfConstants(CMQL_execution_environment *pEE,id_d_t enum_id, bool& bResult)
{
  // Deal with next first
  if (m_next != 0) {
    if (!m_next->SymbolExistenceOfConstants(pEE, enum_id, bResult))
      return false;
    if (!bResult)
      return true;
  }


  // Then deal with ourselves
  bool bEnumConstExists;
  long dummy_value;
  bool dummy_is_default;
  if (!pEE->pDB->EnumConstExists(*m_ec_name,
			    enum_id,
			    bEnumConstExists,
			    dummy_value,
			    dummy_is_default))
    return false;

  // If it existed, we return with a fail
  bResult = !bEnumConstExists;
  if (!bResult) {
    std::string mql_error;
    mql_error = "The constant " + *m_ec_name + " exists already.\n";
    pEE->pError->mql_append_error(mql_error);
  }

  // Return no database error
  return true;
}

//
// - Assign a value to each ec_declaration, either based on its 
//   position in the sequence, or based on its optional initialization
// 
void CECDeclaration::SymbolAssignValues(void)
{
  long previous = 0;
  SymbolAssignValuesPrivate(previous);
}

void CECDeclaration::SymbolAssignValuesPrivate(long& previous)
{
  // Deal with next before we deal with ourselves
  if (m_next != 0) {
    m_next->SymbolAssignValuesPrivate(previous);
  } 

  // Then deal with ourselves:
  // - If we have an explicit initialization, use that.
  // - Else use previous + 1
  if (m_opt_ec_initialization != 0) {
    m_value = *m_opt_ec_initialization;
  } else {
    m_value = previous + 1;
  }

  // Set previous so that the next one in the list can use it
  previous = m_value;

  // Return
  return;
}


//
// - Check that at most one declaration has the DEFAULT keyword.
//
// - If none has the DEFAULT keyword, set the m_is_default
//   of the first one to true.
//
void CECDeclaration::WeedDefaultPrivate(CMQL_execution_environment *pEE, bool& bResult, bool& has_met_default)
{
  // If we have already met the default, and this is a default too, 
  // return with a fail
  if (has_met_default && m_is_default) {
    bResult = false;
    std::string mql_error;
    mql_error = "You cannot have more than one default\n";
    pEE->pError->mql_append_error(mql_error);
    return;
  }

  // Set has_met_default
  if (m_is_default)
    has_met_default = true;

  // Deal with next before we deal with ourselves
  if (m_next != 0) {
    m_next->WeedDefaultPrivate(pEE, bResult, has_met_default);
    if (!bResult)
      return;
  }

  // Then deal with ourselves:
  // If this is the first one, and we haven't met the default yet,
  // set ourselves to be the default
  if (m_next == 0 && !has_met_default)
    m_is_default = true;

  // Set bResult and return
  bResult = true;
  return;
}

bool CECDeclaration::CreateConstants(CMQL_execution_environment *pEE, id_d_t enum_id)
{
  // Deal with next before we deal with ourselves
  if (m_next != 0) {
    if (!m_next->CreateConstants(pEE, enum_id))
      return false;
  }

  // Then deal with ourselves:
  enum_const_info ec_info(*m_ec_name, m_value, m_is_default);
  if (!pEE->pDB->CreateEnumConstant(enum_id, ec_info))
    return false;

  // Return with no database error
  return true;
}



//////////////////////////////////////////////////////////////
//
// CECUpdate 
//
//////////////////////////////////////////////////////////////

CECUpdate::CECUpdate(bool is_default, 
		     std::string* ec_name, 
		     long ec_initialization, 
		     eAddUpdateRemove kind, 
		     CECUpdate* next)
{
  m_is_default = is_default;
  m_ec_name = ec_name;
  m_ec_initialization = ec_initialization;
  m_kind = kind;
  m_next = next;
}

CECUpdate::~CECUpdate()
{
  delete m_next;
  delete m_ec_name;
}



void CECUpdate::WeedDefault(CMQL_execution_environment *pEE, bool& bResult)
{
  bool has_met_default = false;
  WeedDefaultPrivate(pEE, bResult, has_met_default);
  return;
}

//
// - Check that at most one declaration has the DEFAULT keyword.
//
void CECUpdate::WeedDefaultPrivate(CMQL_execution_environment *pEE, bool& bResult, bool& has_met_default)
{
  // If we have already met the default, and this is a default too, 
  // return with a fail
  if (has_met_default && m_is_default) {
    bResult = false;
    std::string mql_error;
    mql_error = "You cannot have more than one default\n";
    pEE->pError->mql_append_error(mql_error);
    return;
  }

  // Set has_met_default
  if (m_is_default)
    has_met_default = true;

  // Deal with next 
  if (m_next != 0) {
    m_next->WeedDefaultPrivate(pEE, bResult, has_met_default);
    if (!bResult)
      return;
  }


  // Set bResult and return
  bResult = true;
  return;
}

bool SymbolConstExists(CMQL_execution_environment *pEE, const std::string& ec_name, id_d_t enum_id, bool& bConstExists)
{
  long dummy_value;
  bool dummy_is_default;
  return pEE->pDB->EnumConstExists(ec_name,
			      enum_id,
			      bConstExists,
			      dummy_value,
			      dummy_is_default);
}

//
// - Check that for all additions, the constants to be added 
//   do not exist
//
// - Check that for all removals and updates, the constants
//   involved do exist.
//
bool CECUpdate::SymbolConstantExistence(CMQL_execution_environment *pEE, id_d_t enum_id, bool& bResult)
{
  // Deal with next first
  if (m_next != 0) {
    if (!m_next->SymbolConstantExistence(pEE, enum_id, bResult))
      return false;
    if (!bResult)
      return true;
  }

  // Then deal with ourselves
  bool bConstExists;
  switch (m_kind) {
  case kAdd:
    if (!SymbolConstExists(pEE, *m_ec_name, enum_id, bConstExists))
      return false;
    if (bConstExists) {
      bResult = false;
      std::string mql_error;
      mql_error = "Enumeration constant " + *m_ec_name + " exists already.  Cannot add.\n";
      pEE->pError->mql_append_error(mql_error);
      return true;
    }
    break;
  case kUpdate:
    if (!SymbolConstExists(pEE, *m_ec_name, enum_id, bConstExists))
      return false;
    if (!bConstExists) {
      bResult = false;
      std::string mql_error;
      mql_error = "Enumeration constant " + *m_ec_name + " does not exist.  Cannot update.\n";
      pEE->pError->mql_append_error(mql_error);
      return true;
    }
    break;
  case kRemove:
    if (!SymbolConstExists(pEE, *m_ec_name, enum_id, bConstExists))
      return false;
    if (!bConstExists) {
      bResult = false;
      std::string mql_error;
      mql_error = "Enumeration constant " + *m_ec_name + " does not exist.  Cannot remove.\n";
      pEE->pError->mql_append_error(mql_error);
      return true;
    }
    break;
  }
  bResult = true;
  return true;
}

//
// - Check whether the current default is being removed.
//
bool CECUpdate::SymbolDefaultIsRemoved(CMQL_execution_environment *pEE, id_d_t enum_id, bool& bDefaultIsRemoved)
{
  // Deal with next first
  if (m_next != 0) {
    if (!m_next->SymbolDefaultIsRemoved(pEE, enum_id, bDefaultIsRemoved))
      return false;
    if (bDefaultIsRemoved)
      return true;
  }

  // Then deal with ourselves
  if (m_kind == kRemove) {
    bool dummy_const_exists;
    long dummy_value;
    bool is_default;
    if (!pEE->pDB->EnumConstExists(*m_ec_name,
			      enum_id,
			      dummy_const_exists,
			      dummy_value,
			      is_default))
      return false;
    if (is_default) {
      bDefaultIsRemoved = true;
      return true;
    }
  }
  
  // Set result and return
  bDefaultIsRemoved = false;
  return true;
}

//
// Returns true on one of the ec_updates in the list
// has a "default" modifier.
//
bool CECUpdate::has_default(void)
{
  // First deal with next
  if (m_next != 0) {
    bool bResult;
    bResult = m_next->has_default();
    if (bResult)
      return true;
  } 

  // Then deal with ourselves
  return m_is_default;
}

// Returns a reference to the ec_name that is the default.
// Assumes that one is the default.
const std::string& CECUpdate::get_default(void) const
{
  // If we are the default, return the ec_name
  if (m_is_default)
    return *m_ec_name;

  if (m_next != 0) {
    return m_next->get_default();
  } else {
    // We reached the end of the list without finding the default.
    assert(false);
    return *m_ec_name;  // Just so we don't get a warning
  }
}

//
// - Add all the constants to be added.
//
// - Update all the constants to be updated.
//
// - Remove all the constants to be removed.
//
bool CECUpdate::ExecAddUpdateRemove(CMQL_execution_environment *pEE, id_d_t enum_id)
{
  // First deal with next
  if (m_next != 0) {
    if (!m_next->ExecAddUpdateRemove(pEE, enum_id))
      return false;
  }

  // Then deal with ourselves
  enum_const_info ec_info(*m_ec_name, m_ec_initialization, m_is_default);
  long value;      // Current value of enum const
  bool dummy_is_default; // Dummy variable
  bool bExists;

  // Get current value (if constant exists)
  if (!pEE->pDB->EnumConstExists(*m_ec_name, enum_id, bExists, 
				 value, dummy_is_default))
    return false;

  switch (m_kind) {
  case kAdd:
    if (!pEE->pDB->CreateEnumConstant(enum_id, ec_info))
      return false;
    break;
  case kUpdate:
    // If it doesn't exist, something may be gruesomely wrong,
    // since have checked for existence previously.
    // However, since Emdros is a multi-user system,
    // it could be that someone has removed it before us
    // after we have checked for existence, so we should handle 
    // this gracefully.
    if (bExists) {
      if (!pEE->pDB->UpdateEnumConst(*m_ec_name, enum_id, value, m_ec_initialization))
	return false;
    }
    break;
  case kRemove:
    // If it doesn't exist, something may be gruesomely wrong,
    // since have checked for existence previously.
    // However, since Emdros is a multi-user system,
    // it could be that someone has removed it before us
    // after we have checked for existence, so we should handle 
    // this gracefully.
    if (bExists) {
      // Drop it!
      if (!pEE->pDB->DropEnumConst(enum_id, *m_ec_name, value))
	return false;
    }
    break;
  }

  // Return no database error
  return true;
}




//////////////////////////////////////////////////////////////
//
// CFeatureAssignment 
//
//////////////////////////////////////////////////////////////

CFeatureAssignment::CFeatureAssignment(std::string* feature_name,
				       CExpression* expression,
				       CFeatureAssignment* next)
{
  m_feature_name = feature_name;
  m_expression = expression;
  m_next = next;
  m_type = 0;
}

CFeatureAssignment::~CFeatureAssignment()
{
  delete m_next;
  delete m_feature_name;
  delete m_expression;
  delete m_type;
}

//
// - Check that the feature "self" is not assigned to
//
void CFeatureAssignment::WeedSelfNotAssigned(CMQL_execution_environment *pEE, bool& bResult)
{
  // First deal with next
  if (m_next != 0) {
    m_next->WeedSelfNotAssigned(pEE, bResult);
    if (!bResult)
      return;
  }

  // Then deal with ourselves
  bResult = strcmp_nocase("self", *m_feature_name) != 0;
  if (!bResult) {
    pEE->pError->mql_append_error("Cannot assign to feature 'self', since this feature is read-only.\n");
  }
}

//
// Returns true if a member of the list has a feature of name
// feature_name.  The comparison is done caseinsensitively.
// 
bool CFeatureAssignment::list_has_feature(const std::string& feature_name)
{
  // First deal with next
  if (m_next != 0) {
    if (m_next->list_has_feature(feature_name))
      return true;
  }

  // Then deal with ourselves
  return strcmp_nocase(*m_feature_name, feature_name) == 0;
}

bool CFeatureAssignment::SymbolNoFeatureMayBeComputed(CMQL_execution_environment *pEE, bool& bResult)
{
  // Deal with next first
  if (m_next != 0) {
    if (!m_next->SymbolNoFeatureMayBeComputed(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }

  // Then deal with ourselves
  bResult = !m_is_computed;
  if (!bResult) {
    pEE->pError->mql_append_error("Feature " + *m_feature_name + " is a computed feature.  Cannot assign to a computed feature.\n");
  } 
  return true;
}


//
// Make a std::list<feature_info> from the list
//
bool CFeatureAssignment::ExecMakeFeatureList(CMQL_execution_environment *pEE, std::list<feature_info>& features)
{
  // First deal with next
  if (m_next != 0) {
    m_next->ExecMakeFeatureList(pEE, features);
  }

  // Then deal with ourselves
  std::string feature_value;
  if (!m_expression->get_as_string(pEE, feature_value, true))
    return false;
  feature_info f_info(*m_feature_name,               // feature name
		      m_type->get_type_id(),         // feature type
		      feature_value,                 // feature value
		      false);                        // is_computed
  features.push_back(f_info);
  return true;
}

//
// - Assign a type based on m_feature_name
//
bool CFeatureAssignment::TypeAssignTypes(CMQL_execution_environment *pEE, id_d_t object_type_id)
{
  // First deal with next
  if (m_next != 0) {
    if (!m_next->TypeAssignTypes(pEE, object_type_id))
      return false;
  }

  // Then deal with ourselves
  bool bDummyFeatureExists;
  id_d_t feature_type_id;
  std::string dummy_default_value;
  bool dummy_is_computed;
  if (!pEE->pDB->FeatureExists(*m_feature_name,
			  object_type_id,
			  bDummyFeatureExists, // We know it exists (symbolchecking)
			  feature_type_id,
			  dummy_default_value,
			  dummy_is_computed))
    return false;

  bool bEnumExists;
  std::string enum_name;
  std::string *pEnum_name;

  // Decode feature type
  id_d_t ft;
  long string_length;
  decode_feature_type_id(feature_type_id, ft, string_length);

  // Decide type
  switch (ft) {
  case FEATURE_TYPE_INTEGER:
    m_type = new CType(kInteger);
    break;
  case FEATURE_TYPE_STRING:
    m_type = new CType(kString, string_length);
    break;
  case FEATURE_TYPE_ASCII:
    m_type = new CType(kASCII, string_length);
    break;
  case FEATURE_TYPE_ID_D:
    m_type = new CType(kID_D);
    break;
  default:
    if (!pEE->pDB->EnumExists(feature_type_id, bEnumExists, enum_name))
      return false;
    pEnum_name = new std::string(enum_name);
    m_type = new CType(pEnum_name);
    break;
  }

  // Assign type_id internally to the CType
  if (!m_type->TypeAssignTypeID(pEE))
    return false;

  return true;
}

//
// - Check for type compatibility between feature and expression
//
bool CFeatureAssignment::TypeTypeCompatibility(CMQL_execution_environment *pEE, bool& bResult)
{
  // First deal with next
  if (m_next != 0) {
    if (!m_next->TypeTypeCompatibility(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }

  // Then deal with ourselves
  if (!m_expression->TypeTypeCompatibility(pEE, m_type, bResult))
    return false;
  if (!bResult) {
    pEE->pError->mql_append_error("There was type incompatibility between the feature " + *m_feature_name + " and its assignment.\n");
    return true;
  }

  // If we got this far, there were no DB errors and no compiler errors
  bResult = true;
  return true;
}



//////////////////////////////////////////////////////////////
//
// CObjectSpec 
//
//////////////////////////////////////////////////////////////

CObjectSpec::CObjectSpec(std::string* object_type_name,
			 CFeatureAssignment* feature_assignment_list)
{
  m_object_type_name = object_type_name;
  m_feature_assignments = feature_assignment_list;
}

CObjectSpec::~CObjectSpec()
{
  delete m_object_type_name;
  delete m_feature_assignments;
}

//
// - Check that the feature "self" is not assigned to
//
void CObjectSpec::WeedSelfNotAssigned(CMQL_execution_environment *pEE, bool& bResult)
{
  if (m_feature_assignments != 0)
    m_feature_assignments->WeedSelfNotAssigned(pEE, bResult);
  else
    bResult = true;
  return;
}

// Local helper function
void LocalWeedObjectTypeNameNotSpecial(CMQL_execution_environment *pEE, std::string special_type, const std::string& object_type_name, bool& bResult)
{
  bResult = strcmp_nocase(special_type, object_type_name) != 0;
  if (!bResult) {
    pEE->pError->mql_append_error("Cannot create object of type " + special_type + ", since this object type is a special object type, implicitly defined.\n");
  }
}

//
// - Check that "object_type_name" is not all_m, pow_m, or any_m
// 
void CObjectSpec::WeedObjectTypeNameNotSpecial(CMQL_execution_environment *pEE, bool& bResult)
{
  LocalWeedObjectTypeNameNotSpecial(pEE, "all_m", *m_object_type_name, bResult);
  if (!bResult) {
    return;
  }

  LocalWeedObjectTypeNameNotSpecial(pEE, "pow_m", *m_object_type_name, bResult);
  if (!bResult) {
    return;
  }

  LocalWeedObjectTypeNameNotSpecial(pEE, "any_m", *m_object_type_name, bResult);
  if (!bResult) {
    return;
  }
}

//
// - Check that the object type exists.  In doing so, store its
//   type_id in the AST.
// 
bool CObjectSpec::SymbolObjectTypeExists(CMQL_execution_environment *pEE, bool& bTypeExists)
{
  if (!pEE->pDB->ObjectTypeExists(*m_object_type_name, bTypeExists, m_object_type_id))
    return false;
  return true;
}

//
// - Make sure that all features are given a value.  If a feature 
//   is not given an explicit value, then use the default value.
//   If a feature exists for the object type, but is not in the list
//   of feature assignments, then add it to the list of feature 
//   assignments.
//
bool CObjectSpec::SymbolAllFeaturesMustBeAssigned(CMQL_execution_environment *pEE)
{
  // Get features of object type
  std::list<feature_info> l;
  if (!pEE->pDB->GetFeaturesForObjectType(m_object_type_id, l))
    return false;

  // Iterate through list
  std::list<feature_info>::const_iterator ci(l.begin());
  std::list<feature_info>::const_iterator cend(l.end());
  while (ci != cend) {
    bool bMakeFeatureAssignment;
    if (m_feature_assignments == 0 && !ci->get_is_computed())
      bMakeFeatureAssignment = true;
    else 
      bMakeFeatureAssignment = !ci->get_is_computed() 
	&& !m_feature_assignments->list_has_feature(ci->get_name());
    if (bMakeFeatureAssignment) {
      //
      // Make feature name
      //
      std::string* pFeature_name;
      pFeature_name = new std::string(ci->get_name());

      //
      // Make expression
      //
      CExpression* pExpr;
      std::string* pString;

      // Decode feature type
      id_d_t ft;
      long string_length;
      decode_feature_type_id(ci->get_type(), ft, string_length);
      
      // Decide type
      switch (ft) {
      case FEATURE_TYPE_INTEGER:
	pExpr = new CExpression(string2long(ci->get_default_value()));
	break;
      case FEATURE_TYPE_ID_D:
	pExpr = new CExpression(string2long(ci->get_default_value()));
	break;
      case FEATURE_TYPE_STRING:
	pString = new std::string(ci->get_default_value());
	pExpr = new CExpression(pString, kExprString);
	break;
      case FEATURE_TYPE_ASCII:
	pString = new std::string(ci->get_default_value());
	pExpr = new CExpression(pString, kExprString);
	break;
      default: // Enumerations
	pString = new std::string(ci->get_default_value());
	pExpr = new CExpression(pString, kExprIdentifier);
	break;
      }

      //
      // Make feature assignment
      //
      CFeatureAssignment* pNewFA;
      pNewFA = new CFeatureAssignment(pFeature_name, pExpr, m_feature_assignments);

      //
      // Insert it at the beginning of the list
      //
      m_feature_assignments = pNewFA;
    }
    ++ci;
  }

  // If we got this far, there were no errors
  return true;
}

std::list<feature_info>::const_iterator LocalListHasFeature(const std::list<feature_info>& l, const std::string& feature_name)
{
  std::list<feature_info>::const_iterator ci(l.begin());
  std::list<feature_info>::const_iterator cend(l.end());
  while (ci != cend) {
    if (strcmp_nocase(ci->get_name(), feature_name) == 0)
      return ci;
    ++ci;
  }
  // If we got this far, the feature was not in the list
  return cend;
}

//
// - All features must be exist.
//
bool CObjectSpec::SymbolAllFeaturesMustExist(CMQL_execution_environment *pEE, bool& bResult)
{
  // Get features of object type
  std::list<feature_info> l;
  if (!pEE->pDB->GetFeaturesForObjectType(m_object_type_id, l))
    return false;

  // Assume they all exist until otherwise is proven
  bResult = true;

  // Iterate through list of feature assignments.
  // Find all the ones that aren't there before returning
  CFeatureAssignment* pFA = m_feature_assignments;
  while (pFA != 0) {
    std::list<feature_info>::const_iterator ci;
    std::list<feature_info>::const_iterator cend(l.end());
    ci = LocalListHasFeature(l, pFA->get_feature_name());
    if (ci == cend) {
      // We don't have the feature
      bResult = false;
      pEE->pError->mql_append_error("Object type " + *m_object_type_name + " does not have feature " + pFA->get_feature_name() + ".\n");
    } else {
      // We do have the feature.  Set info.
      pFA->set_is_computed(ci->get_is_computed());
    }
      
    // Get next
    pFA = pFA->Next();
  }

  // If we got this far, there were no db errors.
  // bResult has already been set.
  return true;
}

bool CObjectSpec::SymbolNoFeatureMayBeComputed(CMQL_execution_environment *pEE, bool& bResult)
{
  // Only perform check if there are feature assignments
  if (m_feature_assignments != 0) {
    if (!m_feature_assignments->SymbolNoFeatureMayBeComputed(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  } 

  // If we got this far, there were no compiler errors and no DB errors
  bResult = true;
  return true;
}

//
// - Assign a type to each feature assignment
//
bool CObjectSpec::TypeAssignTypes(CMQL_execution_environment *pEE)
{
  if (m_feature_assignments != 0)
    return m_feature_assignments->TypeAssignTypes(pEE, m_object_type_id);
  else /* The list of features specs may be empty, and the type may have no features. */
    return true; 
}

//
// - Check for type compatibility between feature and expression
//
bool CObjectSpec::TypeTypeCompatibility(CMQL_execution_environment *pEE, bool& bResult)
{
  if (m_feature_assignments != 0)
    return m_feature_assignments->TypeTypeCompatibility(pEE, bResult);
  else /* The list of features specs may be empty, and the type may have no features. */
    return true; 
}


//
// Make a std::list<feature_info> from the list
//
bool CObjectSpec::ExecMakeFeatureList(CMQL_execution_environment *pEE, std::list<feature_info>& features)
{
  if (m_feature_assignments != 0)
    return m_feature_assignments->ExecMakeFeatureList(pEE, features);
  else /* The list of features specs may be empty, and the type may have no features. */
    return true; 
}





//////////////////////////////////////////////////////////////
//
// CID_D 
//
//////////////////////////////////////////////////////////////

CID_D::CID_D(id_d_t id_d, CID_D* next)
{
  m_id_d = id_d;
  m_next = next;
}

CID_D::~CID_D()
{
  delete m_next;
}

//
// - Check that no id_d in the list is NIL
// 
void CID_D::WeedNoneIsNIL(CMQL_execution_environment *pEE, bool& bResult)
{
  // First check next
  if (m_next != 0) {
    m_next->WeedNoneIsNIL(pEE, bResult);
    if (!bResult)
      return;
  }

  // Then deal with ourselves
  bResult = m_id_d != NIL;
  if (!bResult) {
    pEE->pError->mql_append_error("Cannot have NIL in the list of id_ds.\n");
  }
}

//
// - Check that all the objects exist in the given type.
//   The parameter object_type_id must be the object type id 
//   of object_type_name.
//
bool CID_D::SymbolObjectsExistInType(CMQL_execution_environment *pEE, const std::string& object_type_name, id_d_t object_type_id, bool& bResult)
{
  // First deal with next
  if (m_next != 0) {
    if (!m_next->SymbolObjectsExistInType(pEE, object_type_name, object_type_id, bResult))
      return false;
    if (!bResult)
      return true;
  }

  // Then deal with ourselves
  if (!pEE->pDB->ObjectID_DExistsInType(m_id_d, object_type_name, bResult))
    return false;
  if (!bResult) {
    pEE->pError->mql_append_error("Object with id_d " + id_d2string(m_id_d) + " does not exist in object type " + object_type_name + ".\n");
    return true;
  } else {
    // Set m_object_type_id
    m_object_type_id = object_type_id;
  }

  // If we got this far, there were no DB errors and no compiler-errors
  bResult = true;
  return true;
}

//
// - Check that all the objects exist
//
bool CID_D::SymbolObjectsExist(CMQL_execution_environment *pEE, bool& bResult)
{
  // First deal with next
  if (m_next != 0) {
    if (!m_next->SymbolObjectsExist(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }

  // Then deal with ourselves
  if (!pEE->pDB->ObjectID_DExists(m_id_d, bResult, m_object_type_id))
    return false;
  if (!bResult) {
    pEE->pError->mql_append_error("Object with id_d " + id_d2string(m_id_d) + " does not exist.\n");
    return true;
  }

  // If we got this far, there were no DB errors and no compiler-errors
  bResult = true;
  return true;
}

bool CID_D::SymbolAllAreOfSameType(CMQL_execution_environment *pEE, id_d_t object_type_id, bool& bResult)
{
  // First deal with next
  if (m_next != 0) {
    if (!m_next->SymbolAllAreOfSameType(pEE, object_type_id, bResult))
      return false;
    if (!bResult)
      return true;
  }

  // Then deal with ourselves
  bResult = m_object_type_id == object_type_id;
  if (!bResult) {
    pEE->pError->mql_append_error("Objects in id_d list are not all of the same type.\n");
    return true;
  }
  
  // If we got this far, there were no DB errors and no compiler-errors
  bResult = true;
  return true;
}

void CID_D::ExecMakeList(std::list<id_d_t>& result)
{
  // First deal with next
  if (m_next != 0) {
    m_next->ExecMakeList(result);
  }

  // Then deal with ourselves
  result.push_back(m_id_d);
}

//////////////////////////////////////////////////////////////
//
// CFeature 
//
//////////////////////////////////////////////////////////////

CFeature::CFeature(std::string* feature, CFeature* next)
{
  m_feature = feature;
  m_next = next;
}

CFeature::~CFeature()
{
  delete m_next;
  delete m_feature;
}

bool CFeature::SymbolFeaturesExist(CMQL_execution_environment *pEE, id_d_t object_type_id, bool& bResult)
{
  // Deal with next first
  if (m_next != 0) {
    if (!m_next->SymbolFeaturesExist(pEE, object_type_id, bResult))
      return false;
    if (!bResult)
      return true;
  }

  // Then deal with ourselves
  std::string dummy_default_value;
  if (!pEE->pDB->FeatureExists(*m_feature,
			  object_type_id, 
			  bResult,
			  m_feature_type_id, dummy_default_value,
			  m_is_computed))
    return false;
  if (!bResult) {
    pEE->pError->mql_append_error("Feature does not exist for the given object type.\n");
    return true;
  }

  // If we got this far, there were no DB errors and no compiler errors
  bResult = true;
  return true;
}

bool CFeature::TypeFeatureName(CMQL_execution_environment *pEE, bool& bResult)
{
  // Deal with next first
  if (m_next != 0) {
    if (!m_next->TypeFeatureName(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }

  // Then deal with ourselves
  if (!pEE->pDB->Type_id_to_TableColumnType(m_feature_type_id,
					    m_tc_type,
					    m_enum_name))
    return false;
  if (!bResult) {
    pEE->pError->mql_append_error("Enumeration does not exist for the given object type.\n");
    return true;
  }

  // If we got this far, there were no DB errors and no compiler errors
  bResult = true;
  return true;
}

void CFeature::ExecMakeFeatureList(std::list<feature_info>& feature_infos)
{
  // Deal with next first
  if (m_next != 0) {
    m_next->ExecMakeFeatureList(feature_infos);
  }

  // Then deal with ourselves
  feature_infos.push_back(feature_info(*m_feature, 
				       m_feature_type_id, 
				       "",  // Default value
				       m_is_computed));
}

void CFeature::ExecMakeNameList(CMQLResult *pResult)
{
  // Deal with next first
  if (m_next != 0) {
    m_next->ExecMakeNameList(pResult);
  }

  //
  // Then deal with ourselves
  //
  pResult->append_header(*m_feature, m_tc_type, m_enum_name);
}


//////////////////////////////////////////////////////////////
//
// CByMonads
//
//////////////////////////////////////////////////////////////

CByMonads::CByMonads(CMonadSetElement* monad_set) : m_monad_set(monad_set)
{
}

CByMonads::~CByMonads()
{
  delete m_monad_set;
}

// 
// - Check that all the ranges of monads are positive and monotonic
//
void CByMonads::MyWeed(CMQL_execution_environment *pEE, bool& bResult)
{
  // Monad-ranges should be monotonic
  m_monad_set->WeedMonotonic(pEE, bResult);
  if (!bResult)
    return;

  // Monad-ranges should be positive
  m_monad_set->WeedPositive(pEE, bResult);
  if (!bResult)
    return;

  // If we got this far, there were no DB errors and no compiler errors
  bResult = true;
  return;
}

//
// - Build the set of monads
//
bool CByMonads::MyMonads(CMQL_execution_environment *pEE, bool& bResult)
{
  //
  // - Build the set of monads
  //
  m_monad_set->MonadsBuildSet(m_monads);
  
  // There were no DB errors and no compiler errors
  bResult = true;
  return true;
} 

bool CByMonads::GetObjectsByMonads(CMQL_execution_environment *pEE, const std::string& object_type_name, std::list<id_d_t>& objects)
{
  // Get objects
  CTable intermediate_result;
  if (!pEE->pDB->GetObjectsHavingMonadsInMonadSet(object_type_name,
						   m_monads,
						   intermediate_result))
    return false;

  TableIterator i = intermediate_result.iterator();
  while (i.hasNext()) {
    // Get id_d in string form
    std::string str_id_d = intermediate_result.get_column(i, 3);

    // Put at back
    objects.push_back(string2id_d(str_id_d));

    // Advance inst iterator
    i.next();
  }

  // If we got this far, there were no DB errors
  return true;
}









