/*
 * mql_query.cpp
 *
 * MQL query classes
 *
 * Ulrik Petersen
 * Created: 2/27-2001
 * Last update: 3/16-2004
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2004  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <mql_query.h>
#include <mql_error.h>
#include <mql_symboltable.h>
#include <mql_extern.h>
#include <mql_R.h>
#include <string_func.h>
#include <assert.h>
#include <mql_execution_environment.h>
#include <cstdlib>
#include <logging.h>
#include <monads.h>


////////////////////////////////////////////////////////////
//
// CGap_block 
//
////////////////////////////////////////////////////////////
CGap_block::CGap_block(eRetrieval retrieval, CBlocks *opt_blocks)
  : m_retrieval(retrieval), m_opt_blocks(opt_blocks)
{
}

CGap_block::~CGap_block()
{
  delete m_opt_blocks;
}

void CGap_block::WeedTildeOperator(CMQL_execution_environment *pEE, bool& bResult)
{
  if (m_opt_blocks) {
    m_opt_blocks->WeedTildeOperator(pEE, bResult);
  }
}

bool CGap_block::Symbol(CMQL_execution_environment *pEE, bool& bResult)
{
  //
  // - Check the blocks
  //
  if (m_opt_blocks != 0) {
    if (!m_opt_blocks->Symbol(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }

  
  // If we got this far, there were no DB errors and no compiler errors
  bResult = true;
  return true;
}

bool CGap_block::Type(CMQL_execution_environment *pEE, bool& bResult)
{
  // Check opt_blocks
  if (m_opt_blocks != 0) {
    if (!m_opt_blocks->Type(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }

  // If we got this far, there were no DB errors and no compiler-errors
  bResult = true;
  return true;
}





////////////////////////////////////////////////////////////
//
// CBlock 
//
////////////////////////////////////////////////////////////
CBlock::CBlock(COpt_gap_block* opt_gap_block)
  : m_kind(kOptGapBlock), 
    m_gap_block(0),
    m_opt_gap_block(opt_gap_block),
    m_object_block(0)
{
}

CBlock::CBlock(CGap_block* gap_block)
  : m_kind(kGapBlock), 
    m_gap_block(gap_block),
    m_opt_gap_block(0),
    m_object_block(0)
{
}

CBlock::CBlock(CObject_block* object_block)
  : m_kind(kObjectBlock), 
    m_gap_block(0),
    m_opt_gap_block(0),
    m_object_block(object_block)
{
}

CBlock::~CBlock()
{
  delete m_gap_block;
  delete m_opt_gap_block;
  delete m_object_block;
}

void CBlock::WeedTildeOperator(CMQL_execution_environment *pEE, bool& bResult)
{
  switch(m_kind) {
  case kObjectBlock:
    m_object_block->WeedTildeOperator(pEE, bResult);
    break;
  case kOptGapBlock:
    m_opt_gap_block->WeedTildeOperator(pEE, bResult);
    break;
  case kGapBlock:
    m_gap_block->WeedTildeOperator(pEE, bResult);
    break;
  }
}


void CBlock::WeedFirstLast(CMQL_execution_environment *pEE, bool& bResult, bool is_first, bool is_last)
{
  switch(m_kind) {
  case kObjectBlock:
    m_object_block->WeedFirstLast(pEE, bResult, is_first, is_last);
    break;
  case kOptGapBlock:
  case kGapBlock:
    bResult = true;
    break;
  }
}

bool CBlock::Symbol(CMQL_execution_environment *pEE, bool& bResult)
{
  switch(m_kind) {
  case kObjectBlock:
    if (!m_object_block->Symbol(pEE, bResult))
      return false;
    if (!bResult)
      return true;
    break;
  case kOptGapBlock:
    if (!m_opt_gap_block->Symbol(pEE, bResult))
      return false;
    if (!bResult)
      return true;
    break;
  case kGapBlock:
    if (!m_gap_block->Symbol(pEE, bResult))
      return false;
    if (!bResult)
      return true;
    break;
  }
  bResult = true;
  return true;
}

bool CBlock::Type(CMQL_execution_environment *pEE, bool& bResult)
{
  switch(m_kind) {
  case kObjectBlock:
    if (!m_object_block->Type(pEE, bResult))
      return false;
    if (!bResult)
      return true;
    break;
  case kOptGapBlock:
    if (!m_opt_gap_block->Type(pEE, bResult))
      return false;
    if (!bResult)
      return true;
    break;
  case kGapBlock:
    if (!m_gap_block->Type(pEE, bResult))
      return false;
    if (!bResult)
      return true;
    break;
  }
  bResult = true;
  return true;
}

// Is it an opt_gap_block or a gap_block?
bool CBlock::is_gap_or_opt_gap_block(void)
{
  return m_kind == kOptGapBlock || m_kind == kGapBlock;
}

void CBlock::CalculateMMap(String2COBPtrMMap& mmap)
{
  switch(m_kind) {
  case kObjectBlock:
    m_object_block->CalculateMMap(mmap);
    break;
  case kOptGapBlock:
    // Don't do anything
    break;
  case kGapBlock:
    // Don't do anything
    break;
  }
}

bool CBlock::AggregateQuery(CMQL_execution_environment *pEE, const CSet_of_monad_ms& characteristic_set, String2COBPtrMMap& mmap)
{
  bool bDoIt = false;
  std::string characteristic_string;
  int no_of_object_blocks;
  std::pair<String2COBPtrMMap::iterator,String2COBPtrMMap::iterator> mypair;
  String2COBPtrMMap::iterator p;
  switch(m_kind) {
  case kObjectBlock:
    // Do it if there is more than one with this characteristic string
    characteristic_string = m_object_block->get_characteristic_string();
    no_of_object_blocks = 0;    
    mypair = mmap.equal_range(characteristic_string);
    for (p = mypair.first; p != mypair.second; ++p) {
      ++no_of_object_blocks;
    }
   
    if (no_of_object_blocks > 1) {
      bDoIt = true;
    }


    // Do it if there are feature constraints which yield
    // something which can be tested for by SQL constraints
    if (!bDoIt && m_object_block->get_pre_query_string().length() > 0) {
      bDoIt = true;
    }

    // Do it if opt_blocks is non-null 
    // (and we need to recurse down through it)
    if (!bDoIt && m_object_block->get_opt_blocks() != 0) {
      bDoIt = true;
    }

    // Shall we do it?
    if (bDoIt) {
      return m_object_block->AggregateQuery(pEE, characteristic_set, mmap);
    }
    break;
  case kOptGapBlock:
    // Don't do it, since we don't have an object_type_name to do it for (?)
    break;
  case kGapBlock:
    // Don't do it, since we don't have an object_type_name to do it for (?)
    break;
  }

  // If we came this far, there were no DB errors
  return true;
}

bool CBlock::HasFirstLast() const
{
  switch(m_kind) {
  case kObjectBlock:
    return m_object_block->HasFirstLast();
    break;
  case kOptGapBlock:
  case kGapBlock:
    return false; // Gap blocks can't have first/last
    break;
  }

  // If we got here, there was no First or Last in this object
  return false;
}

////////////////////////////////////////////////////////////
//
// CRest_of_block_str 
//
////////////////////////////////////////////////////////////
CRest_of_block_str::CRest_of_block_str(CBlock* block)
  : m_block(block),
    m_rest_of_block_str(0)
{
}

CRest_of_block_str::CRest_of_block_str(CBlock* block, 
				       CRest_of_block_str* rest_of_block_str)
  : m_block(block),
    m_rest_of_block_str(rest_of_block_str)
{
}

CRest_of_block_str::~CRest_of_block_str()
{
  delete m_block;
  delete m_rest_of_block_str;
}

void CRest_of_block_str::WeedTildeOperator(CMQL_execution_environment *pEE, bool& bResult)
{
  if (is_block()) {
    m_block->WeedTildeOperator(pEE, bResult);
  } else {
    m_block->WeedTildeOperator(pEE, bResult);
    if (!bResult)
      return;
    m_rest_of_block_str->WeedTildeOperator(pEE, bResult);
  }
}


void CRest_of_block_str::WeedFirstLast(CMQL_execution_environment *pEE, bool& bResult, bool is_first, bool is_last)
{
  if (is_block()) {
    m_block->WeedFirstLast(pEE, bResult, is_first, is_last);
  } else {
    m_block->WeedFirstLast(pEE, bResult, is_first, false);
    if (!bResult)
      return;
    m_rest_of_block_str->WeedFirstLast(pEE, bResult, false, is_last);
  }
}

bool CRest_of_block_str::Symbol(CMQL_execution_environment *pEE, bool& bResult)
{
  if (is_block()) {
    if (!m_block->Symbol(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  } else {
    if (!m_block->Symbol(pEE, bResult)) 
      return false;
    if (!bResult)
      return true;
    if (!m_rest_of_block_str->Symbol(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }
  bResult = true;
  return true;
}

bool CRest_of_block_str::Type(CMQL_execution_environment *pEE, bool& bResult)
{
  if (is_block()) {
    if (!m_block->Type(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  } else {
    if (!m_block->Type(pEE, bResult)) 
      return false;
    if (!bResult)
      return true;
    if (!m_rest_of_block_str->Type(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }
  bResult = true;
  return true;
}

void CRest_of_block_str::CalculateMMap(String2COBPtrMMap& mmap)
{
  m_block->CalculateMMap(mmap);
  if (m_rest_of_block_str != 0) {
    m_rest_of_block_str->CalculateMMap(mmap);
  }
}

bool CRest_of_block_str::AggregateQuery(CMQL_execution_environment *pEE, const CSet_of_monad_ms& characteristic_set, String2COBPtrMMap& mmap)
{
  // Do block
  if (!m_block->AggregateQuery(pEE, characteristic_set, mmap))
    return false;

  // Do rest_of_block_str
  if (m_rest_of_block_str != 0) {
    if (!m_rest_of_block_str->AggregateQuery(pEE, characteristic_set, mmap))
      return false;
  }

  // If we came this far, there were no DB errors
  return true;
}

bool CRest_of_block_str::HasFirstLast() const
{
  if (m_block->HasFirstLast()) {
    return true;
  } else if (m_rest_of_block_str != 0) {
    return m_rest_of_block_str->HasFirstLast();
  } else {
    // If we got here, there was no First or Last in this object
    return false;
  }
}


////////////////////////////////////////////////////////////
//
// CObject_reference_usage 
//
////////////////////////////////////////////////////////////
CObject_reference_usage::CObject_reference_usage(std::string* object_reference, 
					     std::string* feature_name)
  : m_object_reference(object_reference),
    m_feature_name(feature_name)
{
}

CObject_reference_usage::~CObject_reference_usage()
{
  delete m_object_reference;
  delete m_feature_name;
}

//
// - Check that object references have been declared 
//   before being used.
//
// - Check that feature on object reference usage exists 
//   for the given object type.  At the same time, store
//   the information in the symbol table that this feature
//   will be used for this object reference.
//
bool CObject_reference_usage::Symbol(CMQL_execution_environment *pEE, bool& bResult)
{
  //
  // - Check that object references have been declared 
  //   before being used.
  //
  if (!pEE->pSymbolTable->exists(*m_object_reference, m_index_in_symboltable)) {
    pEE->pError->mql_append_error("The object reference has not been declared in the object reference usage '" + *m_object_reference + "." + *m_feature_name + "'.\n");
    bResult = false;
    return true;
  }

  //
  // - Check that feature on object reference usage exists 
  //   for the given object type.  At the same time, store
  //   the information in the symbol table that this feature
  //   will be used for this object reference.
  //
  CMQLSymbolTableEntry* pSTE = pEE->pSymbolTable->get(m_index_in_symboltable);
  bool bFeatureExists;
  id_d_t feature_type_id;
  std::string default_value;
  bool is_computed;
  if (!pEE->pDB->FeatureExists(*m_feature_name,
			  pSTE->get_object_type_id(),
			  bFeatureExists,
			  feature_type_id,
			  default_value,
			  is_computed))
    return false;
  if (!bFeatureExists) {
    pEE->pError->mql_append_error("The feature does not exist on the object type of the object reference in the object reference usage '" + *m_object_reference + "." + *m_feature_name + "'.\n");
    bResult = false;
    return true;
  }

  // Create feature-info
  m_feature_info = feature_info(*m_feature_name, feature_type_id, default_value, is_computed);

  // Add feature_info to symbol table entry
  m_index_in_symboltable_entry = pSTE->add_feature(m_feature_info);

  // Get object block
  CObject_block_common *pObjectBlock = pEE->pSymbolTable->get_object_block(m_index_in_symboltable);

  // Add feature_info to CObject_block_common's CMQLObject, 
  // if not already there
  // XXX: This must be uncommented-out if
  // R_object_reference_declaration_inst is to work.
  pObjectBlock->get_object()->add_feature(m_feature_info);

  // If we got this far, there were no DB errors and no compiler-errors
  bResult = true;
  return true;
}



////////////////////////////////////////////////////////////
//
// CValue 
//
////////////////////////////////////////////////////////////
CValue::CValue(std::string* str, eValueKind kind) // for enum_const and string
  : m_emdf_value(0),
    m_object_reference_usage(0),
    m_enum_const(0),
    m_integer(0),
    m_string(0)
{
  assert(kind == kValEnumConst || kind == kValString);
  m_kind = kind;
  switch (kind) {
  case kValEnumConst:
    m_enum_const = str;
    break;
  case kValString:
    m_string = str;
    break;
  default:
    assert(false);
    break;
  }
}

CValue::CValue(long integer)
  : m_emdf_value(0),
    m_object_reference_usage(0),
    m_enum_const(0),
    m_integer(integer),
    m_string(0),
    m_kind(kValInteger)
{
}

CValue::CValue(CObject_reference_usage* object_reference_usage)
  : m_emdf_value(0),
    m_object_reference_usage(object_reference_usage),
    m_enum_const(0),
    m_integer(0),
    m_string(0),
    m_kind(kValObjectReferenceUsage)
{
}

CValue::~CValue()
{
  delete m_enum_const;
  delete m_string;
  delete m_object_reference_usage;
  delete m_emdf_value;
}

//
// - Check that object references have been declared 
//   before being used.
//
// - Check that feature on object reference usage exists 
//   for the given object type.  At the same time, store
//   the information in the symbol table that this feature
//   will be used for this object reference.
//
bool CValue::Symbol(CMQL_execution_environment *pEE, id_d_t feature_type_id, bool& bResult)
{
  bool bEnumConstExists;
  bool dummy_is_default;
  switch (m_kind) {
  case kValEnumConst:
    if (!pEE->pDB->EnumConstExists(*m_enum_const,
				   feature_type_id,
				   bEnumConstExists,
				   m_enum_const_value,
				   dummy_is_default))
      return false;
    if (!bEnumConstExists) {
      bResult = false;
      pEE->pError->mql_append_error("The enumeration constant " + *m_enum_const + " does not exist\nin the enumeration of the feature.\n");
      return true;
    }

    // Now we know the enum id
    m_enum_id = feature_type_id;
    break;
  case kValInteger:
    break;
  case kValString:
    break;
  case kValObjectReferenceUsage:
    if (!m_object_reference_usage->Symbol(pEE, bResult))
      return false;
    if (!bResult)
      return true;
    break;
  }

  // If we got this far, there were no DB errors and no compiler-errors
  bResult = true;
  return true;
}

const CEMdFValue* CValue::get_emdf_value(void)
{
  if (m_emdf_value == 0) {
    switch (m_kind) {
    case kValEnumConst:
      m_emdf_value = new CEMdFValue(kEVEnum, m_enum_const_value);
      break;
    case kValInteger:
      m_emdf_value = new CEMdFValue(kEVInt, m_integer);
      break;
    case kValString:
      m_emdf_value = new CEMdFValue(*m_string);
      break;
    case kValObjectReferenceUsage:
      assert(false);
      break;
    }
  }
  return m_emdf_value;
}

std::string CValue::get_as_string(CMQL_execution_environment *pEE) const
{
  CMQLSymbolTableEntry* pSTE;
  std::string result;
  const CEMdFValue* pValue;
  switch (m_kind) {
  case kValEnumConst:
    result = long2string(m_enum_const_value);
    break;
  case kValInteger:
    result = long2string(m_integer);
    break;
  case kValString:
    result = *m_string;
    break;
  case kValObjectReferenceUsage:
    assert(pEE);
    pSTE = pEE->pSymbolTable->get(m_object_reference_usage->get_index_in_symboltable());
    assert(pSTE != 0);
    pSTE->get_feature_value(m_object_reference_usage->get_index_in_symboltable_entry(), &pValue);
    pValue->get_as_string(result);
    break;
  }

  // Return result;
  return result;
}

const std::string& CValue::get_enum_const()
{
  assert(m_kind == kValEnumConst);
  return *m_enum_const;
}

long CValue::get_integer()
{
  assert(m_kind == kValInteger);
  return m_integer;
}

const std::string& CValue::get_string()
{
  assert(m_kind == kValString);
  return *m_string;
}

CObject_reference_usage* CValue::get_object_reference_usage()
{
  assert(m_kind == kValObjectReferenceUsage);
  return m_object_reference_usage;
}



////////////////////////////////////////////////////////////
//
// CFeature_comparison 
//
////////////////////////////////////////////////////////////
CFeature_comparison::CFeature_comparison(std::string* feature_name,
					 eComparisonOp comparison_op,
					 CValue* value)
  : m_feature_name(feature_name),
    m_comparison_op(comparison_op),
    m_value(value),
    m_pcre(NULL), m_pcre_extra(NULL), 
    m_ovector(0), m_ovectorsize(0)
{
}

CFeature_comparison::~CFeature_comparison()
{
  delete m_feature_name;
  delete m_value;

  // Regular expressions
  if (m_pcre != NULL)
    pcre_free(m_pcre);
  if (m_pcre_extra != NULL)
    pcre_free(m_pcre_extra);
  delete[] m_ovector;
}

//
// - If this is the tilde operator, check that the value is a string.
//
// - Also check/compile regular expression
//
void CFeature_comparison::WeedTildeOperator(CMQL_execution_environment *pEE, bool& bResult)
{
  if (m_comparison_op == kTilde || m_comparison_op == kNotTilde) {
    if (m_value->get_kind() == kValString) {
      //
      // Compile
      //
      const char *pcre_error;
      int error_offset;
      int options = 0;
      m_pcre = pcre_compile(m_value->get_string().c_str(), options, &pcre_error, &error_offset, NULL);
      if (m_pcre == NULL) {
	pEE->pError->mql_append_error(std::string("Error compiling regular expression.  PCRE reported:\n") + pcre_error + "\n");
	pEE->pError->mql_append_error(std::string("Error occured at characterposition ") + int2string(error_offset) + " in the pattern.\n");
	bResult = false;
	return;
      }
      //
      // Study
      //
      options = 0;
      m_pcre_extra = pcre_study(m_pcre, options, &pcre_error);
      if (pcre_error != NULL) {
	pEE->pError->mql_append_error(std::string("Error studying regular expression.  PCRE reported:\n") + pcre_error + "\n");
	bResult = false;
	return;
      }

      // 
      // Make ovector
      //
      int capture_count;
      int fullinfo_return_value;
      if ((fullinfo_return_value = pcre_fullinfo(m_pcre, m_pcre_extra, PCRE_INFO_CAPTURECOUNT, &capture_count)) != 0) {
	pEE->pError->mql_append_error(std::string("Error getting info from regular expression.\nmql_query.cpp:CFeature_comparison::WeedTildeOperator: pcre_fullinfo returned ") + int2string(fullinfo_return_value) + "\n");
	m_ovector = 0;
	m_ovectorsize = 0;
	bResult = false;
	return;
      } else {
	m_ovectorsize = (capture_count + 1)*3;
	m_ovector = new int[m_ovectorsize];
      }

      // If we got this far, there were no compiler errors
      bResult = true;
    }
    else {
      pEE->pError->mql_append_error("The '~' and '!~' tilde operators (regular expression matching) can only be used with strings on the right-hand side.\n");
      bResult = false;
    }
  } else {
    bResult = true;
  }
}

//
// - Check that feature_name is part of the enclosing object type.
//   In doing so, store its feature type_id in the local object.
//
// - Symbol-checking of value
//
bool CFeature_comparison::Symbol(CMQL_execution_environment *pEE, id_d_t enclosing_object_type_id, bool& bResult)
{
  //
  // - Check that feature_name is part of the enclosing object type.
  //   In doing so, store its feature info in the local object.
  //
  bool bFeatureExists;
  id_d_t feature_type_id;
  std::string default_value;
  bool is_computed;
  if (!pEE->pDB->FeatureExists(*m_feature_name,
			  enclosing_object_type_id,
			  bFeatureExists,
			  feature_type_id,
			  default_value,
			  is_computed))
    return false;
  if (!bFeatureExists) {
    pEE->pError->mql_append_error("Feature " + *m_feature_name + " does not exist on its enclosing object type.\n");
    bResult = false;
    return true;
  }
  m_feature_info = feature_info(*m_feature_name, feature_type_id, default_value, is_computed);

  // Check that it is in the enclosing object
  
  //
  // - Symbol-checking of value
  //
  if (!m_value->Symbol(pEE, feature_type_id, bResult)) 
    return false; 
  if (!bResult)
    return true;

  // If we got this far, we had no DB errors and no compiler-errors
  bResult = true;
  return true;
}

void CFeature_comparison::SymbolAddFeatures(CMQLObject* pObj)
{
  pObj->add_feature(m_feature_info);
  m_feature_index = pObj->get_feature_index(*m_feature_name);
}

bool CFeature_comparison::Type(CMQL_execution_environment *pEE, bool& bResult)
{
  // Assume everything went well
  bResult = true;

  // Decode feature type
  id_d_t ft;
  long string_length;
  id_d_t feature_type_id = m_feature_info.get_type();
  decode_feature_type_id(feature_type_id, ft, string_length);

  // What is value's kind?
  CMQLSymbolTableEntry* pSTE;
  switch (m_value->get_kind()) {
  case kValEnumConst:
    bResult = ft == m_value->get_enum_id();
    if (!bResult) {
      pEE->pError->mql_append_error("The feature " + *m_feature_name + " is being compared with an enumeration constant.  The enumeration constant does not belong to the enumeration which is the type of the feature.\n");
    }
    break;
  case kValInteger:
    bResult = ft == FEATURE_TYPE_INTEGER
      || ft == FEATURE_TYPE_ID_D;
    if (!bResult) {
      pEE->pError->mql_append_error("The feature " + *m_feature_name + " is being compared with an integer.  The feature is neither an integer nor an id_d.\n");
    }
    break;
  case kValString:
    bResult = ft == FEATURE_TYPE_STRING || ft == FEATURE_TYPE_ASCII;
    if (!bResult) {
      pEE->pError->mql_append_error("The feature " + *m_feature_name + " is being compared with a string.  The feature is not a string.\n");
    }
    break;
  case kValObjectReferenceUsage:
    pSTE = pEE->pSymbolTable->get(m_value->get_object_reference_usage()->get_index_in_symboltable());
    bResult = feature_type_id == pSTE->get_feature_type_id(m_value->get_object_reference_usage()->get_index_in_symboltable_entry());
    if (!bResult) {
      pEE->pError->mql_append_error("The feature " + *m_feature_name + " is being compared with an object reference usage whose feature is of a different type.\n");
    }
    break;
  }

  // If we got this far, there were no DB errors
  return true;
}


bool CFeature_comparison::compare(CMQL_execution_environment *pEE, const CEMdFValue *left_value)
{
  // Compare
  bool bResult;
  int pcre_result;
  int start_offset = 0;
  std::string right_value;
  std::string left_value_str;
  const CEMdFValue *pRightValue = 0;
  CMQLSymbolTableEntry *pSTE = 0;
  switch (m_comparison_op) {
  case kEqual:
  case kLessThan:
  case kGreaterThan:
  case kNotEqual:
  case kLessThanOrEqual:
  case kGreaterThanOrEqual:
    switch (m_value->get_kind()) {
    case kValEnumConst:
    case kValInteger:
    case kValString:
      pRightValue = m_value->get_emdf_value();
      bResult = left_value->compare(*pRightValue, m_comparison_op);
      break;
    case kValObjectReferenceUsage:
      pSTE = pEE->pSymbolTable->get((m_value->get_object_reference_usage())->get_index_in_symboltable());
      assert(pSTE != 0);
      pSTE->get_feature_value((m_value->get_object_reference_usage())->get_index_in_symboltable_entry(), &pRightValue);
      assert(pRightValue != 0);
      bResult = left_value->compare(*pRightValue, m_comparison_op);
      break;
    default:
      assert(false);
      break;
    }
    break;
  case kTilde:
    right_value = m_value->get_as_string(pEE);
    left_value_str = left_value->get_string();
    pcre_result = pcre_exec(m_pcre, m_pcre_extra, left_value_str.c_str(), left_value_str.length(), start_offset, PCRE_NOTEMPTY, m_ovector, m_ovectorsize);
    if (pcre_result <= 0) {
      bResult = false;
    } else {
      bResult = true;
    }
    break;
  case kNotTilde:
    right_value = m_value->get_as_string(pEE);
    left_value_str = left_value->get_string();
    pcre_result = pcre_exec(m_pcre, m_pcre_extra, left_value_str.c_str(), left_value_str.length(), start_offset, PCRE_NOTEMPTY, m_ovector, m_ovectorsize);
    if (pcre_result == PCRE_ERROR_NOMATCH) {
      bResult = true;
    } else {
      bResult = false;
    }
    break;
  }

  // Return result
  return bResult;
}


std::string CFeature_comparison::make_constraints(bool bContextIsPositive) const
{
  // Check for self feature
  bool bIsSelf = strcmp_nocase(*m_feature_name, "self") == 0;

  // Feature must be stored, but self is OK, even though it is computed
  if (!bIsSelf && m_feature_info.get_is_computed()) {
    return "";
  }

  // Construct SQL feature name
  std::string SQL_feature_name;
  SQL_feature_name = EMdFDB::encode_feature_name_for_prequery_string(*m_feature_name);

  // Comparison operator must have a counterpart in SQL
  if (m_comparison_op != kEqual
      && m_comparison_op != kLessThan
      && m_comparison_op != kGreaterThan
      && m_comparison_op != kNotEqual
      && m_comparison_op != kLessThanOrEqual
      && m_comparison_op != kGreaterThanOrEqual) {
    return "";
  }


  // Value must be atomic, not an object reference
  if (m_value->get_kind() == kValObjectReferenceUsage) {
    return "";
  }


  // If we got this far, we can return the feature comparison
  std::string result;
  std::string value_string;

  id_d_t type_id_d = m_feature_info.get_type();

  value_string = m_value->get_as_string(0); // This is a hack, but we don't need pEE if we aren't doing object references.  Therefore, we can set it to nil.

  result = SQL_feature_name;
  result += " ";
  result += eComparisonOp2SQLString(m_comparison_op);
  result += " ";
  if (feature_type_id_is_STRING(type_id_d)) {
    value_string = string2dbstring(value_string);
  } else if (feature_type_id_is_ASCII(type_id_d)) {
    value_string = encodeASCIIstring(value_string);
  }
  result += "'" + value_string + "'"; 

  return result;
}




////////////////////////////////////////////////////////////
//
// CFFactor 
//
////////////////////////////////////////////////////////////
CFFactor::CFFactor(CFFactor* ffactor)
  : m_ffactor(ffactor),
    m_ffeatures(0),
    m_feature_comparison(0)
{
}

CFFactor::CFFactor(CFFeatures* ffeatures)
  : m_ffactor(0),
    m_ffeatures(ffeatures),
    m_feature_comparison(0)
{
}

CFFactor::CFFactor(CFeature_comparison* feature_comparison)
  : m_ffactor(0),
    m_ffeatures(0),
    m_feature_comparison(feature_comparison)
{
}

CFFactor::~CFFactor()
{
  delete m_ffactor;
  delete m_ffeatures;
  delete m_feature_comparison;
}

void CFFactor::WeedTildeOperator(CMQL_execution_environment *pEE, bool& bResult)
{
  if (is_NOT()) {
    m_ffactor->WeedTildeOperator(pEE, bResult);
  } else if (is_parenthesis()) {
    m_ffeatures->WeedTildeOperator(pEE, bResult);
  } else { // is_feature_comparison
    m_feature_comparison->WeedTildeOperator(pEE, bResult);
  }
}

bool CFFactor::Symbol(CMQL_execution_environment *pEE, id_d_t enclosing_object_type_id, bool& bResult)
{
  if (is_NOT()) {
    return m_ffactor->Symbol(pEE, enclosing_object_type_id, bResult);
  } else if (is_parenthesis()) {
    return m_ffeatures->Symbol(pEE, enclosing_object_type_id, bResult);
  } else { // is_feature_comparison
    return m_feature_comparison->Symbol(pEE, enclosing_object_type_id, bResult);
  }
}

void CFFactor::SymbolAddFeatures(CMQLObject* pObj)
{
  if (is_NOT()) {
    m_ffactor->SymbolAddFeatures(pObj);
  } else if (is_parenthesis()) {
    m_ffeatures->SymbolAddFeatures(pObj);
  } else { // is_feature_comparison
    m_feature_comparison->SymbolAddFeatures(pObj);
  }
}

bool CFFactor::Type(CMQL_execution_environment *pEE, bool& bResult)
{
  if (is_NOT()) {
    return m_ffactor->Type(pEE, bResult);
  } else if (is_parenthesis()) {
    return m_ffeatures->Type(pEE, bResult);
  } else { // is_feature_comparison
    return m_feature_comparison->Type(pEE, bResult);
  }
}

std::string CFFactor::make_constraints(bool bContextIsPositive) const
{
  if (is_NOT()) {
    std::string ffactor = m_ffactor->make_constraints(!bContextIsPositive);
    if (ffactor.length() == 0) {
      return "";
    } else {
      return std::string("NOT ") + ffactor;
    }
  } else if (is_parenthesis()) {
    std::string ffeatures = m_ffeatures->make_constraints(bContextIsPositive);
    if (ffeatures.length() == 0) {
      return "";
    } else {
      return std::string("(") + ffeatures + ")";
    }
  } else { // is_feature_comparison
    return m_feature_comparison->make_constraints(bContextIsPositive);
  }
}


////////////////////////////////////////////////////////////
//
// CFTerm 
//
////////////////////////////////////////////////////////////
CFTerm::CFTerm(CFFactor* ffactor)
  : m_ffactor(ffactor),
    m_fterm(0)
{
}

CFTerm::CFTerm(CFFactor* ffactor, CFTerm* fterm)
  : m_ffactor(ffactor),
    m_fterm(fterm)
{
}

CFTerm::~CFTerm()
{
  delete m_ffactor;
  delete m_fterm;
}

void CFTerm::WeedTildeOperator(CMQL_execution_environment *pEE, bool& bResult)
{
  if (is_ffactor()) {
    m_ffactor->WeedTildeOperator(pEE, bResult);
  } else {
    m_ffactor->WeedTildeOperator(pEE, bResult);
    if (!bResult)
      return;
    m_fterm->WeedTildeOperator(pEE, bResult);
  }
}
bool CFTerm::Symbol(CMQL_execution_environment *pEE, id_d_t enclosing_object_type_id, bool& bResult)
{
  if (is_ffactor()) {
    return m_ffactor->Symbol(pEE, enclosing_object_type_id, bResult);
  } else {
    if (!m_ffactor->Symbol(pEE, enclosing_object_type_id, bResult))
      return false;
    if (!bResult)
      return true;
    return m_fterm->Symbol(pEE, enclosing_object_type_id, bResult);
  }
}

void CFTerm::SymbolAddFeatures(CMQLObject* pObj)
{
  if (is_ffactor()) {
    m_ffactor->SymbolAddFeatures(pObj);
  } else {
    m_ffactor->SymbolAddFeatures(pObj);
    m_fterm->SymbolAddFeatures(pObj);
  }
}

bool CFTerm::Type(CMQL_execution_environment *pEE, bool& bResult)
{
  if (is_ffactor()) {
    return m_ffactor->Type(pEE, bResult);
  } else {
    if (!m_ffactor->Type(pEE, bResult))
      return false;
    if (!bResult)
      return true;
    return m_fterm->Type(pEE, bResult);
  }
}

std::string CFTerm::make_constraints(bool bContextIsPositive) const
{
  if (is_ffactor()) {
    return m_ffactor->make_constraints(bContextIsPositive);
  } else {
    std::string A, B;
    A = m_ffactor->make_constraints(bContextIsPositive);
    B = m_fterm->make_constraints(bContextIsPositive);
    if (!bContextIsPositive) {
      // In a negative context, AND works like OR.  Therefore,
      // if even one feature cannot be prequeried, neither 
      // must be prequeried. This is because it would miss
      // an extension to the query.
      if (A.length() == 0 || B.length() == 0) {
	return "";
      } else {
	return A + " AND " + B;
      }
    } else { // bContextIsPositive == true
      // In a positive context, it does no harm to leave one out. 
      // This is because taking one away from an AND does not restrict 
      // but rather enlarge the output from the SQL engine, and hence
      // it doesn't matter; the non-prequeriable feature
      // can be checked in the backend.
      if (A.length() == 0 && B.length() == 0) {
	return "";
      } else if (A.length() > 0 && B.length() == 0) {
	return A;
      } else if (A.length() == 0 && B.length() > 0) {
	return B;
      } else {
	return A + " AND " + B;
      }
    }
  }
}


////////////////////////////////////////////////////////////
//
// CFFeatures 
//
////////////////////////////////////////////////////////////
CFFeatures::CFFeatures(CFTerm* fterm)
  : m_fterm(fterm),
    m_ffeatures(0),
    m_constraints(0)
{
}

CFFeatures::CFFeatures(CFFeatures* ffeatures, CFTerm* fterm)
  : m_fterm(fterm),
    m_ffeatures(ffeatures),
    m_constraints(0)
{
}

CFFeatures::~CFFeatures()
{
  delete m_fterm;
  delete m_ffeatures;
  delete m_constraints;
}

void CFFeatures::WeedTildeOperator(CMQL_execution_environment *pEE, bool& bResult)
{
  if (is_fterm()) {
    m_fterm->WeedTildeOperator(pEE, bResult);
  } else {
    m_ffeatures->WeedTildeOperator(pEE, bResult);
    if (!bResult)
      return;
    m_fterm->WeedTildeOperator(pEE, bResult);
  }
}

bool CFFeatures::Symbol(CMQL_execution_environment *pEE, id_d_t enclosing_object_type_id, bool& bResult)
{
  if (is_fterm()) {
    return m_fterm->Symbol(pEE, enclosing_object_type_id, bResult);
  } else {
    if (!m_ffeatures->Symbol(pEE, enclosing_object_type_id, bResult))
      return false;
    if (!bResult)
      return true;
    return m_fterm->Symbol(pEE, enclosing_object_type_id, bResult);
  }
}

void CFFeatures::SymbolAddFeatures(CMQLObject* pObj)
{
  if (is_fterm()) {
    m_fterm->SymbolAddFeatures(pObj);
  } else {
    m_ffeatures->SymbolAddFeatures(pObj);
    m_fterm->SymbolAddFeatures(pObj);
  }
}


bool CFFeatures::Type(CMQL_execution_environment *pEE, bool& bResult)
{
  if (is_fterm()) {
    return m_fterm->Type(pEE, bResult);
  } else {
    if (!m_ffeatures->Type(pEE, bResult))
      return false;
    if (!bResult)
      return true;
    return m_fterm->Type(pEE, bResult);
  }
}

const std::string& CFFeatures::get_constraints(void)
{
  if (m_constraints == 0) {
    m_constraints = new std::string;
    *m_constraints = make_constraints(true);
  } 
  return *m_constraints;
}

std::string CFFeatures::make_constraints(bool bContextIsPositive) const
{
  if (is_fterm()) {
    return m_fterm->make_constraints(bContextIsPositive);
  } else {
    std::string A, B;
    A = m_ffeatures->make_constraints(bContextIsPositive);
    B = m_fterm->make_constraints(bContextIsPositive);
    if (bContextIsPositive) {
      // In a positive context, if even one cannot be prequeried,
      // neither must be prequeried. This is because it would miss
      // an extension to the query.
      if (A.length() == 0 || B.length() == 0) {
	return "";
      } else {
	return A + " OR " + B;
      }
    } else { // bContextIsPositive == false
      // In a negative context, OR works like AND, and hence
      // it does no harm to leave one out. This is because
      // taking one away from an AND does not restrict but rather
      // enlarge the output from the SQL engine, and hence
      // it doesn't matter; the non-prequeriable feature
      // can be checked in the backend.
      if (A.length() == 0 && B.length() == 0) {
	return "";
      } else if (A.length() > 0 && B.length() == 0) {
	return A;
      } else if (A.length() == 0 && B.length() > 0) {
	return B;
      } else {
	return A + " OR " + B;
      }
    }
  }
}



////////////////////////////////////////////////////////////
//
// CObject_block_common 
//
////////////////////////////////////////////////////////////
CObject_block_common::CObject_block_common(std::string* object_type_name,
					   std::string* object_reference,
					   eRetrieval retrieval,
					   eFirstLast first_last,
					   CFFeatures* feature_constraints,
					   CFeature* pFeature_retrieval,
					   CBlocks* opt_blocks)
  : m_object_type_name(object_type_name),
    m_object_reference(object_reference),
    m_retrieval(retrieval),
    m_first_last(first_last),
    m_feature_constraints(feature_constraints),
    m_pFeature_retrieval(pFeature_retrieval),
    m_opt_blocks(opt_blocks),
    m_object_reference_index(StrSTEPairVec::no_element),
    m_inst(0)
{
  m_object = new CMQLObject();
}

CObject_block_common::~CObject_block_common()
{
  delete m_object_type_name;
  delete m_object_reference;
  delete m_feature_constraints;
  delete m_pFeature_retrieval;
  delete m_opt_blocks;
  delete m_object; 
  delete m_inst; 
}

void CObject_block_common::WeedTildeOperator(CMQL_execution_environment *pEE, bool& bResult)
{
  if (m_feature_constraints != 0) {
    m_feature_constraints->WeedTildeOperator(pEE, bResult);
    if (!bResult)
      return;
  }
  if (m_opt_blocks != 0) {
    m_opt_blocks->WeedTildeOperator(pEE, bResult);
    if (!bResult)
      return;
  }
}

//
// - First check ourselves
//
// - Then check the opt_blocks
//
void CObject_block_common::WeedFirstLast(CMQL_execution_environment *pEE, bool& bResult, bool is_first, bool is_last)
{
  switch (m_first_last) {
  case kFirst:
    bResult = is_first;
    if (!bResult) {
      pEE->pError->mql_append_error("The 'first' modifier must only be used for the first object block in a blocks.\n");
    }
    break;
  case kLast:
    bResult = is_last;
    if (!bResult) {
      pEE->pError->mql_append_error("The 'last' modifier must only be used for the last object block in a blocks.\n");
    }
    break;
  case kFirstAndLast:
    bResult = is_first && is_last;
    if (!bResult) {
      pEE->pError->mql_append_error("The 'first and last' modifier must only be used object blocks that are both first and last in their blocks.\n");
    }
    break;
  case kNoFirstLast:
    bResult = true;
    break;
  }

  if (bResult && m_opt_blocks != 0) {
    m_opt_blocks->Weed(pEE, bResult);
  }
}

//
// Checks whether object type exists.
// Sets m_object_type_id if it does.
bool CObject_block_common::SymbolObjectTypeExists(CMQL_execution_environment *pEE, bool& bObjectTypeExists)
{
  bool bDBResult;
  bDBResult = pEE->pDB->ObjectTypeExists(*m_object_type_name,
					 bObjectTypeExists, m_object_type_id);


  // Was there a DB error?
  if (!bDBResult) {
    pEE->pError->mql_append_error("DB error checking existence of object type name '" + *m_object_type_name + "'.\n");
    return false;
  }

  // Was the object type there?
  if (!bObjectTypeExists) {
    pEE->pError->mql_append_error("Object type name '" + *m_object_type_name + "' does not exist.\n");
  }

  // Set m_object_type_index
  std::string OTN = NormalizeOTName(*m_object_type_name);
  m_object_type_index = pEE->addObjectTypeName(OTN);

  // If we got this far, there were no DB errors.
  return true;
}


//
// - Check that object_type_name is a valid type
//
// - Check that the object reference declaration is only declared once
//
// - Declare object reference
//
// - Check FFeatures
// 
// - Add all features used to local symbol table
//
// - Check the opt_blocks
//
bool CObject_block_common::Symbol(CMQL_execution_environment *pEE, bool& bResult)
{
  //
  // - Check that object_type_name is a valid type
  //
  if (!SymbolObjectTypeExists(pEE, bResult))
    return false;
  if (!bResult)
    return true;

  //
  // - Check that the object reference declaration is only declared once
  //
  if (m_object_reference->length() > 0) {
    short int dummy_id;
    if (pEE->pSymbolTable->exists(*m_object_reference, dummy_id)) {
      pEE->pError->mql_append_error("Object reference " + *m_object_reference + " has already been declared once.  An object reference must not be declared more than once.\n");
      bResult = false;
      return true;
    }
  }

  // 
  // - Declare object reference
  // 
  if (m_object_reference->length() > 0) {
    CMQLSymbolTableEntry* pSTE;
    pSTE = new CMQLSymbolTableEntry(*m_object_reference,
				    *m_object_type_name,
				    m_object_type_id);
    set_object_reference_index(pEE->pSymbolTable->insert(pSTE, this));
  }
  
  //
  // - Check FFeatures
  // 
  if (m_feature_constraints != 0) {
    if (!m_feature_constraints->Symbol(pEE, m_object_type_id, bResult))
      return false;
    if (!bResult)
      return true;
  }

  //
  // - Check feature_retrieval
  //
  if (m_pFeature_retrieval != 0) {
    if (!m_pFeature_retrieval->SymbolFeaturesExist(pEE, m_object_type_id, bResult))
      return false;
    if (!bResult)
      return true;
  }

  //
  // - Add all features used to local symbol table
  //
  if (m_feature_constraints != 0) {
    m_feature_constraints->SymbolAddFeatures(m_object);
  }

  //
  // - Check the blocks
  //
  if (m_opt_blocks != 0) {
    if (!m_opt_blocks->Symbol(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }

  
  // If we got this far, there were no DB errors and no compiler errors
  bResult = true;
  return true;
}

//
// - First check the ffeatures
//
// - Then check the opt_blocks
//
bool CObject_block_common::Type(CMQL_execution_environment *pEE, bool& bResult)
{
  // Check ffeatures
  if (m_feature_constraints != 0) {
    if (!m_feature_constraints->Type(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }

  // Check opt_blocks
  if (m_opt_blocks != 0) {
    if (!m_opt_blocks->Type(pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }

  // If we got this far, there were no DB errors and no compiler-errors
  bResult = true;
  return true;
}

bool CObject_block_common::make_inst(CMQL_execution_environment *pEE, const CSet_of_monad_ms& Su, String2COBPtrMMap& mmap)
{
  // Get characteristic string
  std::string characteristic_string = get_characteristic_string();

  // See if we can find it already
  std::pair<String2COBPtrMMap::iterator,String2COBPtrMMap::iterator> mypair = mmap.equal_range(characteristic_string);
  String2COBPtrMMap::iterator p;
  bool bDoContinue = true;
  for (p = mypair.first; bDoContinue && p != mypair.second; ++p) {
    CInst *pInst = (p->second)->m_inst;
    if (pInst != 0) {
      m_inst = new CInst(pInst);
      bDoContinue = false;
    }
  }

  if (m_inst == 0) {
    // We didn't find it.  Calculate it.
    try {
      m_inst = R_inst(pEE, Su, this);
      m_inst->set_is_aggregate(true);
    } catch (CEMdFDBException e) {
      return false;
    }
  }
  return true;
}

void CObject_block_common::CalculateMMap(String2COBPtrMMap& mmap)
{
  // Calculate pre-query string
  calculate_pre_query_string();

  // Get characteristic string
  std::string characteristic_string;
  calculate_characteristic_string();
  characteristic_string = get_characteristic_string();

  // Add this to the mmap
  mmap.insert(String2COBPtrMMap::value_type(characteristic_string, this));

  // Recurse down through inner blocks
  if (m_opt_blocks != 0) {
    m_opt_blocks->CalculateMMap();
  }
}

void CObject_block_common::calculate_pre_query_string(void)
{
  if (m_feature_constraints == 0) {
    m_pre_query_string = "";
  } else {
    m_pre_query_string = m_feature_constraints->get_constraints();
  }
}

void CObject_block_common::calculate_characteristic_string(void)
{
  // Get object type name
  m_characteristic_string = *m_object_type_name + ",";

  // Get pre_query_string
  m_characteristic_string += get_pre_query_string();

  // Get vector of feature names
  const std::vector<std::string>& feature_name_vector = m_object->get_feature_names();

  // Add feature names to set
  std::vector<std::string>::const_iterator civ;
  std::vector<std::string>::const_iterator civend(feature_name_vector.end());
  for (civ = feature_name_vector.begin(); civ != civend; ++civ) {
    m_characteristic_string += "," + *civ;
  }
}

bool CObject_block_common::AggregateQuery(CMQL_execution_environment *pEE, const CSet_of_monad_ms& characteristic_set, String2COBPtrMMap& mmap)
{
  // Write to log
  // LOG_WRITE_TIME("CObject_block_common::AggregateQuery", *m_object_type_name + " Making inst.");

  // Do it for this
  if (!make_inst(pEE, characteristic_set, mmap))
    return false;
  // LOG_WRITE_TIME("CObject_block_common::AggregateQuery", *m_object_type_name + " Inst-making finished.");

  // Recurse down through the opt_blocks
  if (m_opt_blocks != 0) {
    // Get big-union of inst.
    CSet_of_monad_ms CS;
    m_inst->big_union(CS);

    // Recurse downwards, but only if CS is not empty!
    // If it is empty, there are no monads within which to look,
    // and we will get a NoMonadsException.
    // Also, it means that the query from this object_block_common
    // is going to return no results, since an empty big_union means
    // there are no objects in the inst.  This means that there will be
    // no matches inside it either, so there is no point in 
    // aggregately querying inside.
    if (!CS.is_empty()) {
      // LOG_WRITE_TIME("CObject_block_common::AggregateQuery ", *m_object_type_name + " Recursing down through opt_blocks.");
      if (!m_opt_blocks->AggregateQuery(pEE, CS))
	return false;
      // LOG_WRITE_TIME("CObject_block_common::AggregateQuery ", *m_object_type_name + " Recursion finished.");
    }
  }

  // If we came this far, there were no DB errors
  return true;
}


bool CObject_block_common::HasFirstLast() const
{
  return m_first_last != kNoFirstLast;
}



////////////////////////////////////////////////////////////
//
// CObject_block_first 
//
////////////////////////////////////////////////////////////
CObject_block_first::CObject_block_first(std::string* object_type_name,
					 std::string* object_reference,
					 eRetrieval retrieval,
					 eFirstLast first_last,
					 CFFeatures* feature_constraints,
					 CFeature* pFeature_retrieval,
					 CBlocks* opt_blocks)
  : CObject_block_common(object_type_name,
			 object_reference,
			 retrieval,
			 first_last,
			 feature_constraints,
			 pFeature_retrieval,
			 opt_blocks)
{
}

CObject_block_first::~CObject_block_first()
{
}

////////////////////////////////////////////////////////////
//
// CObject_block 
//
////////////////////////////////////////////////////////////
CObject_block::CObject_block(std::string* object_type_name,
			     std::string* object_reference,
			     eRetrieval retrieval,
			     eFirstLast last,
			     CFFeatures* feature_constraints,
			     CFeature *pFeature_retrieval,
			     CBlocks* opt_blocks)
  : CObject_block_common(object_type_name,
			 object_reference,
			 retrieval,
			 last,
			 feature_constraints,
			 pFeature_retrieval,
			 opt_blocks)
{
}

CObject_block::~CObject_block()
{
}



////////////////////////////////////////////////////////////
//
// CBlock_str 
//
////////////////////////////////////////////////////////////
CBlock_str::CBlock_str(CObject_block_first* object_block_first)
  : m_object_block_first(object_block_first),
    m_rest_of_block_str(0)
{
}

CBlock_str::CBlock_str(CObject_block_first* object_block_first,
	   CRest_of_block_str* rest_of_block_str)
  : m_object_block_first(object_block_first),
    m_rest_of_block_str(rest_of_block_str)
{
}

CBlock_str::~CBlock_str()
{
  delete m_object_block_first;
  delete m_rest_of_block_str;
}

void CBlock_str::WeedTildeOperator(CMQL_execution_environment *pEE, bool& bResult)
{
  if (is_object_block_first()) {
    m_object_block_first->WeedTildeOperator(pEE, bResult);
  } else {
    m_object_block_first->WeedTildeOperator(pEE, bResult);
    if (!bResult)
      return;
    m_rest_of_block_str->WeedTildeOperator(pEE, bResult);
  }
}


void CBlock_str::WeedFirstLast(CMQL_execution_environment *pEE, bool& bResult, bool is_first, bool is_last)
{
  if (is_object_block_first()) {
    m_object_block_first->WeedFirstLast(pEE, bResult, is_first, is_last);
  } else {
    m_object_block_first->WeedFirstLast(pEE, bResult, is_first, false);
    if (!bResult)
      return;
    m_rest_of_block_str->WeedFirstLast(pEE, bResult, false, is_last);
  }
}


bool CBlock_str::Symbol(CMQL_execution_environment *pEE, bool& bResult)
{
  if (is_object_block_first()) {
    return m_object_block_first->Symbol(pEE, bResult);
  } else {
    if (!m_object_block_first->Symbol(pEE, bResult))
      return false;
    if (!bResult)
      return true;
    return m_rest_of_block_str->Symbol(pEE, bResult);
  }
}

bool CBlock_str::Type(CMQL_execution_environment *pEE, bool& bResult)
{
  if (is_object_block_first()) {
    return m_object_block_first->Type(pEE, bResult);
  } else {
    if (!m_object_block_first->Type(pEE, bResult))
      return false;
    if (!bResult)
      return true;
    return m_rest_of_block_str->Type(pEE, bResult);
  }
}

const std::string& CBlock_str::get_leftmost_object_type_name()
{
  return m_object_block_first->get_object_type_name();
}

void CBlock_str::CalculateMMap(String2COBPtrMMap& mmap) 
{
  m_object_block_first->CalculateMMap(mmap);
  if (m_rest_of_block_str != 0) {
    m_rest_of_block_str->CalculateMMap(mmap);
  }
}

bool CBlock_str::AggregateQuery(CMQL_execution_environment *pEE, const CSet_of_monad_ms& characteristic_set, String2COBPtrMMap& mmap)
{
  // Do it for this object
  if (!m_object_block_first->AggregateQuery(pEE, characteristic_set, mmap))
    return false;

  // Recurse down through the rest_of_block_str
  if (m_rest_of_block_str != 0) {
    if (!m_rest_of_block_str->AggregateQuery(pEE, characteristic_set, mmap))
      return false;
  }

  return true;
}

bool CBlock_str::HasFirstLast() const
{
  if (m_object_block_first->HasFirstLast()) {
    return true;
  } else if (m_rest_of_block_str != 0) {
    return m_rest_of_block_str->HasFirstLast();
  } else {
    // If we got here, there was no First or Last in this object
    return false;
  }
}


////////////////////////////////////////////////////////////
//
// CPower 
//
////////////////////////////////////////////////////////////
// m_limit is 0
CPower::CPower()
  : m_limit(0)
{
}

CPower::CPower(long limit)
  : m_limit(limit)
{
}

CPower::~CPower()
{
}




////////////////////////////////////////////////////////////
//
// CBlock_string 
//
////////////////////////////////////////////////////////////
CBlock_string::CBlock_string(CBlock_str* block_str)
  : m_block_str(block_str),
    m_power(0),
    m_block_string(0)
{
}

CBlock_string::CBlock_string(CBlock_str* block_str, 
			     CPower* power, 
			     CBlock_string* block_string)
  : m_block_str(block_str),
    m_power(power),
    m_block_string(block_string)
{
}

CBlock_string::~CBlock_string()
{
  delete m_block_str;
  delete m_power;
  delete m_block_string;
}

void CBlock_string::WeedTildeOperator(CMQL_execution_environment *pEE, bool& bResult)
{
  if (is_block_str()) {
    m_block_str->WeedTildeOperator(pEE, bResult);
  } else {
    m_block_str->WeedTildeOperator(pEE, bResult);
    if (!bResult)
      return;
    m_block_string->WeedTildeOperator(pEE, bResult);
  }
}


void CBlock_string::WeedFirstLast(CMQL_execution_environment *pEE, bool& bResult, bool is_first, bool is_last)
{
  if (is_block_str()) {
    m_block_str->WeedFirstLast(pEE, bResult, is_first, is_last);
  } else {
    m_block_str->WeedFirstLast(pEE, bResult, is_first, false);
    if (!bResult)
      return;
    m_block_string->WeedFirstLast(pEE, bResult, false, is_last);
  }
}

bool CBlock_string::Symbol(CMQL_execution_environment *pEE, bool& bResult)
{
  if (is_block_str()) {
    return m_block_str->Symbol(pEE, bResult);
  } else {
    if (!m_block_str->Symbol(pEE, bResult))
      return false;
    if (!bResult)
      return true;
    return m_block_string->Symbol(pEE, bResult);
  }
}

bool CBlock_string::Type(CMQL_execution_environment *pEE, bool& bResult)
{
  if (is_block_str()) {
    return m_block_str->Type(pEE, bResult);
  } else {
    if (!m_block_str->Type(pEE, bResult))
      return false;
    if (!bResult)
      return true;
    return m_block_string->Type(pEE, bResult);
  }
}

const std::string& CBlock_string::get_leftmost_object_type_name()
{
  return m_block_str->get_leftmost_object_type_name();
}

CObject_block_first *CBlock_string::get_leftmost_object_block_first()
{
  return m_block_str->get_object_block_first();
}

void CBlock_string::CalculateMMap(String2COBPtrMMap& mmap) 
{
  m_block_str->CalculateMMap(mmap);
  if (m_block_string != 0) {
    m_block_string->CalculateMMap(mmap);
  }
}

bool CBlock_string::AggregateQuery(CMQL_execution_environment *pEE, const CSet_of_monad_ms& characteristic_set, String2COBPtrMMap& mmap)
{
  // Do block_str
  if (!m_block_str->AggregateQuery(pEE, characteristic_set, mmap))
    return false;

  // Do block_string after power
  if (m_block_string != 0) {
    if (!m_block_string->AggregateQuery(pEE, characteristic_set, mmap))
      return false;
  }

  // If we got this far, there were no DB errors
  return true;
}

bool CBlock_string::HasFirstLast() const
{
  if (m_block_str->HasFirstLast()) {
    return true;
  } else if (m_block_str != 0) {
    return m_block_str->HasFirstLast();
  } else {
    // If we got here, there was no First or Last in this object
    return false;
  }
}



////////////////////////////////////////////////////////////
//
// CBlocks 
//
////////////////////////////////////////////////////////////
CBlocks::CBlocks(CBlock_string* block_string)
  : m_block_string(block_string)
{
}

CBlocks::~CBlocks()
{
  delete m_block_string;
}

void CBlocks::WeedTildeOperator(CMQL_execution_environment *pEE, bool& bResult)
{
  m_block_string->WeedTildeOperator(pEE, bResult);
}

void CBlocks::Weed(CMQL_execution_environment *pEE, bool& bResult)
{
  this->WeedTildeOperator(pEE, bResult);
  if (!bResult)
    return;
  m_block_string->WeedFirstLast(pEE, bResult, true, true);
}

bool CBlocks::Symbol(CMQL_execution_environment *pEE, bool& bResult)
{
  return m_block_string->Symbol(pEE, bResult);
}

bool CBlocks::Type(CMQL_execution_environment *pEE, bool& bResult)
{
  return m_block_string->Type(pEE, bResult);
}

const std::string& CBlocks::get_leftmost_object_type_name()
{
  return m_block_string->get_leftmost_object_type_name();
}

CObject_block_first *CBlocks::get_leftmost_object_block_first()
{
  return m_block_string->get_leftmost_object_block_first();
}

bool CBlocks::AggregateQuery(CMQL_execution_environment *pEE, const CSet_of_monad_ms& characteristic_set)
{
  return m_block_string->AggregateQuery(pEE, characteristic_set, m_mmap);
}

bool CBlocks::HasFirstLast() const
{
  return m_block_string->HasFirstLast();
}

void CBlocks::CalculateMMap(void) 
{
  m_block_string->CalculateMMap(m_mmap);
}



////////////////////////////////////////////////////////////
//
// CTopograph 
//
////////////////////////////////////////////////////////////
CTopograph::CTopograph(CBlocks* blocks)
  : m_blocks(blocks)
{
}

CTopograph::~CTopograph()
{
  delete m_blocks;
}

void CTopograph::Weed(CMQL_execution_environment *pEE, bool& bResult)
{
  m_blocks->Weed(pEE, bResult);
}

bool CTopograph::Symbol(CMQL_execution_environment *pEE, bool& bResult)
{ 
  // Symbol-check
  return m_blocks->Symbol(pEE, bResult);
}

bool CTopograph::Type(CMQL_execution_environment *pEE, bool& bResult)
{
  // Type-check
  return m_blocks->Type(pEE, bResult);
}

bool CTopograph::AggregateQuery(CMQL_execution_environment *pEE, const CSet_of_monad_ms& characteristic_set)
{
  return m_blocks->AggregateQuery(pEE, characteristic_set);
}

bool CTopograph::HasFirstLast() const
{
  return m_blocks->HasFirstLast();
}

void CTopograph::CalculateMMap(void)
{
  m_blocks->CalculateMMap();
}
