/*
 * conn.h
 *
 * Database connection abstract base class.
 *
 * Ulrik Petersen
 * Created: 1/27-2001
 * Last update: 6/9-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/




#ifndef EMDF_CONN__H__
#define EMDF_CONN__H__

#include <string>
#include <sstream>

class EMdFNULLValueException {};

class EMdFConnection {
public:
    // Constructor and destructor
    EMdFConnection();
    virtual ~EMdFConnection();

    // Status of connection
    // Returns true on connection good, false on connection bad
    virtual bool ConnectionOk(void) = 0;

    // Executing
    virtual bool Exec(const std::string& query) = 0;

    // Get last error message
    // Returns true on error message available 
    // (even if it is "OK"), false on error 
    // message not available.
    virtual bool ErrorMessage(std::string& out) = 0;

    // Transactions
    virtual bool BeginTransaction(void) = 0;
    virtual bool CommitTransaction(void) = 0;
    virtual bool AbortTransaction(void) = 0;

    // Tuple-counts
    // Both return -1 on not appliccable or error
    virtual long NoOfResultTuples(void) = 0;  // SELECT
    virtual long NoOfCommandTuples(void) = 0; // INSERT, UPDATE, DELETE

    // Field count
    // Returns -1 on no fields.
    virtual int NoOfFields(void) = 0;

    // Tuple-access
    // First tuple is 0
    // First field is 0.
    virtual bool AccessTuple(std::string& out, long tuple_no, int field_no) = 0;
    virtual bool AccessTuple(long tuple_no, 
			     int field_no, 
			     const char* calling_method,
			     bool& result) = 0;
    virtual bool AccessTuple(long tuple_no, 
			     int field_no, 
			     const char* calling_method,
			     long& result) = 0;
    virtual bool AccessTuple(long tuple_no, 
			     int field_no, 
			     const char* calling_method,
			     std::string& result) = 0;

    // Release results so they don't take up memory
    virtual void ReleaseResults(void) = 0;

    // UseDatabase
    virtual bool UseDatabase(const std::string& database_name);

    // Vacuum
    virtual bool Vacuum(bool bAnalyze) { return true; };

    // Local error messages
    void append_local_error(std::string error);
    void clear_local_error(void);
    std::string get_local_error(void);

protected:

private:
    // Make the assignment operator and copy constructor private
    // because they are not supported.
    EMdFConnection(const EMdFConnection&);
    EMdFConnection& operator=(const EMdFConnection&);

    // Local error message
    std::string m_local_errormessage;
};


#endif // EMDF_CONN__H__
