/*
 * mql_types.h
 *
 * Definitions of AST-classes for MQL
 *
 * Ulrik Petersen
 * Created: 2/27-2001
 * Last update: 3/22-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/




#ifndef MQL_TYPES__H__
#define MQL_TYPES__H__

#ifndef SWIG
#include <mql_result.h>
#include <emdf.h>
#include <string>
#include <infos.h>
#include <monads.h>
#include <assert.h>


class CTopograph; // Forward declaration
class CMQL_execution_environment; // Forward declaration


typedef enum { kAdd, kUpdate, kRemove } eAddUpdateRemove;

#endif // !defined SWIG

// Base class for all statements
class CStatement {
private:
  bool m_bDeleteResult; 
protected:
  CMQLResult* m_result;
  CMQL_execution_environment *m_pEE;
public:
  CStatement(CMQL_execution_environment *pEE);
  virtual ~CStatement();
  CMQLResult* get_result();
#ifndef SWIG
  virtual void Weed(bool& bResult);
  virtual bool Symbol(bool& bResult);
  virtual bool Type(bool& bResult);
  virtual bool Monads(bool& bResult);
  virtual bool Exec();
  void set_delete_result(bool bDoDelete);
#endif // !defined SWIG
};

#ifndef SWIG
typedef enum { kInteger, kString, kASCII, kID_D, kEnumeration } eType;

// Helper class
class CType {
private:
  eType m_type;
  std::string* m_Enumeration;
  id_d_t m_type_id;
  long m_string_length;
public:
  CType(std::string* Enumeration); // For Enumeration
  CType(eType type, long string_length);  // for String and ASCII
  CType(eType type); // For all others
  virtual ~CType();
  eType get_type(void) const { return m_type; };
  const std::string& get_Enumeration(void) const { assert(m_type == kEnumeration); return *m_Enumeration; };
  id_d_t get_type_id(void) const { return m_type_id; };
  void WeedStringLength(CMQL_execution_environment *pEE, bool& bResult);
  bool SymbolEnumerationExists(CMQL_execution_environment *pEE, bool& bResult);
  bool TypeAssignTypeID(CMQL_execution_environment *pEE);
  long get_string_length() { return m_string_length; };
};

typedef enum { kExprInteger, kExprString, kExprIdentifier } eExprType;

// Helper class
class CExpression {
private:
  eExprType m_type;
  std::string* m_string;
  long m_integer;
  id_d_t m_enum_id;
public:
  CExpression(long integer); // For kExprInteger
  CExpression(std::string* str, eExprType type); // For kExprString and kExprIdentifier
  virtual ~CExpression();
  eExprType get_kind() { return m_type; };
  long get_integer();
  const std::string& get_string();
  const std::string& get_identifier();
  bool get_as_string(CMQL_execution_environment *pEE, std::string& result, bool bConvertEnumConstToInteger);
  bool TypeTypeCompatibility(CMQL_execution_environment *pEE, CType* type, bool& bResult);
};

// Helper class
class CMonadSetElement {
private:
  monad_m m_mse_first, m_mse_last;
  CMonadSetElement* m_next;
public:
  CMonadSetElement(monad_m mse_first, monad_m mse_last, CMonadSetElement* next);
  CMonadSetElement(monad_m mse_singleton, CMonadSetElement* next);
  virtual ~CMonadSetElement();
  monad_m get_mse_first() { return m_mse_first; };
  monad_m get_mse_last() { return m_mse_last; };
  CMonadSetElement* get_Next() { return m_next; };
  void set_Next(CMonadSetElement* next) { m_next = next; };
  void WeedMonotonic(CMQL_execution_environment *pEE, bool& bResult);
  void WeedPositive(CMQL_execution_environment *pEE, bool& bResult);
  void MonadsBuildSet(CSet_of_monad_ms& monads);
};


// Helper class: id_d_list, id_d
class CID_D {
private:
  CID_D* m_next;
  id_d_t m_id_d;
  id_d_t m_object_type_id;
public:
  CID_D(id_d_t id_d, CID_D* next);
  virtual ~CID_D();
  id_d_t get_id_d() { return m_id_d; };
  id_d_t get_object_type_id() { return m_object_type_id; };
  CID_D* Next() { return m_next; };
  void set_Next(CID_D* next) { m_next = next; };
  void WeedNoneIsNIL(CMQL_execution_environment *pEE, bool& bResult);
  void ExecMakeList(std::list<id_d_t>& result);
  bool SymbolObjectsExist(CMQL_execution_environment *pEE, bool& bResult);
  bool SymbolObjectsExistInType(CMQL_execution_environment *pEE, const std::string& object_type_name, id_d_t object_type_id, bool& bResult);
  bool SymbolAllAreOfSameType(CMQL_execution_environment *pEE, id_d_t object_type_id, bool& bResult);
};

typedef enum { kFocus, kAll } eFocusSpec;
typedef enum { kFullSheaf, kFlatSheaf } sheaf_return_type_t;


// Base class for all query statements
class CQueryBase {
protected:
  eFocusSpec m_focus;
  CMonadSetElement* m_in;
  std::string* m_using_segment;
  sheaf_return_type_t m_sheaf_return_type;
  CTopograph* m_query;
  CSet_of_monad_ms m_Su;
public:
  CQueryBase(eFocusSpec focus, 
	     CMonadSetElement* in_spec, 
	     std::string* using_segment,
	     sheaf_return_type_t sheaf_return_type,
	     CTopograph* query);
  virtual ~CQueryBase();
  virtual void MyWeed(CMQL_execution_environment *pEE, bool& bResult);
  virtual bool MySymbol(CMQL_execution_environment *pEE, bool& bResult);
  virtual bool MyType(CMQL_execution_environment *pEE, bool& bResult);
  virtual bool MyMonads(CMQL_execution_environment *pEE, bool& bResult);
  virtual bool MyExec(CMQL_execution_environment *pEE, CMQLResult **ppMQLResult);
};

// Helper class: feature_assignment
class CFeatureAssignment {
private:
  CFeatureAssignment* m_next;
  std::string* m_feature_name;
  CExpression* m_expression;
  CType* m_type;
  bool m_is_computed;
public:
  CFeatureAssignment(std::string* feature_name,
		     CExpression* expression,
		     CFeatureAssignment* next);
  virtual ~CFeatureAssignment();
  void WeedSelfNotAssigned(CMQL_execution_environment *pEE, bool& bResult);
  const std::string& get_feature_name() { return *m_feature_name; };
  CExpression* get_expression() { return m_expression; };
  CFeatureAssignment* Next() { return m_next; };
  void set_Next(CFeatureAssignment* next) { m_next = next; };
  bool list_has_feature(const std::string& feature_name);
  bool SymbolNoFeatureMayBeComputed(CMQL_execution_environment *pEE, bool& bResult);
  bool TypeAssignTypes(CMQL_execution_environment *pEE, id_d_t object_type_id);
  bool ExecMakeFeatureList(CMQL_execution_environment *pEE, std::list<feature_info>& features);
  bool TypeTypeCompatibility(CMQL_execution_environment *pEE, bool& bResult);
  void set_is_computed(bool is_computed) { m_is_computed = is_computed; };
  bool get_is_computed(void) const { return m_is_computed; };
};

// Helper class
class CFeature {
private:
  CFeature* m_next;
  std::string* m_feature;
  id_d_t m_feature_type_id;
  bool m_is_computed;
  TableColumnType m_tc_type;
  std::string m_enum_name;
public:
  CFeature(std::string* feature, CFeature* next);
  virtual ~CFeature();
  const std::string& get_feature() { return *m_feature; };
  CFeature* Next() { return m_next; };
  void set_Next(CFeature* next) { m_next = next; };
  bool SymbolFeaturesExist(CMQL_execution_environment *pEE, id_d_t object_type_id, bool& bResult);
  bool TypeFeatureName(CMQL_execution_environment *pEE, bool& bResult);
  void ExecMakeFeatureList(std::list<feature_info>& feature_infos);
  void ExecMakeNameList(CMQLResult *pResult);
};


#endif // !defined SWIG
#endif /* MQL_TYPES__H__ */
