/*
 * smart_vector.h
 *
 * CSmartVector template
 *
 * Ulrik Petersen
 * Created: 1/18-2003
 * Last update: 6/10-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/



#ifndef SMART_VECTOR__H__
#define SMART_VECTOR__H__

#include <vector>
#include <functional>

/////////////////////////////////////////////////////////////
//
// CSmartVector template
//
// CSmartVector takes a class type and a predicate as template 
// arguments. The predictate must be an "equals" predicate which
// compares two T*'s for equality.
//
// short int add_value(T* v, P p);
// - add_value only adds v if there is no member in the vector
//   already which compares equal with v using p.  It returns
//   the index of the existing value if present, and the index
//   of the new value if added.
//
// short int get_index(T* v, P p);
// - Gets the index of the value that compares equal to v using p.
//   returns no_element if not present.
//
// T* operator[](short int index);
// - Index access operator.  There is no special bounds checking.
//
// void set_value(T* v, short int index);
// - Deletes the current value at index and sets to v at index.
//   Assumes that index is valid.
//
//
//


template<class T, class P = std::equal_to<T> > class CSmartVector {
private:
  std::vector<T*> m_vec;
  bool m_bDelete;
public:
  static const short int no_element;
  CSmartVector(bool bDelete) : m_bDelete(bDelete) {};
  ~CSmartVector(void);
  // Returns index of value
  short int add_value(T* v, P p, bool& /* out */ bWasAdded);
  short int get_index(T* v, P p);
  T* operator[](short int index);
  void set_value(T* v, short int index);
  short int size(void) const { return m_vec.size(); };
};

template<class T, class P> const short int CSmartVector<T,P>::no_element = -1;

template <class T, class P> CSmartVector<T,P>::~CSmartVector<T,P>(void)
{
  for (unsigned int i = 0; i < m_vec.size(); i++) {
    if (m_bDelete) {
      delete m_vec[i];
    }
    m_vec[i] = 0;
  }
  m_vec.resize(0);
}

  // Returns index of value
template <class T, class P> short int CSmartVector<T,P>::add_value(T* v, P p, bool& /* out */ bWasAdded)
{
  bWasAdded = false;
  typename std::vector<T*>::const_iterator ci(m_vec.begin());
  typename std::vector<T*>::const_iterator ciend(m_vec.end());
  short int index = 0;
  while (ci != ciend) {
    if (p((*ci), v)) {
      return index;
    }
    index++;
    ci++;
  }

  // It wasn't there, so add it
  bWasAdded = true;
  m_vec.push_back(v);
  return index;
}

template <class T, class P> short int CSmartVector<T,P>::get_index(T* v, P p)
{
  typename std::vector<T*>::const_iterator ci(m_vec.begin());
  typename std::vector<T*>::const_iterator ciend(m_vec.end());
  short int index = 0;
  while (ci != ciend) {
    if (p((*ci), v)) {
      return index;
    }
    index++;
    ci++;
  }
  return no_element;
}

template <class T, class P> T* CSmartVector<T,P>::operator[](short int index) 
{
  return m_vec[index];
}

template <class T, class P> void CSmartVector<T,P>::set_value(T* v, short int index)
{
  m_vec[index] = v;
}



#endif // SMART_VECTOR__H__
