/*
 * emdftry.cpp
 *
 * Trial of EMdF framework
 * Created: 1/27-2001
 * Last update: 6/23-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <iostream>
#include <string>
#include <infos.h>
#include <default_password.h>
#include <emdros_environment.h>
#if USE_POSTGRESQL
#include <pgemdfdb.h>
#elif USE_MYSQL
#include <mysqlemdfdb.h>
#endif


#define DB_ERROR_MESSAGE  { \
    std::string dberr_msg = pDB->ErrorMessage(); \
    std::string err_msg; \
    if (dberr_msg.length() > 0) { \
      err_msg = "DB Error message = '" + dberr_msg + "'\n"; \
    } else { \
      err_msg = "No DB error message available.\n"; \
    }; \
    std::cerr << err_msg; \
    result = 1; \
    bContinue = false; \
    bAllOK = false; \
}

#define VALUE_ERROR { \
    result = 2; \
    bContinue = false; \
    bAllOK = false; \
}

void show_version(std::ostream& ostr)
{
  ostr << "emdftry from Emdros version " << EMDROS_VERSION 
       << " on " << EMdFDB::BackendName()<< std::endl;
}


void print_usage(std::ostream& ostr)
{
  show_version(ostr);
  ostr << "Usage:" << std::endl;
  ostr << "   emdftry [options] " << std::endl;
  ostr << "OPTIONS:" << std::endl;
  ostr << "   --help               Show this help" << std::endl; 
  ostr << "   -V , --version       Show version" << std::endl;
  ostr << "   -h , --host host     Set hostname to connect to" << std::endl;
  ostr << "   -u , --user user     Set database user to connect as (default: 'emdf')" << std::endl;
  ostr << "   -p , --password pwd  Set password to use for database user" << std::endl;
  ostr << "DEFAULTS:" << std::endl;
  ostr << "   -h localhost" << std::endl;
  ostr << "   -u emdf" << std::endl;
}

bool parse_argument(int argc, char* argv[], int &i, std::string& hostname, std::string& user, std::string& password)
{
  if (argv[i][0] != '-') {
    return false;
  } else { // [0] == '-'
    std::string argument(argv[i]);
    if (argument == "--help") {
      print_usage(std::cout);
      exit(0);
    } else if (argument == "-V" || argument == "--version") {
      show_version(std::cout);
      exit(0);
    } else if (argument == "-h" || argument == "--host") {
      if (i < (argc-1)) {
	++i;
	hostname = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a hostname as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " vlager" << std::endl;
	return false;
      }
    } else if (argument == "-u" || argument == "--user") {
      if (i < (argc-1)) {
	++i;
	user = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a user as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " emdf" << std::endl;
	return false;
      }
    } else if (argument == "-p" || argument == "--password") {
      if (i < (argc-1)) {
	++i;
	password = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a password as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " secret" << std::endl;
	return false;
      }
    } else {
      std::cerr << "ERROR: Unrecognized option " << argument << std::endl;
      return false;
    }
    return true;
  }
}

bool parse_arguments(int argc, char* argv[], std::string& hostname, std::string& user, std::string& password)
{
  // Set defaults
  hostname = "localhost";
  user = "emdf";

  // Get default password
#if USE_DEFAULT_PASSWORD
  get_default_password(password);
#else
  password = "";
#endif

  for (int i=1; i < argc; ++i) {
    if (!parse_argument(argc, argv, i, hostname, user, password)) 
      return false;
  }
  
  return true;

}


int main(int argc, char* argv[])
{
#if USE_POSTGRESQL
  PgEMdFDB* pDB;
#elif USE_MYSQL
  MySQLEMdFDB *pDB;
#endif
  std::string newdb;
  int result;
  bool bContinue = true;
  bool bAllOK = true;

  std::string hostname;
  std::string password;
  std::string user;
  if (!parse_arguments(argc, argv, hostname, user, password)) {
    print_usage(std::cerr);
    return 3;
  }
  

  // Make connection
#if USE_POSTGRESQL
  pDB = new PgEMdFDB(hostname, user, password, "emdf");
#elif USE_MYSQL
  pDB = new MySQLEMdFDB(hostname, user, password, "emdf");
#endif
  newdb = "EMdF1";

  // Zero-fill the password
  zero_fill_string(password);

  // Test connection
  if (!pDB->ConnectionOk()) {
    DB_ERROR_MESSAGE;

    std::cerr << "FAILURE: Could not connect to back-end server." << std::endl;
    std::cerr << "This usually has one of the following causes:" << std::endl;
    std::cerr << "  - The back-end is not running." << std::endl;
    std::cerr << "  - The back-end is not configured for incoming TCP/IP connections." << std::endl;
    std::cerr << "  - You did not supply a user/password with the -u/-p options." << std::endl;
    std::cerr << "    Note that the default user is 'emdf'." << std::endl;
    std::cerr << "  - You have not followed the instructions in the 'bootstrapping.txt'" << std::endl;
    std::cerr << "    document." << std::endl;
    std::cerr << "There are also hints in the HOW-TO-USE document on configuring your back-end." << std::endl;
    std::cerr << "     Please see      <http://emdros.org/whattodo.html>" << std::endl;
    std::cerr << "     for more information on this problem." << std::endl;
    std::cerr << "Also, please be sure to download the full documentation, available from" << std::endl;
    std::cerr << "<http://emdros.org/>.  You won't be able to make Emdros useful without it." << std::endl;

    delete pDB;
    return 1;
  }

  // Test string_func.cpp
  if (bContinue) {
    long valueLong;
    monad_m valueMonad_m;
    id_d_t valueID_D;
    valueLong = string2long("100");
    bContinue = bContinue && valueLong == 100;
    if (!bContinue) {
      std::cerr << "FAILURE: string2long failed." << std::endl;
    }
    valueMonad_m = string2monad_m("100");
    bContinue = bContinue && valueMonad_m == 100;
    if (!bContinue) {
      std::cerr << "FAILURE: string2monad_m failed." << std::endl;
    }
    valueID_D = string2id_d(NIL_AS_VISIBLE_STRING);
    bContinue = bContinue && valueID_D == NIL;
    if (!bContinue) {
      std::cerr << "FAILURE: string2id_d failed on NIL." << std::endl;
    }
    valueID_D = string2id_d("10000");
    bContinue = bContinue && valueID_D == 10000;
    if (!bContinue) {
      std::cerr << "FAILURE: string2id_d failed on non-NIL id_d." << std::endl;
    }

    std::string strValue;
    strValue = monad_m2string(1);
    bContinue = bContinue && strValue == "1";
    if (!bContinue) {
      std::cerr << "FAILURE: monad_m2string." << std::endl;
    }
    strValue = id_d2string(NIL);
    bContinue = bContinue && strValue == NIL_AS_VISIBLE_STRING;
    if (!bContinue) {
      std::cerr << "FAILURE: id_d2string failed on NIL." << std::endl;
    }
    strValue = id_d2string(1);
    bContinue = bContinue && strValue == "1";
    if (!bContinue) {
      std::cerr << "FAILURE: id_d2string failed on non-NIL id_d." << std::endl;
    }
    strValue = long2string(1);
    bContinue = bContinue && strValue == "1";
    if (!bContinue) {
      std::cerr << "FAILURE: long2string." << std::endl;
    }
    // strcmp_nocase
    valueLong = strcmp_nocase("WxQt", "WXQt");
    bContinue = bContinue && valueLong == 0;
    if (!bContinue) {
      std::cerr << "FAILURE: strcmp_nocase(\"WxQt\", \"WXQt\")." << std::endl;
    }
    valueLong = strcmp_nocase("WxYq", "WXQt");
    bContinue = bContinue && valueLong == 1;
    if (!bContinue) {
      std::cerr << "FAILURE: strcmp_nocase(\"WxYq\", \"WXQt\")." << std::endl;
    }
    valueLong = strcmp_nocase("WxQt", "WXYq");
    bContinue = bContinue && valueLong == -1;
    if (!bContinue) {
      std::cerr << "FAILURE: strcmp_nocase(\"WxQt\", \"WXYq\")." << std::endl;
    }
    // strcmp_case
    valueLong = strcmp_case("WxQt", "WXQt");
    bContinue = bContinue && valueLong == 1;
    if (!bContinue) {
      std::cerr << "FAILURE: strcmp_case(\"WxQt\", \"WXQt\")." << std::endl;
    }
    valueLong = strcmp_case("WxQt", "WxQt");
    bContinue = bContinue && valueLong == 0;
    if (!bContinue) {
      std::cerr << "FAILURE: strcmp_case(\"WxQt\", \"WxQt\")." << std::endl;
    }
    valueLong = strcmp_case("WXYq", "WXQt");
    bContinue = bContinue && valueLong == 1;
    if (!bContinue) {
      std::cerr << "FAILURE: strcmp_case(\"WXYq\", \"WXQt\")." << std::endl;
    }
    valueLong = strcmp_case("WXQt", "WXYq");
    bContinue = bContinue && valueLong == -1;
    if (!bContinue) {
      std::cerr << "FAILURE: strcmp_case(\"WXQt\", \"WXYq\")." << std::endl;
    }
    
    if (bContinue) {
      std::cout << "SUCCESS: all string_func functions pass." << std::endl;
    } else {
      bAllOK = false;
    }
  }
  

  // Test database creation
  if (bContinue) {
    if (pDB->CreateDatabase(newdb)) {
      std::cout << "SUCCESS: " << newdb << " created!\n";
      result = 0;
    } else {
      if (pDB->DropDatabase(newdb)) {
	if (pDB->CreateDatabase(newdb)) {
	  std::cout << "SUCCESS: " << newdb << " created!\n";
	  result = 0;
	} else {
	  std::cerr << "FAILURE: Could not create " << newdb << ".\n";
	  DB_ERROR_MESSAGE;
	}
      } else {
	std::cerr << "FAILURE: Could not create " << newdb << ", or drop it.\n";
	DB_ERROR_MESSAGE;
      }
    }
  }

  if (bContinue) {
    if (pDB->UseDatabase(newdb)) {
      std::cout << "SUCCESS: Database context switch to database '" << newdb << "' succeeded.\n";
    } else {
      std::cerr << "FAILURE: Database context switch to database '" << newdb << "' failed.\n";
      DB_ERROR_MESSAGE;
    }
  }

  id_d_t id_d;
  if (bContinue) {
    if (pDB->GetNextObjectID_D(id_d)) {
      if (id_d == SEQUENCE_START) {
	std::cout << "SUCCESS: object id_d = " << id_d << " == SEQUENCE_START" << " in GetNextObjectID_D." << std::endl;
      } else {
	std::cerr << "FAILURE: first object id_d = " << id_d << " != SEQUENCE_START == " << SEQUENCE_START << " in GetNextObjectID_D." << std::endl;
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: GetNextObjectID_D failed!\n";
      DB_ERROR_MESSAGE;
    }
  }

  if (bContinue) {
    if (pDB->GetNextObjectID_D(id_d)) {
      std::cout << "SUCCESS: object id_d = " << id_d << " in GetNextObjectID\n";
    } else {
      std::cerr << "FAILURE: GetNextObjectID_D failed!\n";
      DB_ERROR_MESSAGE;
    }
  }

  if (bContinue) {
    if (pDB->GetNextOtherID(id_d)) {
      std::cout << "SUCCESS: other id_d = " << id_d << " in GetNextOtherID\n";
    } else {
      std::cerr << "FAILURE: GetNextOtherID failed!\n";
      DB_ERROR_MESSAGE;
    }
  }
  
  // Create enumeration
  long enum_id;
  if (bContinue) {
    if (pDB->CreateEnum("phrase_type_t", enum_id)) {
      std::cout << "SUCCESS: creating enum \n";
    } else {
      std::cerr << "FAILURE: creating enum \n";
      DB_ERROR_MESSAGE;
    }
  }

  // Check that enum exists
  bool enum_exists;
  long second_enum_id;
  if (bContinue) {
    if (pDB->EnumExists("phrase_type_t", enum_exists, second_enum_id)) {
      if (enum_exists && second_enum_id == enum_id)
	std::cout << "SUCCESS: checking existence of enum\n";
      else {
	std::cerr << "FAILURE: checking existence of enum: enum_exists = " << enum_exists << " second_enum_id = " << second_enum_id << " enum_id = " << enum_id << std::endl;
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE in Checking existence of enum \n";
      DB_ERROR_MESSAGE;
    }
  }

  // Create enumeration constants
  if (bContinue) {
    if (pDB->CreateEnumConstant(enum_id, enum_const_info("PT_NotAppliccable", -1, true))
	&& pDB->CreateEnumConstant(enum_id, enum_const_info("VP", 1, false))
	&& pDB->CreateEnumConstant(enum_id, enum_const_info("NP", 3, false))) {
      std::cout << "SUCCESS: creating enumeration constants\n";
    } else {
      std::cerr << "FAILURE: creating enumeration constants\n";
      DB_ERROR_MESSAGE;
    }
  }
  
  // Check get default
  bool get_default_result;
  std::string default_constant_name;
  if (bContinue) {
    if (pDB->GetDefault(enum_id, get_default_result, default_constant_name)) {
      if (get_default_result && default_constant_name == "PT_NotAppliccable")
	std::cout << "SUCCESS: checking default enumeration constant\n";
      else {
	std::cerr << "FAILURE: checking default enumeration constant: default_constant_name = " << default_constant_name << " get_default_result = " << get_default_result << "\n";
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE in checking default enumeration constant\n";
      DB_ERROR_MESSAGE;
    }
  }

  // Check that constant exists
  bool enum_const_exists_result;
  long NP_enum_const_value;
  bool is_default;
  if (bContinue) {
    if (pDB->EnumConstExists("NP", 
			     enum_id,
			     enum_const_exists_result,
			     NP_enum_const_value,
			     is_default)) {
      if (enum_const_exists_result && NP_enum_const_value == 3 && !is_default)
	std::cout << "SUCCESS: checking existence of enumeration constant\n";
      else {
	std::cerr << "FAILURE: checking existence enumeration constant: result = " << enum_const_exists_result << " value = " << NP_enum_const_value << " is_default = " << is_default << std::endl;
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: checking existence of enumeration constant\n";
      DB_ERROR_MESSAGE;
    }
  }

  // Check that constant exists
  std::string enum_const_name;
  if (bContinue) {
    if (pDB->EnumConstExists(1, 
			     enum_id,
			     enum_const_exists_result,
			     enum_const_name,
			     is_default)) {
      if (enum_const_exists_result && enum_const_name == "VP" && !is_default)
	std::cout << "SUCCESS: checking existence of enumeration constant \"VP\"\n";
      else {
	std::cerr << "FAILURE: checking existence enumeration constant \"VP\": result = " << enum_const_exists_result << " name = " << enum_const_name << " is_default = " << is_default << std::endl;
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: checking existence of enumeration constant \"VP\"\n";
      DB_ERROR_MESSAGE;
    }
  }

  // Check that constant exists
  if (bContinue) {
    if (pDB->GetEnumConstNameFromValue(3,
				       std::string("phrase_type_t"),
				       enum_const_name)) {
      if (enum_const_name == "NP")
	std::cout << "SUCCESS: getting string for enum-value 3\n";
      else {
	std::cerr << "FAILURE: getting string for enum-value 3: name = " << enum_const_name << " is_default = " << is_default << std::endl;
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: getting string for enum-value 3\n";
      DB_ERROR_MESSAGE;
    }
  }

  if (bContinue) {
    if (pDB->UpdateEnumConst("NP", enum_id, NP_enum_const_value, 2)) {
      std::cout << "SUCCESS: updating enumeration constant\n";
    } else {
      std::cerr << "FAILURE: updating enumeration constant\n";
      DB_ERROR_MESSAGE;
    }
  }

  // Check that it was really updated
  if (bContinue) {
    if (pDB->EnumConstExists("NP", 
			     enum_id,
			     enum_const_exists_result,
			     NP_enum_const_value,
			     is_default)) {
      if (enum_const_exists_result && NP_enum_const_value == 2 && !is_default)
	std::cout << "SUCCESS: 2nd checking existence of enumeration constant\n";
      else {
	std::cerr << "FAILURE: 2nd checking existence enumeration constant: result = " << enum_const_exists_result << " value = " << NP_enum_const_value << " is_default = " << is_default << std::endl;
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: 2nd checking existence of enumeration constant\n";
      DB_ERROR_MESSAGE;
    }
  }

  // Check setting default
  if (bContinue) {
    if (pDB->SetDefaultEnumConst(enum_id, "VP")
	&& pDB->GetDefault(enum_id, get_default_result, default_constant_name)) {
      if (get_default_result && default_constant_name == "VP")
	std::cout << "SUCCESS: setting default enum constant\n";
      else {
	std::cerr << "FAILURE: setting default enumeration constant: result = " << get_default_result << " default_constant_name = " << default_constant_name << std::endl;
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: setting default enumeration constant\n";
      DB_ERROR_MESSAGE;
    }
  }

  // Check create object type
  id_d_t phrase_object_type_id;
  std::list<feature_info> feature_infos;
  // feature "phrase_type"
  feature_info f_info_phrase_type("phrase_type", enum_id, "-1", false);
  feature_infos.push_back(f_info_phrase_type);

  // feature "NP_type"
  feature_info f_info_NP_type("NP_type", encode_feature_type_id(FEATURE_TYPE_INTEGER), "0", false);
  feature_infos.push_back(f_info_NP_type);

  // feature "Description"
  feature_info f_info_description("Description", encode_feature_type_id(FEATURE_TYPE_STRING, 40*2), "This is a missing description!",false);
  feature_infos.push_back(f_info_description);

  // feature "Parent"
  feature_info f_info_parent("Parent", encode_feature_type_id(FEATURE_TYPE_ID_D), "0", false);
  feature_infos.push_back(f_info_parent);

  if (bContinue) {
    if (pDB->CreateObjectType("Phrase", feature_infos, phrase_object_type_id)) {
      std::cout << "SUCCESS: creating object type with id_d " << phrase_object_type_id << std::endl;
    } else {
      std::cerr << "FAILURE: creating object type" << std::endl;
      DB_ERROR_MESSAGE;
    }
  }

  // Check existence
  long second_phrase_object_type_id;
  bool object_type_exists_result;
  if (bContinue) {
    if (pDB->ObjectTypeExists("Phrase", object_type_exists_result, second_phrase_object_type_id)) {
      if (object_type_exists_result &&
	  second_phrase_object_type_id == phrase_object_type_id)
	std::cout << "SUCCESS: checking existence of object type with id_d " << phrase_object_type_id << std::endl;
      else {
	std::cerr << "FAILURE: checking existence of object type: result = " << object_type_exists_result << " second_phrase_object_type_id = " << second_phrase_object_type_id << std::endl;
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: checking existence of object type" << std::endl;
      DB_ERROR_MESSAGE;
    }
  }

  // Check FeatureExists
  bool feature_exists_result;
  id_d_t feature_type_id;
  std::string default_feature_value;
  bool is_computed;
  if (bContinue) {
    if (pDB->FeatureExists("Parent", phrase_object_type_id, feature_exists_result, feature_type_id, default_feature_value, is_computed)) {
      if (feature_exists_result && !is_computed 
	  && default_feature_value == "0")
	std::cout << "SUCCESS: checking existence of feature with id_d " << feature_type_id << std::endl;
      else {
	std::cerr << "FAILURE: checking existence of feature: result = " << feature_exists_result << " feature_type_id = " << feature_type_id << " default_feature_value = " <<  default_feature_value << " is_computed = " << is_computed << std::endl;
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: checking existence of feature" << std::endl;
      DB_ERROR_MESSAGE;
    }
  }

  // Create "Word" object type
  feature_infos.erase(feature_infos.begin(), feature_infos.end());
  // feature "surface"
  feature_info f_info_surface("surface", encode_feature_type_id(FEATURE_TYPE_STRING, 40), "", false);
  feature_infos.push_back(f_info_surface);

  // feature "lemma"
  feature_info f_info_lemma("lemma", encode_feature_type_id(FEATURE_TYPE_STRING, 40), "", false);
  feature_infos.push_back(f_info_lemma);

  // feature "Description"
  feature_info f_info_description2("Description", encode_feature_type_id(FEATURE_TYPE_STRING, 40*2), "This is a missing description!", false);
  feature_infos.push_back(f_info_description2);

  // feature "Parent"
  feature_info f_info_parent2("Parent", encode_feature_type_id(FEATURE_TYPE_ID_D), NIL_AS_STRING, false);
  feature_infos.push_back(f_info_parent2);

  long word_object_type_id;
  if (bContinue) {
    if (pDB->CreateObjectType("Word", feature_infos, word_object_type_id)) {
      std::cout << "SUCCESS: creating object type 'Word' with id_d " << word_object_type_id << std::endl;
    } else {
      std::cerr << "FAILURE: creating object type 'Word'" << std::endl;
      DB_ERROR_MESSAGE;
    }
  }

  // Check adding feature
  // feature "Parallax"
  feature_info f_info_parallax("Parallax", encode_feature_type_id(FEATURE_TYPE_ID_D), NIL_AS_STRING, false);
  if (bContinue) {
    if (pDB->AddFeature("Word", word_object_type_id, f_info_parallax)) {
      std::cout << "SUCCESS: Adding feature parallax.\n";
    } else {
      std::cerr << "FAILURE: Adding feature parallax.\n";
      DB_ERROR_MESSAGE;
    }
  }

  // Check dropping feature
  if (bContinue) {
    if (pDB->DropFeature("Word", "Parallax", word_object_type_id)) {
      std::cout << "SUCCESS: Dropping feature parallax.\n";
    } else {
      std::cerr << "FAILURE: Dropping feature parallax.\n";
      DB_ERROR_MESSAGE;
    }
  }

  // Check create covered by relationship
  if (bContinue) {
    if (pDB->CreateCoveredByRelationship(phrase_object_type_id, word_object_type_id)) {
      std::cout << "SUCCESS: creating covered_by relationship" << std::endl;
    } else {
      std::cerr << "FAILURE: creating covered_by relationship" << std::endl;
      DB_ERROR_MESSAGE;
    }
  }

  // Check create buildable from relationship
  if (bContinue) {
    if (pDB->CreateBuildableFromRelationship(phrase_object_type_id, word_object_type_id)) {
      std::cout << "SUCCESS: creating buildable_from relationship" << std::endl;
    } else {
      std::cerr << "FAILURE: creating buildable_from relationship" << std::endl;
      DB_ERROR_MESSAGE;
    }
  }

  // Checking that "Covered_by" relationship exists
  bool covered_by_exists1, covered_by_exists2;
  if (bContinue) {
    if (pDB->CoveredByRelationshipExists(phrase_object_type_id, word_object_type_id, covered_by_exists1)
	&& pDB->CoveredByRelationshipExists(phrase_object_type_id, 2543453, covered_by_exists2)) {
      if (covered_by_exists1 && !covered_by_exists2)
	std::cout << "SUCCESS: checking existence of covered_by relationship" << std::endl;
      else {
	std::cerr << "FAILURE: checking existence of covered_by relationship: 1 = " << covered_by_exists1 << " 2 = " << covered_by_exists2 << std::endl;
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: checking existence of covered_by relationship" << std::endl;
      DB_ERROR_MESSAGE;
    }
  }

  // Checking that "buildable_from" relationship exists
  bool buildable_from_exists1, buildable_from_exists2;
  if (bContinue) {
    if (pDB->BuildableFromRelationshipExists(phrase_object_type_id, word_object_type_id, buildable_from_exists1)
	&& pDB->BuildableFromRelationshipExists(phrase_object_type_id, 2543453, buildable_from_exists2)) {
      if (buildable_from_exists1 && !buildable_from_exists2)
	std::cout << "SUCCESS: checking existence of buildable_from relationship" << std::endl;
      else {
	std::cerr << "FAILURE: checking existence of buildable_from relationship: 1 = " << buildable_from_exists1 << " 2 = " << buildable_from_exists2 << std::endl;
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: checking existence of buildable_from relationship" << std::endl;
      DB_ERROR_MESSAGE;
    }
  }

  // Checking dropping of covered_by relationship
  if (bContinue) {
    if (pDB->DropCoveredByRelationship(phrase_object_type_id, word_object_type_id)) {
      std::cout << "SUCCESS: dropping covered_by relationship" << std::endl;
      
    } else {
      std::cerr << "FAILURE: dropping covered_by relationship" << std::endl;
      DB_ERROR_MESSAGE;
    }
  }

  // Checking dropping of buildable_from relationship
  if (bContinue) {
    if (pDB->DropBuildableFromRelationship(phrase_object_type_id, word_object_type_id)) {
      std::cout << "SUCCESS: dropping buildable_from relationship" << std::endl;
      
    } else {
      std::cerr << "FAILURE: dropping buildable_from relationship" << std::endl;
      DB_ERROR_MESSAGE;
    }
  }

  // Checking dropping of covered_by relationship
  if (bContinue) {
    if (pDB->DropCoveredByRelationship(phrase_object_type_id, word_object_type_id)) {
      std::cout << "SUCCESS: 2nd dropping of covered_by relationship" << std::endl;
      
    } else {
      std::cerr << "FAILURE: 2nd dropping covered_by relationship" << std::endl;
      DB_ERROR_MESSAGE;
    }
  }



  // Checking that "Covered_by" relationship exists
  if (bContinue) {
    if (pDB->CoveredByRelationshipExists(phrase_object_type_id, word_object_type_id, covered_by_exists1)) {
      if (!covered_by_exists1)
	std::cout << "SUCCESS: 2nd checking existence of covered_by relationship" << std::endl;
      else {
	std::cerr << "FAILURE: 2nd checking existence of covered_by relationship: 1 = " << covered_by_exists1 << std::endl;
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: 2nd checking existence of covered_by relationship" << std::endl;
      DB_ERROR_MESSAGE;
    }
  }

  // Checking that "buildable_from" relationship exists
  if (bContinue) {
    if (pDB->BuildableFromRelationshipExists(phrase_object_type_id, word_object_type_id, buildable_from_exists1)) {
      if (!buildable_from_exists1)
	std::cout << "SUCCESS: 2nd checking existence of buildable_from relationship" << std::endl;
      else {
	std::cerr << "FAILURE: 2nd checking existence of buildable_from relationship: 1 = " << buildable_from_exists1 << std::endl;
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: 2nd checking existence of buildable_from relationship" << std::endl;
      DB_ERROR_MESSAGE;
    }
  }

  // Check creation of object
  // Create phrase
  id_d_t phrase1;
  if (bContinue) {
    if (!pDB->GetNextObjectID_D(phrase1)) {
      std::cerr << "FAILURE: Error in getting oid for phrase (GetNextObjectID_D)." << std::endl;
      DB_ERROR_MESSAGE;
    }
  }

  CSet_of_monad_ms monads3;
  monads3.add(1,2);
  std::list<feature_info> features;
  features.push_back(feature_info("phrase_type", encode_feature_type_id(enum_id), "1", false));
  features.push_back(feature_info("NP_type", encode_feature_type_id(FEATURE_TYPE_INTEGER), "0", false));
  features.push_back(feature_info("Description", encode_feature_type_id(FEATURE_TYPE_STRING, 40*2), "This is a missing description!",false));
  features.push_back(feature_info("Parent", FEATURE_TYPE_ID_D, NIL_AS_STRING, false));
  if (bContinue) {
    if (pDB->CreateObject(phrase1, "Phrase", monads3, features)) {
      std::cout << "SUCCESS: creating object of type Phrase" << std::endl;
    } else {
      std::cerr << "FAILURE: creating object of type Phrase" << std::endl;
      DB_ERROR_MESSAGE;
    }
  }

  // Create word1
  id_d_t word1;
  if (bContinue) {
    if (!pDB->GetNextObjectID_D(word1)) {
      std::cerr << "FAILURE: getting id_d for word1 in GetNextObjectID_D." << std::endl;
      DB_ERROR_MESSAGE;
    }
  }
  CSet_of_monad_ms monads1(1);
  features.erase(features.begin(), features.end());
  features.push_back(feature_info("surface", encode_feature_type_id(FEATURE_TYPE_STRING, 40*2), "Very", false));
  features.push_back(feature_info("lemma", encode_feature_type_id(FEATURE_TYPE_STRING), "very", false));
  features.push_back(feature_info("Description", encode_feature_type_id(FEATURE_TYPE_STRING), "Adverb", false));
  features.push_back(feature_info("Parent", FEATURE_TYPE_ID_D, NIL_AS_STRING, false));
  if (bContinue) {
    if (pDB->CreateObject(word1, "Word", monads1, features)) {
      std::cout << "SUCCESS: creating 1st object of type Word" << std::endl;
    } else {
      std::cerr << "FAILURE: creating 1st object of type Word" << std::endl;
      DB_ERROR_MESSAGE;
    }
  }
  
  // Create 2nd word
  features.erase(features.begin(), features.end());
  id_d_t word2;
  if (bContinue) {
    if (!pDB->GetNextObjectID_D(word2)) {
      std::cerr << "FAILURE: Error in getting oid for word2 (GetNextObjectID_D)." << std::endl;
      DB_ERROR_MESSAGE;
    }
  }
  CSet_of_monad_ms monads2(2);
  features.push_back(feature_info("surface", encode_feature_type_id(FEATURE_TYPE_STRING, 40), "good", false));
  features.push_back(feature_info("lemma", encode_feature_type_id(FEATURE_TYPE_STRING, 40), "good", false));
  features.push_back(feature_info("Description", encode_feature_type_id(FEATURE_TYPE_STRING, 40), "Adjective", false));
  features.push_back(feature_info("Parent", encode_feature_type_id(FEATURE_TYPE_ID_D), NIL_AS_STRING, false));
  if (bContinue) {
    if (pDB->CreateObject(word2, "Word", monads2, features)) {
      std::cout << "SUCCESS: creating 2nd object of type Word" << std::endl;
    } else {
      std::cerr << "FAILURE: creating 2nd object of type Word" << std::endl;
      DB_ERROR_MESSAGE;
    }
  }



  // Checking ObjectID_DExists
  bool oidExists;
  id_d_t word1Type;
  if (bContinue) {
    if (pDB->ObjectID_DExists(word1, oidExists, word1Type)) {
      if (oidExists && word1Type == word_object_type_id) {
	std::cout << "SUCCESS: checking existence of id_d" << std::endl;
      } else {
	std::cerr << "FAILURE: checking existence of id_d " << std::endl;
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: checking existence of id_d " << std::endl;
      DB_ERROR_MESSAGE;
    }
  }

  // Check segments
  bool bThereAreSegments;
  if (bContinue) {
    if (!pDB->ThereAreSegments(bThereAreSegments)) {
      std::cerr << "FAILURE: Couldn't check for existence of segments.\n";
      DB_ERROR_MESSAGE;
    }  else {
      if (bThereAreSegments) {
	std::cerr << "FAILURE: There shouldn't have been any segments...\n";
	VALUE_ERROR;
      } else
	std::cout << "SUCCESS: There aren't any segments yet.\n";
    }
  }

  // Make segment
  if (bContinue) {
    if (pDB->CreateSegment("OT", 0, 500000)) {
      std::cout << "SUCCESS: Created segment OT (0,500000)\n";
    } else {
      std::cerr << "FAILURE: Couldn't create segment OT.\n";
      DB_ERROR_MESSAGE;
    }
  }

  if (bContinue) {
    if (!pDB->ThereAreSegments(bThereAreSegments)) {
      std::cerr << "FAILURE: Couldn't check whether there were any segments.\n";
      DB_ERROR_MESSAGE;
    } else {
      if (bThereAreSegments)
	std::cout << "SUCCESS: There are segments.\n";
      else {
	std::cerr << "FAILURE: There should have been segments...\n";
	VALUE_ERROR;
      }
    }
  }

  // Make segment
  if (bContinue) {
    if (pDB->CreateSegment("NT", 700000, 900000)) {
      std::cout << "SUCCESS: Created segment NT (700000,900000)\n";
    } else {
      std::cerr << "FAILURE: Couldn't create segment NT.\n";
      DB_ERROR_MESSAGE;
    }
  }

  // Check for overlap
  bool bNoOverlap;
  if (bContinue) {
    if (pDB->SegmentDoesNotOverlap(600000, 800000, bNoOverlap)) {
      if (!bNoOverlap) {
	std::cout << "SUCCESS: There was overlap with (600000, 800000).\n";
      } else {
	std::cerr << "FAILURE: There should have been overlap with (600000, 800000).\n";
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: Couldn't check for overlap.\n";
      DB_ERROR_MESSAGE;
    }
  }

  // Check for overlap
  if (bContinue) {
    if (pDB->SegmentDoesNotOverlap(650000, 680000, bNoOverlap)) {
      if (bNoOverlap) {
	std::cout << "SUCCESS: There was no overlap with (650000, 680000).\n";
      } else {
	std::cerr << "FAILURE: There shouldn't have been overlap with (650000, 680000).\n";
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: Couldn't check for overlap.\n";
      DB_ERROR_MESSAGE;
    }
  }

  // Check for overlap
  if (bContinue) {
    if (pDB->SegmentDoesNotOverlap(1000000, 1400000, bNoOverlap)) {
      if (bNoOverlap) {
	std::cout << "SUCCESS: There was no overlap with (1000000, 1400000).\n";
      } else {
	std::cerr << "FAILURE: There shouldn't have been overlap with (1000000, 1400000).\n";
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: Couldn't check for overlap.\n";
      DB_ERROR_MESSAGE;
    }
  }

  // Checking for existence of segment
  bool bSegmentExists;
  monad_m NT_first, NT_last;
  if (bContinue) {
    if (pDB->SegmentExists("NT", bSegmentExists, NT_first, NT_last)) {
      if (bSegmentExists) {
	if (NT_first == 700000 && NT_last == 900000) {
	  std::cout << "SUCCESS: Segment NT existed with the right borders.\n";
	}	else {
	  std::cerr << "FAILURE: Segment NT existed with the wrong borders.\n";
	  VALUE_ERROR;
	}
      } else {
	std::cerr << "FAILURE: Segment NT did not exist.\n";
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: Could not check existence of segment.\n";
      DB_ERROR_MESSAGE;
    }
  }

  // Getting segment list
  std::list<segment_info> segment_info_list;
  if (bContinue) {
    if (pDB->GetSegments(segment_info_list)) {
      if (segment_info_list.size() == 2) {
	std::cout << "SUCCESS: Segment info list is:\n";
	std::list<segment_info>::const_iterator sici;
	sici = segment_info_list.begin();
	while (sici != segment_info_list.end()) {
	  std::cout << "         ('" << sici->get_name() << "', " << sici->get_first() << ", " << sici->get_last() << ")\n";
	  ++sici;
	}
      } else {
	std::cerr << "FAILURE: Segment info list wasn't 2 long.\n";
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: Could not get segments.\n";
      DB_ERROR_MESSAGE;
    }
  }


  // Dropping a segment
  if (bContinue) {
    if (pDB->DropSegment("NT")) {
      std::cout << "SUCCESS: dropping segment NT.\n";
    } else {
      std::cerr << "FAILURE: Could not drop segment NT (DropSegment).\n";
      DB_ERROR_MESSAGE;
    }
  }

  // Check that it is gone
  if (bContinue) {
    if (pDB->SegmentExists("NT", bSegmentExists, NT_first, NT_last)) {
      if (bSegmentExists) {
	std::cerr << "FAILURE: Segment NT existed after dropping it (SegmentExists).\n";
	VALUE_ERROR;
      } else {
	std::cout << "SUCCESS: Segment NT did not exist any longer.\n";
      }
    } else {
      std::cerr << "FAILURE: Could not check existence of segment.\n";
      DB_ERROR_MESSAGE;
    }
  }


  // Check inst
  IDDList inst;
  if (bContinue) {
    if (pDB->GetObjectsInUniverse("Word", 0, 3, inst)) {
      if (inst.length() == 2) {
	std::cout << "SUCCESS: checking GetObjectsInUniverse\n";
      } else {
	std::cerr << "FAILURE: Inst existed, but had " << inst.length() << " members.  It should have had 2.\n";
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: Could not get inst\n";
      DB_ERROR_MESSAGE;
    }
  }

  // Check getting objects starting at Sm
  IDDList iddlist;
  if (bContinue) {
    if (pDB->GetObjectsStartingAtSm("Word", 2, iddlist)) {
      if (iddlist.length() == 1) {
	if (*(iddlist.begin()) == word2) {
	  std::cout << "SUCCESS: IDDList had 1 member which was " << word2 << ". (GetObjectsStartingAtSm)." << std::endl;
	} else {
	  std::cerr << "FAILURE: IDDList had 1 member which was not " << word2 << " but " << *(iddlist.begin()) << ". (GetObjectsStartingAtSm)." << std::endl;
	  VALUE_ERROR;
	}
      } else {
	std::cerr << "FAILURE: IDDList was not 1 long (GetObjectsStartingAtSm)." << std::endl;
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: Could not get objects starting at 1. (GetObjectsStartingAtSm)." << std::endl;
      DB_ERROR_MESSAGE;
    }
  }

  // Getting SOM for object word2
  CSet_of_monad_ms som2;
  if (bContinue) {
    if (pDB->GetSOMForObject("Word", word2, som2)) {
      if (som2 == monads2) {
	std::cout << "SUCCESS: Getting SOM for word2 was OK." << std::endl;
      } else {
	std::cerr << "FAILURE: Getting SOM for word2 was wrong." << std::endl;
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: Could not get SOM for word2" << std::endl;
      DB_ERROR_MESSAGE;
    }
  }

  // Getting min_m
  monad_m min_m;
  if (bContinue) {
    if (pDB->Get_min_m(min_m)) {
#define EXPECTED_MIN_M  (1)
      if (min_m == EXPECTED_MIN_M) {
	std::cout << "SUCCESS: min_m is " << min_m << " which it should be." << std::endl;
      } else {
	std::cerr << "FAILURE: min_m is not " << EXPECTED_MIN_M << " but " << min_m << "." << std::endl;
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: Could not get min_m" << std::endl;
      DB_ERROR_MESSAGE;
    }
  }

  // Getting max_m
  monad_m max_m;
  if (bContinue) {
    if (pDB->Get_max_m(max_m)) {
#define EXPECTED_MAX_M  (2)
      if (max_m == EXPECTED_MAX_M) {
	std::cout << "SUCCESS: max_m is " << max_m << " which it should be." << std::endl;
      } else {
	std::cerr << "FAILURE: max_m is not " << EXPECTED_MAX_M << " but " << max_m << "." << std::endl;
	VALUE_ERROR;
      }
    } else {
      std::cerr << "FAILURE: Could not get max_m" << std::endl;
      DB_ERROR_MESSAGE;
    }
  }

  // Get object types
  std::list<std::string> object_types_list;
  if (bContinue) {
    if (pDB->GetObjectTypes(object_types_list)) {
      if (object_types_list.size() == 2) {
	std::cout << "SUCCESS: Object types list is:\n";
      } else {
	std::cerr << "FAILURE: Object types list wasn't 2 long.\n";
	VALUE_ERROR;
      }
      std::list<std::string>::const_iterator otci;
      otci = object_types_list.begin();
      while (otci != object_types_list.end()) {
	std::cout << "         '" << *otci << "'" << std::endl;
	++otci;
      }
    } else {
      std::cerr << "FAILURE: Couldn't get object types.\n";
      DB_ERROR_MESSAGE;
    }
  }

  // Check getting features
  features.erase(features.begin(), features.end());
  if (bContinue) {
    if (pDB->GetFeaturesForObjectType(word_object_type_id,
				      features)) {
      if (features.size() == 5) {
	std::cout << "SUCCESS: Features of object type Word are:\n";
      } else {
	std::cerr << "FAILURE: Feature-list of object type Word wasn't 5 long.\n";
	VALUE_ERROR;
      }
      std::list<feature_info>::const_iterator ci;
      ci = features.begin();
      while (ci != features.end()) {
	std::cout << "         ('" << ci->get_name() << "', " << ci->get_type() << ", '" << ci->get_default_value() << "', " << ci->get_is_computed() << ")" << std::endl;
	++ci;
      }
    } else {
      std::cerr << "FAILURE: Couldn't get list of features for object type word\n";
      DB_ERROR_MESSAGE;
    }
  }

  // Check getting enumerations
  std::list<std::string> enums;
  if (bContinue) {
    if (pDB->GetEnumerations(enums)) {
      if (enums.size() == 1) {
	std::cout << "SUCCESS: Enumerations are:\n";
      } else {
	std::cerr << "FAILURE: Enumeration-list is not 1 long\n";
	VALUE_ERROR;
      }
      std::list<std::string>::const_iterator ci;
      ci = enums.begin();
      while (ci != enums.end()) {
	std::cout << "         '" << *ci << "'" << std::endl;
	++ci;
      }
    } else {
      std::cerr << "FAILURE: Couldn't get list of enumerations\n";
      DB_ERROR_MESSAGE;
    }
  }
    
  // Check enum constants
  std::list<enum_const_info> enum_consts;
  if (bContinue) {
    if (pDB->GetEnumConstants(enum_id, enum_consts)) {
      if (enum_consts.size() == 3) {
	std::cout << "SUCCESS: Enumeration constants are:\n";
      } else {
	std::cerr << "FAILURE: Enumeration constant list wasn't 3 long.\n";
	VALUE_ERROR;
      }
      std::list<enum_const_info>::const_iterator ci;
      ci = enum_consts.begin();
      while (ci != enum_consts.end()) {
	std::cout << "         ('" << ci->get_name() << "', " << ci->get_value() << ", " << ci->get_is_default() << ")" << std::endl;
	++ci;
      }
    } else {
      std::cerr << "FAILURE: Couldn't get enum constants.\n";
      DB_ERROR_MESSAGE;
    }
  }

  // Check getting object types using enumeration
  object_types_list.erase(object_types_list.begin(), object_types_list.end());
  if (bContinue) {
    if (pDB->GetObjectTypesUsingEnumeration(enum_id, object_types_list)) {
      if (object_types_list.size() == 1) {
	std::cout << "SUCCESS: Object types using enumeration are:\n";
      } else {
	std::cerr << "FAILURE: List of object types using enumeration 1 long.\n";
	VALUE_ERROR;
      }
      std::list<std::string>::const_iterator otci;
      otci = object_types_list.begin();
      while (otci != object_types_list.end()) {
	std::cout << "         '" << *otci << "'" << std::endl;
	++otci;
      }
    } else {
      std::cerr << "FAILURE: Couldn't get object types using enumeration.\n";
      DB_ERROR_MESSAGE;
    }
  }

  




  // Drop database
  if (pDB->DropDatabase(newdb)) {
    std::cout << "SUCCESS in dropping database '" << newdb << "'\n";
  } else {
    std::cerr << "FAILURE in dropping database '" << newdb << "'\n";
    DB_ERROR_MESSAGE;
  }

  if (bAllOK) {
    std::cout << "SUCCESS: All tests completed successfully." << std::endl;
  } else {
    std::cerr << "FAILURE: At least one test failed." << std::endl;
  }

  // Clean up
  delete pDB;
  return result;
}


